/*
 *       Lexer for SDL-2000
 *
 * Copyright (C) 2003 Michael Schmitt <Michael.Schmitt@teststep.org>
 *
 * Credits to Andreas Prinz and Ralf Schroeder from Humboldt University of Berlin
 * for testing and bug fixes.
 *
 * This program is free software; you can redistribute it and/or modify it under the terms
 * of the GNU General Public License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program;
 * if not, write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 *
 * $Id: SDLLexer.g,v 1.17 2003/05/21 21:45:24 schmitt Exp $
 *
 * This SDL-2000 grammar is based on the document 'SDL Forum Society Z.100 (11/99) version 3'
 * aka 'TDG04r22.doc' aka 'z.100_1001.pdf' (SDL-2000 as approved by ITU-T in September 2001).
 *
 * For the latest release of all grammar files please have a look at http://www.teststep.org.
 *
 * This file is intended to be used with ANTLR 2.7.2. ANTLR is a parser generator
 * that is based on 'linear approximate' lookahead analysis (a kind of weak LL(k) analysis).
 * ANTLR is free software provided by jGuru.com (MageLang Institute). For further information
 * please refer to http://www.antlr.org. My special thanks go to Terence Parr from jGuru.com.
 *
 * If you make use of this SDL-2000 grammar specification, please send a little email in which
 * you tell me what you are using it for and whether you consider it useful. Any grammar
 * corrections and extensions (e.g. for semantics checks and abstract syntax tree transformation)
 * are welcome.
 *
 * General considerations:
 *
 *   - the original Z.100 lexical rules have been adopted as much as possible since the SDL
 *     standard is likely to undergo further changes in the future which should be easy to
 *     integrate
 *
 * Restrictions:
 *
 *   - macros are not supported (and will never ever be!)
 *   - uppercase keywords are not accepted
 *   - line numbers are not handled correctly if two lines in the input are concatenated by '_'
 *   - delete characters are treated as common control characters
 *   - extended character sets are not supported
 *
 */

header "pre_include_hpp"
{
  // $Id: SDLLexer.g,v 1.17 2003/05/21 21:45:24 schmitt Exp $

  #include <string>
  #include "antlr/CharBuffer.hpp"

  using std::string;
}


header "post_include_hpp"
{
  namespace SDLGrammar
  {
    class SDLCharBuffer : public antlr::CharBuffer
    {
      public:

        SDLCharBuffer( std::istream &input) : CharBuffer( input ) { }

        int getChar()
        {
                 int   c;
          static int   nextc = 0;

          if ( nextc != 0 ) {
            c = nextc;
            nextc = 0;
          } else {
            c = CharBuffer::getChar();
          }

          if ( c != '_' ) {
            return c;
          }

          nextc = CharBuffer::getChar();
          if ( nextc < 0 || nextc > 32 ) {
            return c;
          }

          do {
            nextc = CharBuffer::getChar();
          } while( nextc >= 0 && nextc <= 32 );

          return getChar();
        }
    };
  }
}


header "post_include_cpp"
{
  #include <iostream>

  using std::cerr;
  using std::endl;

  namespace SDLGrammar
  {
    void SDLLexer::reportError( const string &errorMessage )
    {
      reportMessage( errorMessage );
      noOfErrors++;
    }

    void SDLLexer::reportMessage( const string &message ) const
    {
      if ( getFilename().length() > 0 ) {
        cerr << getFilename();
      } else {
        cerr << "<stdin>";
      }

      cerr << ":" << getLine() << ":" << getColumn() << ": error: " << message << endl;
    }

    unsigned int SDLLexer::numberOfErrors() const
    {
      return noOfErrors;
    }

    void SDLLexer::resetErrors()
    {
      noOfErrors = 0;
    }

    const string &SDLLexer::versionInfo()
    {
      static string   info = "$Id: SDLLexer.g,v 1.17 2003/05/21 21:45:24 schmitt Exp $";

      return info;
    }
  }
}


options {
  language  = "Cpp";
  namespace = "SDLGrammar";
  namespaceStd = "std";
  namespaceAntlr = "antlr";
  genHashLines = true;
}


class SDLLexer extends Lexer;

  options {
    k = 2;
    exportVocab = SDLLexer;
    testLiterals = false;
    caseSensitive = true;
  }

  tokens {
    ABSTRACT            = "abstract";
    ACTIVE              = "active";
    ADDING              = "adding";
    AGGREGATION         = "aggregation";
    ALTERNATIVE         = "alternative";
    AND                 = "and";
    ANY                 = "any";
    AS                  = "as";
    ASSOCIATION         = "association";
    ATLEAST             = "atleast";
    BLOCK               = "block";
    BREAK               = "break";
    CALL                = "call";
    CHANNEL             = "channel";
    CHOICE              = "choice";
    COMMENT             = "comment";
    COMPOSITION         = "composition";
    CONNECT             = "connect";
    CONNECTION          = "connection";
    CONSTANTS           = "constants";
    CONTINUE            = "continue";
    CREATE              = "create";
    DCL                 = "dcl";
    DECISION            = "decision";
    DEFAULT             = "default";
    ELSE                = "else";
    ENDALTERNATIVE      = "endalternative";
    ENDBLOCK            = "endblock";
    ENDCHANNEL          = "endchannel";
    ENDCONNECTION       = "endconnection";
    ENDDECISION         = "enddecision";
    ENDEXCEPTIONHANDLER = "endexceptionhandler";
    ENDINTERFACE        = "endinterface";
    ENDMACRO            = "endmacro";
    ENDMETHOD           = "endmethod";
    ENDOBJECT           = "endobject";
    ENDOPERATOR         = "endoperator";
    ENDPACKAGE          = "endpackage";
    ENDPROCEDURE        = "endprocedure";
    ENDPROCESS          = "endprocess";
    ENDSELECT           = "endselect";
    ENDSTATE            = "endstate";
    ENDSUBSTRUCTURE     = "endsubstructure";
    ENDSYNTYPE          = "endsyntype";
    ENDSYSTEM           = "endsystem";
    ENDTYPE             = "endtype";
    ENDVALUE            = "endvalue";
    ENV                 = "env";
    EXCEPTION           = "exception";
    EXCEPTIONHANDLER    = "exceptionhandler";
    EXPORT              = "export";
    EXPORTED            = "exported";
    EXTERNAL            = "external";
    FI                  = "fi";
    FINALIZED           = "finalized";
    FROM                = "from";
    GATE                = "gate";
    HANDLE              = "handle";
    IF                  = "if";
    IMPORT              = "import";
    IN                  = "in";
    INHERITS            = "inherits";
    INPUT               = "input";
    INTERFACE           = "interface";
    JOIN                = "join";
    LITERALS            = "literals";
    LOOP                = "loop";
    MACRO               = "macro";
    MACRODEFINITION     = "macrodefinition";
    MACROID             = "macroid";
    METHOD              = "method";
    METHODS             = "methods";
    MOD                 = "mod";
    NAMECLASS           = "nameclass";
    NEXTSTATE           = "nextstate";
    NODELAY             = "nodelay";
    NONE                = "none";
    NOT                 = "not";
    NOW                 = "now";
    OBJECT              = "object";
    OFFSPRING           = "offspring";
    ONEXCEPTION         = "onexception";
    OPERATOR            = "operator";
    OPERATORS           = "operators";
    OPTIONAL            = "optional";
    OR                  = "or";
    ORDERED             = "ordered";
    OUT                 = "out";
    OUTPUT              = "output";
    PACKAGE             = "package";
    PARENT              = "parent";
    PRIORITY            = "priority";
    PRIVATE             = "private";
    PROCEDURE           = "procedure";
    PROTECTED           = "protected";
    PROCESS             = "process";
    PROVIDED            = "provided";
    PUBLIC              = "public";
    RAISE               = "raise";
    REDEFINED           = "redefined";
    REFERENCED          = "referenced";
    REM                 = "rem";
    REMOTE              = "remote";
    RESET               = "reset";
    RETURN              = "return";
    SAVE                = "save";
    SELECT              = "select";
    SELF                = "self";
    SENDER              = "sender";
    SET                 = "set";
    SIGNAL              = "signal";
    SIGNALLIST          = "signallist";
    SIGNALSET           = "signalset";
    SIZE                = "size";
    SPELLING            = "spelling";
    START               = "start";
    STATE               = "state";
    STOP                = "stop";
    STRUCT              = "struct";
    SUBSTRUCTURE        = "substructure";
    SYNONYM             = "synonym";
    SYNTYPE             = "syntype";
    SYSTEM              = "system";
    TASK                = "task";
    THEN                = "then";
    THIS                = "this";
    TIMER               = "timer";
    TO                  = "to";
    TRY                 = "try";
    TYPE                = "type";
    USE                 = "use";
    VALUE               = "value";
    VIA                 = "via";
    VIRTUAL             = "virtual";
    WITH                = "with";
    XOR                 = "xor";

    ResultSign              = "->";
    RangeSign               = "..";
    CompositeBeginSign      = "(.";
    CompositeEndSign        = ".)";
    ConcatenationSign       = "//";
    HistoryDashSign         = "-*";
    GreaterThanOrEqualsSign = ">=";
    ImpliesSign             = "=>";
    IsAssignedSign          = ":=";
    LessThanOrEqualsSign    = "<=";
    NotEqualsSign           = "/=";
    QualifierBeginSign      = "<<";
    QualifierEndSign        = ">>";

    ExclamationMark         = "!";
    QuotationMark           = "\"";
    LeftParenthesis         = "(";
    RightParenthesis        = ")";
    Asterisk                = "*";
    PlusSign                = "+";
    Comma                   = ",";
    Hyphen                  = "-";
    FullStop                = ".";
    Solidus                 = "/";
    Colon                   = ":";
    Semicolon               = ";";
    LessThanSign            = "<";
    EqualsSign              = "=";
    GreaterThanSign         = ">";
    LeftSquareBracket       = "[";
    RightSquareBracket      = "]";
    LeftCurlyBracket        = "{";
    RightCurlyBracket       = "}";
    NumberSign              = "#";
  }

{
  public:

    unsigned int numberOfErrors() const;
    void resetErrors();

    static const string &versionInfo();

  protected:

    void reportError( const string &errorMessage );
    void reportMessage( const string &message ) const;

  private:

    unsigned int   noOfErrors;
}


LexicalUnit { int type; } :

    Name                      { $setType(testLiteralsTable(Name)); }
  | QuotedOperationName       { $setType(QuotedOperationName); }
  | ( BitString ) =>
    BitString                 { $setType(BitString); }        // Test 'BitString' ...
  | ( HexString ) =>
    HexString                 { $setType(HexString); }        // and 'HexString' first
  | CharacterString           { $setType(CharacterString); }
  | ( "/*" ) =>
    Note                      { $setType(antlr::Token::SKIP); }
  | CommentBody               { $setType(antlr::Token::SKIP); }
  | ( CompositeSpecial ) =>   // no warning despite ambiguity due to bug in ANTLR!
    type = CompositeSpecial   { $setType(type); }
  | type = Special            { $setType(type); }

  ;

protected Name :

    ( ( '0' .. '9' )+ '.' ) =>
    ( '0' .. '9' )+ ( '.' ( '0' .. '9' )+ )*
  | ( '_' )* Word ( '_' ( Word )? )*   // somewhat simpler than in the SDL-2000 standard

  ;

protected Word :

  ( Alphanumeric )+

  ;

protected Alphanumeric :

    'A' .. 'Z'
  | 'a' .. 'z'
  | '0' .. '9'

  ;

protected QuotedOperationName :

  '"'!
  (   InfixOperationName
    | MonadicOperationName )
  '"'!

  ;

protected InfixOperationName :

    "or"  | "OR"
  | "xor" | "XOR"
  | "and" | "AND"
  | "in"  | "IN"
  | "mod" | "MOD"
  | "rem" | "REM"
  | '+'   | '-'   | '*'
  | "/="  | "<="  | ">="
  | "//"  | "=>"  | '/'
  | '='   | '>'   | '<'

  ;

protected MonadicOperationName :

    // '-'   // already in InfixOperationName
  | "not"
  | "NOT"

  ;

protected CharacterString :

  '\''!
  (   GeneralTextCharacter
    | Special
    | ( '\'' '\'' ) =>
      '\'' '\''! )*
  '\''!

  ;

protected HexString :

  '\''!
  (   '0' .. '9'
    | 'a' .. 'f'
    | 'A' .. 'F' )*
  '\''! ( 'H'! | 'h'! )

  ;

protected BitString :

  '\''!
  (   '0'
    | '1' )*
  '\''! ( 'B'! | 'b'! )

  ;

protected Note :

  '/'! '*'! NoteText '*'! '/'!   // additional '*'s are consumed in 'NoteText'

  ;

protected NoteText :

  ( options { greedy = false; } :
    (   GeneralTextCharacter
      | OtherSpecial
      | '#'
      | '*'
      | '/'
      | '\'' ) )*

  ;

protected Text :

  (   GeneralTextCharacter
    | Special
    | '\'' )*

  ;

protected GeneralTextCharacter :

    Alphanumeric
  | OtherCharacter
  | Space

  ;

protected CommentBody :

  '/'! '#'! CommentText '#'! '/'!

  ;

protected CommentText :

  ( options { greedy = false; } :
    (   GeneralTextCharacter
      | OtherSpecial
      | '*'
      | '#'
      | '/'
      | '\'' ) )*

  ;

protected CompositeSpecial returns [ int type ] :

    "->" { type = ResultSign; }
  | ".." { type = RangeSign; }
  | "(." { type = CompositeBeginSign; }
  | ".)" { type = CompositeEndSign; }
  | "//" { type = ConcatenationSign; }
  | "-*" { type = HistoryDashSign; }
  | ">=" { type = GreaterThanOrEqualsSign; }
  | "=>" { type = ImpliesSign; }
  | ":=" { type = IsAssignedSign; }
  | "<=" { type = LessThanOrEqualsSign; }
  | "/=" { type = NotEqualsSign; }
  | "<<" { type = QualifierBeginSign; }
  | ">>" { type = QualifierEndSign; }

  ;

protected Special returns [ int type ] :

    '/' { type = Solidus; }
  | '*' { type = Asterisk; }
  | '#' { type = NumberSign; }
  | type = OtherSpecial

  ;

protected OtherSpecial returns [ int type ] :

    '!' { type = ExclamationMark; }
  | '(' { type = LeftParenthesis; }
  | ')' { type = RightParenthesis; }
  | '+' { type = PlusSign; }
  | ',' { type = Comma; }
  | '-' { type = Hyphen; }
  | '.' { type = FullStop; }
  | ':' { type = Colon; }
  | ';' { type = Semicolon; }
  | '<' { type = LessThanSign; }
  | '=' { type = EqualsSign; }
  | '>' { type = GreaterThanSign; }
  | '[' { type = LeftSquareBracket; }
  | ']' { type = RightSquareBracket; }
  | '{' { type = LeftCurlyBracket; }
  | '}' { type = RightCurlyBracket; }

  ;

protected OtherCharacter :

    '"'  | '$' | '%' | '&' | '?' | '@'
  | '\\' | '^' | '_' | '`' | '|' | '~'

  ;

protected Space :

    '\000' .. '\011'      // Control characters
  | '\n' { newline(); }
  | '\013' .. '\037'      // Control characters
  | ' '

  ;

WhiteSpace :

  ( Space )+

  { $setType(antlr::Token::SKIP); }

  ;
