/*
 *       Parser for SDL-2000
 *
 * Copyright (C) 2003 Michael Schmitt <Michael.Schmitt@teststep.org>
 *
 * Credits to Andreas Prinz and Ralf Schroeder from Humboldt University of Berlin
 * for testing and bug fixes.
 *
 * This program is free software; you can redistribute it and/or modify it under the terms
 * of the GNU General Public License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program;
 * if not, write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 *
 * $Id: SDLParser.g,v 1.57 2003/05/21 21:31:39 schmitt Exp $
 *
 * This SDL-2000 grammar is based on the document 'SDL Forum Society Z.100 (11/99) version 3'
 * aka 'TDG04r22.doc' aka 'z.100_1001.pdf' (SDL-2000 as approved by ITU-T in September 2001).
 *
 * For the latest release of all grammar files please have a look at http://www.teststep.org.
 *
 * This file is intended to be used with ANTLR 2.7.2. ANTLR is a parser generator
 * that is based on 'linear approximate' lookahead analysis (a kind of weak LL(k) analysis).
 * ANTLR is free software provided by jGuru.com (MageLang Institute). For further information
 * please refer to http://www.antlr.org. My special thanks go to Terence Parr from jGuru.com.
 *
 * If you make use of this SDL-2000 grammar specification, please send a little email in which
 * you tell me what you are using it for and whether you consider it useful. Any grammar
 * corrections and extensions (e.g. for semantics checks and abstract syntax tree transformation)
 * are welcome.
 *
 * General considerations:
 *
 *   - the original Z.100 grammar rules have been adopted as much as possible since the SDL
 *     standard is likely to undergo further changes in the future which should be easy to
 *     integrate
 *   - 'xxxReference's need to be put in front of 'xxxDefinition's
 *     since they can only be distinguished by infinite lookahead over 'xxxReference'
 *   - 'xxxTypeDefinition's are put in front of 'xxxDefinition' for simpler predicates
 *   - 'textualTypebasedXxxDefinition's need to be put in front of 'xxxDefinition's
 *     since they can only be distinguished by the ':' character in the header of the former one
 *   - the SDL parser constructs an abstract syntax tree (AST) based on the token stream delivered
 *     by 'SDLLexer'. However, as no type information about names is available at this stage of
 *     processing, several decisions have to be postponed. E.g., the parser cannot always
 *     determine whether a statement represents a remote or a 'normal' procedure call. Such
 *     ambiguities are resolved by a succeeding tree parser that operates on the intermediate AST.
 *     For a full list of unresolvable nondeterminisms please see below
 *
 * Unsolved problems:
 *
 *   - macros are not supported (and will never ever be!)
 *   - 'package' is not handled correctly in 'sdlSpecification'
 *   - check for comments on syntactical restrictions in the textual part of Z.100
 *   - take construct transformations into account
 *   - build symbol tables
 *   - imprecise line/column information for 'typeReferenceProperties', 'associationEnd',
 *     'primary', 'sortContextParameter' and 'rangeCheck'
 *
 * General problems, inconsistencies and weaknesses of the Z.100 SDL standard:
 *
 *  1. the optional 'dataTypeDefinitionBody' in the first alternative of 'dataTypeDefinition'
 *     is not delimited; this is problematic since 'dataTypeDefinitionBody' may contain a
 *     'dataTypeDefinition' itself; the same problem occurs also within 'interfaceDefinition';
 *     Solution: Clearly state that a 'dataTypeDefinitionBody' belongs to the innermost definition
 *  2. the same name can be used for entities of different kinds; this is unusual and leads
 *     to conflicts with other (programming) languages; moreover it causes several problems
 *     within SDL itself:
 *        - CALL must be mandatory sometimes (but not always) -> complex static semantics
 *        - how can we resolve the potential ambiguities in 'linkedTypeIdentifier',
 *          'channelEndpoint' (BTW: What is the entity kind of 'state'), 'createBody',
 *          'typeReferenceProperties', 'attributeProperty', 'behaviourProperty' ???
 *     Surprisingly, on the other hand a procedure name is not allowed to be reused for procedures
 *     with different signatures :-( But this is a common feature in OO languages like C++ and Java
 *  3. redundant curly brackets in the EBNF of various rules, e.g. in 'name', 'loopStep' and
 *     'compositeStateList'; very confusing
 *  4. 'compositeStateList', 'stateListToken' and 'typebasedStatePartitionHeading'
 *     belong to the concrete textual grammar
 *  5. 'textualTypebasedStatePartitionDefinition' is included in 'statePartitioning' and
 *     hence should not be listed separately in rule 'selectDefinition' (fixed)
 *  6. 'fieldName' should derive to 'name' of type 'field' (the type is unspecified in Z.100; fixed)
 *
 * Parser specific weaknesses of the Z.100 SDL standard:
 *
 *  1. 'procedureBody' must be mandatory in 'procedureDefinition' in order not to
 *     confuse LL(k) parsers as 'procedureBody' may derive to the empty string
 *  2. minor problem: 'operatorList' and 'methodList' should be rewritten for LL(3) parsers;
 *     the same applies for 'structureDefinition' and 'choiceDefinition' (see rules below)
 *  3. the different representations of 'procedureDefinition' and 'operationDefinition' should be
 *     aligned for much simpler parsing (this does not mean only transforming the grammar rules
 *     but changing the language itself)
 *
 * Alternatives commented out/combined for making the grammar nonambiguous
 *    (in cases where one alternative subsumes another)
 *
 *  1. In rule 'state' : 'basicState' (subsumed by 'compositeStateApplication')
 *  2. In rule 'sort' : 'pidSort' (subsumed by 'basicSort')
 *  3. In rule 'actualContextParameter': 'identifier' (subsumed by 'primary')
 *  4. In rule 'linkedTypeIdentifier': identifiers
 *  5. In rule 'channelEndpoint': identifiers
 *  6. In rule 'signalListItem': identifiers
 *  7. In rule 'point': 'stateExitPoint' (subsumed by 'stateEntryPoint')
 *  8. In rule 'createBody': identifiers
 *  9. In rule 'procedureCallBody': identifier (subsumed by 'typeExpression')
 * 10. In rule 'destination': 'identifier' and 'THIS' (subsumed by 'expression')
 * 11. In rule 'viaPath': identifiers
 * 12. In rule 'loopStep': 'procedureCallBody' (subsumed by 'expression')
 * 13. In rule 'basicSort': 'syntype' (subsumed by 'identifier')
 * 14. In rule 'primary': 'fieldName' and 'synonym' (subsumed by 'operatorApplication')
 * 15. In rule 'answer': 'informalText' (subsumed by 'rangeCondition')
 * 16. In rule 'alternativeQuestion': 'informalText' (subsumed by 'simpleExpression')
 * 17. In rule 'question': 'informalText' (subsumed by 'expression')
 *
 * Alternatives guarded by predicates because the grammar is _inherently_ ambig
 *    (in cases where one alternative parses a substring(!) of another or
 *     some input (but not all) is accepted by more than one alternative)
 *
 *  1. In rule 'typeReferenceProperties': 'attributeProperty' vs. 'behaviourProperty'
 *  2. In rule 'attributeProperty': 'variableProperty' vs. 'fieldProperty' vs.
 *                                  'interfaceVariableProperty'
 *  3. In rule 'behaviourProperty': 'operationProperty' vs. 'procedureProperty' vs.
 *                                  'signalProperty' vs. 'exceptionProperty' vs.
 *                                  'timerProperty'
 *  4. In rule 'signalListItem': identifiers
 *  5. In rule 'action1': 'procedureCall' vs. 'remoteProcedureCall'
 *  6. In rule 'statement': 'callStatement' vs. 'expressionStatement'
 *  7. In rule 'renamePair': 'operationName' vs. 'literalName'
 *  8. In rule 'expression': 'valueReturningProcedureCall' vs. 'operand'
 *  9. In rule 'primary': 'operatorApplication' vs. 'literal' vs 'activePrimary'
 * 10. In rule 'primaryExtension': 'operationIdentifier' vs. '*field* Name'
 * 11. In rule 'valueReturningProcedureCall': 'procedureCallBody' vs. 'remoteProcedureCallBody'
 * 12. In rule 'stateListToken' : 'typebasedCompositeState' vs. 'Name ...' (substring)
 * 13. In rule 'question': 'ANY' vs. 'expression' (substring)
 *
 */

header "pre_include_hpp"
{
  // $Id: SDLParser.g,v 1.57 2003/05/21 21:31:39 schmitt Exp $

  #include <string>
  #include "InfoAST.hpp"
  #include "SDLName.h"

  using std::string;
}


header "post_include_cpp"
{
  #include <iostream>

  using std::cerr;
  using std::endl;

  namespace SDLGrammar
  {
    void SDLParser::reportError( const string &errorMessage )
    {
      reportMessage( errorMessage );
      noOfErrors++;
    }

    void SDLParser::reportError( const antlr::RecognitionException &ex )
    {
      cerr << ex.toString().c_str() << endl;
      noOfErrors++;
    }

    void SDLParser::reportMessage( const string &message )
    {
      if ( getFilename().length() > 0 ) {
        cerr << getFilename();
      } else {
        cerr << "<stdin>";
      }

      cerr << ":" << LT( 1 ) -> getLine() << ":" << LT( 1 ) -> getColumn() << ": error: " << message << endl;
    }

    unsigned int SDLParser::numberOfErrors() const
    {
      return noOfErrors;
    }

    void SDLParser::resetErrors()
    {
      noOfErrors = 0;
    }

    const string &SDLParser::versionInfo()
    {
      static string   info = "$Id: SDLParser.g,v 1.57 2003/05/21 21:31:39 schmitt Exp $";

      return info;
    }
  }

  #define ADDVOIDCHILD(tokenast)     astFactory -> addASTChild( currentAST, tokenast )
  #define ADDCHILD(tokenAST,token) { const antlr::RefAST &tmpAST = tokenAST; \
                                     SETLINECOLUMN( tmpAST, token ); \
                                     astFactory -> addASTChild( currentAST, tmpAST ); }
  #define MAKEROOT(tokenAST,token) { const antlr::RefAST &tmpAST = tokenAST; \
                                     SETLINECOLUMN( tmpAST, token ) \
                                     astFactory -> makeASTRoot( currentAST, tmpAST ); }
  #define SETLINECOLUMN(AST,token) { antlr::RefInfoAST( AST ) -> setLine( token -> getLine() ); \
                                     antlr::RefInfoAST( AST ) -> setColumn( token -> getColumn() ); }
}


options {
  language  = "Cpp";
  namespace = "SDLGrammar";
  namespaceStd = "std";
  namespaceAntlr = "antlr";
  genHashLines = false;         // too many compiler warnings when generating #line statements
}


class SDLParser extends Parser;

  options {
    k = 3;
    importVocab = SDLLexer;
    exportVocab = SDLParser;
    defaultErrorHandler = true;
    buildAST = true;
    codeGenMakeSwitchThreshold = 10;
  }

  // The following section defines tokens to be used for new root nodes in the AST
  //
  // Tokens that are not allowed in general for ambiguity reasons:
  //
  //   ActivePrimary, AttributeProperty, BaseType, BasicSort, BasicState, BehaviourProperty,
  //   CallStatement, CompositeStateApplication, CompositeStateList, ExceptionProperty, Expression
  //   ExpressionStatement, FieldName, FieldProperty, FormalParameter, InformalText,
  //   InterfaceVariableProperty, Literal, LiteralIdentifier, LiteralName, Operand, Operand0,
  //   Operand1, Operand2, Operand3, Operand4, Operand5, OperationApplication, OperationIdentifier,
  //   OperationName, OperationProperty, OperatorApplication, PidSort, Primary, PrimaryExtension,
  //   ProcedureCall, ProcedureCallBody, ProcedureProperty, ProcedureSignature, RemoteProcedureCall,
  //   RemoteProcedureCallBody, SignalProperty, StateEntryPoint, StateExitPoint,
  //   StateList, StateListToken, Synonym, Syntype, TimerProperty, TypeExpression,
  //   ValueReturningProcedureCall, VariableAccess, VariableProperty
  //
  //   Note 1: 'FormalParameter' and 'ProcedureSignature' cannot be used for root nodes as they
  //           conflict with 'SortList' (caused by ambiguities in 'behaviourProperty')
  //   Note 2: If one of the tokens is needed in some parts of the AST, a corresponding token
  //           with prefix 'Exc' is used ('Exc' denotes that the token is used exceptionally)

  tokens
  {
    Abstract;                     ActionStatement;              ActualContextParameters;
    ActualParameters;             Adding;                       AgentBody;
    AgentConstraint;              AgentFormalParameters;        AgentInstantiation;
    AgentStructure;               AgentTypeConstraint;          AggregationAggregateEndBoundKind;
    AggregationNotBoundKind;      AggregationPartEndBoundKind;  AggregationTwoEndsBoundKind;
    AlgorithmActionStatement;     AlgorithmAnswerPart;          AlgorithmElsePart;
    AnchoredSort;                 AnswerPart;                   Any;
    AnyExpression;                Argument;                     Arguments;
    Assignment;                   AssignmentStatement;          Association;
    AssociationEnd;               AssociationEndBoundKind;      AssociationNotBoundKind;
    AssociationTwoEndsBoundKind;  AsteriskConnectList;          AsteriskExceptionHandlerList;
    AsteriskExceptionStimulusList;                              AsteriskInputList;
    AsteriskSaveList;             AsteriskStateList;
    Atleast;                      Block;                        BlockContextParameter;
    BlockDefinition;              BlockHeading;                 BlockReference;
    BlockType;                    BlockTypeContextParameter;    BlockTypeDefinition;
    BlockTypeHeading;             BlockTypeReference;           BreakStatement;
    ChannelDefinition;            ChannelEndpoint;              ChannelIdentifiers;
    ChannelPath;                  ChannelToChannelConnection;   ChoiceDefinition;
    ChoiceOfSort;                 ClosedRange;                  CommConDestination;
    CommConTimer;                 Comment;                      CompositeStateBody;
    CompositeStateGraph;          CompositeStateHeading;        CompositeStateReference;
    CompositeStateTypeConstraint; CompositeStateTypeContextParameter;
    CompositeStateTypeDefinition; CompositeStateTypeHeading;    CompositeStateTypeReference;
    CompositionCompositeEndBoundKind;                           CompositionNotBoundKind;
    CompositionPartEndBoundKind;  CompositionTwoEndsBoundKind;  CompoundStatement;
    ConditionalExpression;        ConnectList;                  ConnectPart;
    Constraint;                   ContinuousSignal;             CreateRequest;
    DashNextstate;                DataTypeDefinition;           DataTypeDefinitionBody;
    DataTypeHeading;              DataTypeReference;            DataTypeSpecialization;
    Decision;                     DecisionBody;                 DecisionStatement;
    DecisionStatementBody;        Default;                      DefaultInitialization;
    DefinitionSelection;          DefinitionSelectionList;      ElsePart;
    EmptyStatement;               EnablingCondition;            End;
    EndpointConstraint;           Env;                          ExcCallStatement;
    Exception;
    ExceptionContextParameter;    ExceptionDefinition;          ExceptionDefinitionItem;
    ExceptionHandler;             ExceptionHandlerList;         ExceptionRaise;
    ExceptionStatement;           ExceptionStimulus;            ExceptionStimulusList;
    ExpandedSort;                 Export;                       Exported;
    ExportedAs;                   ExpressionList;               ExternalChannelIdentifiers;
    ExternalOperationDefinition;  ExternalProcedureDefinition;  ExternalSynonymDefinitionItem;
    Field;                        FieldDefaultInitialization;   FieldsOfSort;
    Finalized;                    FormalContextParameters;      FormalOperationParameters;
    FormalVariableParameters;     FreeAction;                   GateConstraint;
    GateContextParameter;         GateDefinition;               Handle;
    HandleStatement;              HistoryDashNextstate;         Identifier;
    IfStatement;                  ImportExpression;             In;
    InOut;                        InnerEntryPoint;              InputList;
    InputPart;                    Interface;                    InterfaceConstraint;
    InterfaceContextParameter;    InterfaceDefinition;          InterfaceGateDefinition;
    InterfaceHeading;             InterfaceProcedureDefinition; InterfaceReference;
    InterfaceSpecialization;      InterfaceUseList;             InterfaceVariableDefinition;
    InternalSynonymDefinitionItem;Join;                         Label;
    LabelledStatement;            LiteralList;                  Literals;
    Local;                        LocalVariablesOfSort;         LoopBreakStatement;
    LoopClause;                   LoopContinueStatement;        LoopStatement;
    LoopStep;                     LoopVariableDefinition;       LoopVariableIndication;
    Method;                       MethodList;                   Methods;
    NameClassLiteral;             NameClassOperation;           NamedNumber;
    Nextstate;                    Nodelay;                      NowExpression;
    NumberOfInstances;            Object;                       Offspring;
    OnException;                  OpAnd;                        OpBinaryMinus;
    OpConcat;                     OpDivide;                     OpEqual;
    OpGreater;                    OpGreaterOrEqual;             OpImplies;
    OpIn;                         OpLess;                       OpLessOrEqual;
    OpMod;                        OpNot;                        OpNotEqual;
    OpOr;                         OpPlus;                       OpRangeCheck;
    OpRem;                        OpStar;                       OpTimes;
    OpUnaryMinus;                 OpXor;                        OpenRange;
    OperationBody;                OperationDefinition;          OperationHeading;
    OperationParameters;          OperationPreamble;            OperationReference;
    OperationResult;              OperationSignature;           OperationSignatureInConstraint;
    Operator;                     OperatorList;                 Operators;
    Optional;                     Ordered;                      Out;
    OuterEntryPoint;              Output;                       Package;
    PackageDefinition;            PackageHeading;               PackageInterface;
    PackageReference;             PackageUseClause;             ParametersOfSort;
    Parent;                       PartialRegularExpression;     PathItem;
    PriorityInput;                PriorityInputList;
    Private;                      Procedure;                    ProcedureBody;
    ProcedureConstraint;          ProcedureContextParameter;    ProcedureDefinition;
    ProcedureFormalParameters;    ProcedureHeading;             ProcedurePreamble;
    ProcedureReference;           ProcedureResult;              ProcedureSignatureInConstraint;
    Process;                      ProcessContextParameter;      ProcessDefinition;
    ProcessHeading;               ProcessReference;             ProcessType;
    ProcessTypeContextParameter;  ProcessTypeDefinition;        ProcessTypeHeading;
    ProcessTypeReference;         Protected;                    Public;
    Qualifier;                    Raise;                        RaiseStatement;
    Raises;                       RangeCondition;               Redefined;
    ReferenceSort;                RegularInterval;              Remote;
    RemoteProcedure;              RemoteProcedureContextParameter;
    RemoteProcedureDefinition;    RemoteProcedureReject;        RemoteVariableContextParameter;
    RemoteVariableDefinition;     RenamePair;                   Renaming;
    Reset;                        ResetClause;                  Result;
    Return;                       ReturnStatement;              SDLSpecification;
    SavePart;                     SelectDefinition;             Self;
    Sender;                       Set;                          SetClause;
    Signal;                       SignalConstraint;             SignalContextParameter;
    SignalDefinition;             SignalDefinitionItem;         Signallist;
    SignalList;                   SignalListDefinition;         SignalListItem;
    SignalReference;              SizeConstraint;               SomeExpression;
    SomePrimaryExtension;         SomeProcedureCall;
    SomeSort;                     SortConstraint;               SortContextParameter;
    SortList;                     SortSignature;                Specialization;
    SpellingTerm;                 SpontaneousTransition;        Start;
    State;                        StateAggregation;             StateAggregationHeading;
    StateAggregationTypeDefinition;                             StateAggregationTypeHeading;
    StateConnectionPoints;        StateEntryPoints;             StateExitPoints;
    StateExpression;              StatePartitionConnection;     StateType;
    Stimulus;                     Stop;                         StructureDefinition;
    StructurePrimary;             Synonym;                      SynonymContextParameter;
    SynonymDefinition;            SyntypeDefinition;            System;
    SystemDefinition;             SystemHeading;                SystemType;
    SystemTypeDefinition;         SystemTypeHeading;            SystemTypeReference;
    Task;                         TerminatorStatement;          This;
    Timer;                        TimerActiveExpression;        TimerContextParameter;
    TimerDefaultInitialization;   TimerDefinition;              TimerDefinitionItem;
    Transition;                   TransitionOption;             Type;
    TypeReferenceProperties;      TypebasedBlockDefinition;     TypebasedCompositeState;
    TypebasedProcessDefinition;   TypebasedStatePartitionDefinition;
    TypebasedSystemDefinition;    ValidInputSignalSet;          Value;
    Variable;                     VariableContextParameter;     VariableDefinition;
    VariableDefinitionStatement;  VariableField;                VariableIndex;
    VariablesOfSort;              ViaGate;                      ViaPath;
    Virtual;                      VirtualityConstraint;         Void;
  }

{
  public:

    unsigned int numberOfErrors() const;
    void resetErrors();

    static const string &versionInfo();

  protected:

    void reportError( const string &errorMessage );
    void reportError( const antlr::RecognitionException& ex );
    void reportMessage( const string &message );

  private:

    unsigned int   noOfErrors;
}

/* 6.2.2 Macro definition, page 18 */

macroDefinition :

  MACRODEFINITION

  { throw new antlr::SemanticException( "Macros are not supported by this SDL-2000 parser.\n" ); }

  ;

/* 6.3 Visibility rules, names and identifiers, page 19 */

nameDef[ SDLName::Type type ] :

  t:Name

  { /*symTable.registerName( t -> getText(), type )*/ }

  ;

identifier :

  ( qualifier { ## = #( [Identifier], ## ); } )? Name

  ;

qualifier :

  start:QualifierBeginSign! pathItem ( "/"! pathItem )* QualifierEndSign!

  { ## = #( [Qualifier], ## ); SETLINECOLUMN( ##, start ); }

  ;

stringName :

    CharacterString
  | BitString
  | HexString

  ;

pathItem :

  ( scopeUnitKind )? Name

  { ## = #( [PathItem], ## ); }

  ;

scopeUnitKind! :

    start01:PACKAGE!       { ## = #[Package];     SETLINECOLUMN( ##, start01 ); }
  | start02:SYSTEM! TYPE!  { ## = #[SystemType];  SETLINECOLUMN( ##, start02 ); }
  | start03:SYSTEM!        { ## = #[System];      SETLINECOLUMN( ##, start03 ); }
  | start04:BLOCK!         { ## = #[Block];       SETLINECOLUMN( ##, start04 ); }
  | start05:BLOCK! TYPE!   { ## = #[BlockType];   SETLINECOLUMN( ##, start05 ); }
  | start06:PROCESS!       { ## = #[Process];     SETLINECOLUMN( ##, start06 ); }
  | start07:PROCESS! TYPE! { ## = #[ProcessType]; SETLINECOLUMN( ##, start07 ); }
  | start08:STATE!         { ## = #[State];       SETLINECOLUMN( ##, start08 ); }
  | start09:STATE! TYPE!   { ## = #[StateType];   SETLINECOLUMN( ##, start09 ); }
  | start10:PROCEDURE!     { ## = #[Procedure];   SETLINECOLUMN( ##, start10 ); }
  | start11:SIGNAL!        { ## = #[Signal];      SETLINECOLUMN( ##, start11 ); }
  | start12:TYPE!          { ## = #[Type];        SETLINECOLUMN( ##, start12 ); }
  | start13:OPERATOR!      { ## = #[Operator];    SETLINECOLUMN( ##, start13 ); }
  | start14:METHOD!        { ## = #[Method];      SETLINECOLUMN( ##, start14 ); }
  | start15:INTERFACE!     { ## = #[Interface];   SETLINECOLUMN( ##, start15 ); }

  ;

/* 6.4 Informal text, page 23 */

informalText :  //  no node type change due to ambiguity in various rules

  CharacterString

  ;

/* 6.7 Comment, page 24 */

end { bool cmt = false; } :

  ( comment { cmt = true; } )? start:Semicolon!

  { ## = #( [End], ## ); if ( !cmt ) SETLINECOLUMN( ##, start ); }

  ;

comment! :

  start:COMMENT! c:CharacterString

  { c -> setType( Comment ); ## = #c; SETLINECOLUMN( ##, start ); }

  ;

/* 7.1 Framework, page 25 */

sdlSpecification :

  (   ( ( packageUseClause )* PACKAGE ) =>
      package
    | systemSpecification )
  // ( package )*     package is also included in referencedDefinition
  ( referencedDefinition )*
  EOF!

  { ## = #( [SDLSpecification], ## ); }

  ;

systemSpecification :

  textualSystemSpecification

  ;

package :

  packageDefinition

  ;

textualSystemSpecification :

    (   SYSTEM typebasedSystemHeading | BLOCK typebasedBlockHeading
      | PROCESS typebasedProcessHeading ) =>
    textualTypebasedAgentDefinition
  | agentDefinition

  ;

/* 7.2 Package, page 26 */

packageDefinition :

  ( packageUseClause )*
  packageHeading end
  ( entityInPackage )*
  ENDPACKAGE! ( /*package*/ Name! )? end

  { ## = #( [PackageDefinition], ## ); }

  ;

packageHeading :

  start:PACKAGE! ( qualifier )? nameDef[ SDLName::package ]
  ( packageInterface )?

  { ## = #( [PackageHeading], ## ); SETLINECOLUMN( ##, start ); }

  ;

entityInPackage :

    ( agentTypeReference ) =>
    agentTypeReference
  | ( packageReference ) =>
    packageReference
  | ( signalReference ) =>
    signalReference
  | ( dataTypeReference ) =>
    dataTypeReference
  | ( procedureReference ) =>
    procedureReference
  | ( compositeStateTypeReference ) =>
    compositeStateTypeReference
  | ( interfaceReference ) =>
    interfaceReference
  | ( ( packageUseClause )* ( SYSTEM TYPE | typePreamble ( BLOCK TYPE | PROCESS TYPE ) ) ) =>
    agentTypeDefinition
  | ( ( packageUseClause )* PACKAGE ) =>
    packageDefinition
  | ( typePreamble SIGNAL ) =>
    signalDefinition
  | signalListDefinition
  | remoteVariableDefinition
  | ( typePreamble ( VALUE | OBJECT ) TYPE | ( virtuality )? INTERFACE | SYNTYPE | SYNONYM ) =>
    dataDefinition
  | ( PROCEDURE | ( packageUseClause )* procedureHeading ) =>
    procedureDefinition
  | remoteProcedureDefinition
  | compositeStateTypeDefinition
  | exceptionDefinition
  | selectDefinition
  | macroDefinition
  | association

  ;

packageReference :

  start:PACKAGE! ( qualifier )? /*package*/ Name REFERENCED! end

  { ## = #( [PackageReference], ## ); SETLINECOLUMN( ##, start ); }

  ;

packageUseClause :

  start:USE! /*package*/ identifier ( "/"! definitionSelectionList )? end

  { ## = #( [PackageUseClause], ## ); SETLINECOLUMN( ##, start ); }

  ;

definitionSelectionList :

  definitionSelection ( ","! definitionSelection )*

  { ## = #( [DefinitionSelectionList], ## ); }

  ;

definitionSelection :

  ( selectedEntityKind )? Name

  { ## = #( [DefinitionSelection], ## ); }

  ;

selectedEntityKind! :

    start01:SYSTEM! TYPE!      { ## = #[SystemType];      SETLINECOLUMN( ##, start01 ); }
  | start02:BLOCK! TYPE!       { ## = #[BlockType];       SETLINECOLUMN( ##, start02 ); }
  | start03:PROCESS! TYPE!     { ## = #[ProcessType];     SETLINECOLUMN( ##, start03 ); }
  | start04:PACKAGE!           { ## = #[Package];         SETLINECOLUMN( ##, start04 ); }
  | start05:SIGNAL!            { ## = #[Signal];          SETLINECOLUMN( ##, start05 ); }
  | start06:PROCEDURE!         { ## = #[Procedure];       SETLINECOLUMN( ##, start06 ); }
  | start07:REMOTE! PROCEDURE! { ## = #[RemoteProcedure]; SETLINECOLUMN( ##, start07 ); }
  | start08:TYPE!              { ## = #[Type];            SETLINECOLUMN( ##, start08 ); }
  | start09:SIGNALLIST!        { ## = #[Signallist];      SETLINECOLUMN( ##, start09 ); }
  | start10:STATE! TYPE!       { ## = #[StateType];       SETLINECOLUMN( ##, start10 ); }
  | start11:SYNONYM!           { ## = #[Synonym];         SETLINECOLUMN( ##, start11 ); }
  | start12:REMOTE!            { ## = #[Remote];          SETLINECOLUMN( ##, start12 ); }
  | start13:EXCEPTION!         { ## = #[Exception];       SETLINECOLUMN( ##, start13 ); }
  | start14:INTERFACE!         { ## = #[Interface];       SETLINECOLUMN( ##, start14 ); }

  ;

packageInterface :

  start:PUBLIC! definitionSelectionList

  { ## = #( [PackageInterface], ## ); SETLINECOLUMN( ##, start ); }

  ;

/* 7.3 Referenced definition, page 30 */

referencedDefinition :

  definition

  ;

definition :

    ( ( packageUseClause )* PACKAGE ) =>
    packageDefinition
  | ( ( packageUseClause )* ( SYSTEM TYPE | typePreamble ( BLOCK TYPE | PROCESS TYPE ) ) ) =>
    agentTypeDefinition
  | ( ( packageUseClause )* ( SYSTEM | BLOCK | PROCESS ) ) =>
    agentDefinition
  | ( ( packageUseClause )* ( virtuality )? STATE ( AGGREGATION )? TYPE ) =>
    compositeStateTypeDefinition
  | ( ( packageUseClause )* ( compositeStateHeading | stateAggregationHeading ) ) =>
    compositeState
  | ( PROCEDURE | ( packageUseClause )* procedureHeading ) =>
    procedureDefinition
  | operationDefinition
  | macroDefinition

  ;

/* 8.1.1.1 Agent types, page 36 */

agentTypeDefinition :

    ( ( packageUseClause )* SYSTEM ) =>
    systemTypeDefinition
  | ( ( packageUseClause )* typePreamble BLOCK ) =>
    blockTypeDefinition
  | processTypeDefinition

  ;

typePreamble returns [ bool preamble = false ] :

  (   virtuality { preamble = true; }
    | abstract   { preamble = true; } )?

  ;

agentTypeAdditionalHeading :   // no root node because 'agentTypeAdditionalHeading'
                               // may derive to the empty string
  ( formalContextParameters )? ( virtualityConstraint )?
  agentAdditionalHeading

  ;

agentTypeReference :

    systemTypeReference
  | ( typePreamble BLOCK ) =>
    blockTypeReference
  | processTypeReference

  ;

/* 8.1.1.2 System type, page 33 */

systemTypeDefinition :

  ( packageUseClause )*
  systemTypeHeading end agentStructure
  ENDSYSTEM! TYPE! ( ( qualifier! )? /*systemType*/ Name! )? end

  { ## = #( [SystemTypeDefinition], ## ); }

  ;

systemTypeHeading :

  start:SYSTEM! TYPE! ( qualifier )? nameDef[ SDLName::systemType ] agentTypeAdditionalHeading

  { ## = #( [SystemTypeHeading], ## ); SETLINECOLUMN( ##, start ); }

  ;

systemTypeReference :

  start:SYSTEM! TYPE! /*systemType*/ identifier typeReferenceProperties

  { ## = #( [SystemTypeReference], ## ); SETLINECOLUMN( ##, start ); }

  ;

/* 8.1.1.3 Block type, page 34 */

blockTypeDefinition :

  ( packageUseClause )*
  blockTypeHeading end agentStructure
  ENDBLOCK! TYPE! ( ( qualifier! )? /*blockType*/ Name! )? end

  { ## = #( [BlockTypeDefinition], ## ); }

  ;

blockTypeHeading { bool preamble; } :

  preamble = typePreamble
  start:BLOCK! TYPE! ( qualifier )? nameDef[ SDLName::blockType ]
  agentTypeAdditionalHeading

  { ## = #( [BlockTypeHeading], ## ); if ( !preamble ) SETLINECOLUMN( ##, start ); }

  ;

blockTypeReference { bool preamble; } :

  preamble = typePreamble
  start:BLOCK! TYPE! /*blockType*/ identifier typeReferenceProperties

  { ## = #( [BlockTypeReference], ## ); if ( !preamble ) SETLINECOLUMN( ##, start ); }

  ;

/* 8.1.1.4 Process type, page 34 */

processTypeDefinition :

  ( packageUseClause )*
  processTypeHeading end agentStructure
  ENDPROCESS! TYPE! ( ( qualifier! )? /*processType*/ Name! )? end

  { ## = #( [ProcessTypeDefinition], ## ); }

  ;

processTypeHeading { bool preamble; } :

  preamble = typePreamble
  start:PROCESS! TYPE! ( qualifier )? nameDef[ SDLName::processType ]
  agentTypeAdditionalHeading

  { ## = #( [ProcessTypeHeading], ## ); if ( !preamble ) SETLINECOLUMN( ##, start ); }

  ;

processTypeReference { bool preamble; } :

  preamble = typePreamble
  start:PROCESS! TYPE! /*processType*/ identifier typeReferenceProperties

  { ## = #( [ProcessTypeReference], ## ); if ( !preamble ) SETLINECOLUMN( ##, start ); }

  ;

/* 8.1.1.5 Composite state type, page 35 */

compositeStateTypeDefinition :

    ( ( packageUseClause )* ( virtuality )? STATE TYPE ) =>
    ( packageUseClause )*
    compositeStateTypeHeading end SUBSTRUCTURE!
    ( validInputSignalSet )?
    ( gateInDefinition )*
    ( stateConnectionPoints )*
    loopstar_entityInCompositeState_followedBy_compositeStateBodyOrEndSub
    compositeStateBody
    ENDSUBSTRUCTURE! STATE! TYPE! ( ( qualifier! )? /*compositeStateType*/ Name! )? end

    { ## = #( [CompositeStateTypeDefinition], ## ); }

  | ( packageUseClause )*
    stateAggregationTypeHeading end SUBSTRUCTURE!
    ( validInputSignalSet )?
    ( gateInDefinition )*
    ( stateConnectionPoints )*
    loopstar_entityInCompositeState_followedBy_stateAggregationBodyOrEndSub
    stateAggregationBody
    ENDSUBSTRUCTURE! STATE! TYPE! ( ( qualifier! )? /*compositeStateType*/ Name! )? end

    { ## = #( [StateAggregationTypeDefinition], ## ); }

  ;

loopstar_entityInCompositeState_followedBy_compositeStateBodyOrEndSub :

    ( ONEXCEPTION | START | state | EXCEPTIONHANDLER | CONNECTION | ENDSUBSTRUCTURE ) =>
    // end of loop
  | entityInCompositeState loopstar_entityInCompositeState_followedBy_compositeStateBodyOrEndSub

  ;

loopstar_entityInCompositeState_followedBy_stateAggregationBodyOrEndSub :

    (   compositeStateReference
      | ( textualTypebasedStatePartitionDefinition ) => textualTypebasedStatePartitionDefinition
      | ( packageUseClause )* ( compositeStateHeading | stateAggregationHeading )
      | statePartitionConnection
      | ENDSUBSTRUCTURE ) =>
    // end of loop
  | entityInCompositeState loopstar_entityInCompositeState_followedBy_stateAggregationBodyOrEndSub

  ;

compositeStateTypeHeading { bool virt = false; } :

  ( virtuality { virt = true; } )?
  start:STATE! TYPE! ( qualifier )? nameDef[ SDLName::compositeStateType ]
  ( formalContextParameters )? ( virtualityConstraint )?
  ( specialization )? ( agentFormalParameters )?

  { ## = #( [CompositeStateTypeHeading], ## ); if ( !virt ) SETLINECOLUMN( ##, start ); }

  ;

stateAggregationTypeHeading { bool virt = false; } :

  ( virtuality { virt = true; } )?
  start:STATE! AGGREGATION! TYPE! ( qualifier )? nameDef[ SDLName::compositeStateType ]
  ( formalContextParameters )? ( virtualityConstraint )?
  ( specialization )? ( agentFormalParameters )?

  { ## = #( [StateAggregationTypeHeading], ## ); if ( !virt ) SETLINECOLUMN( ##, start ); }

  ;

compositeStateTypeReference { bool preamble; } :

  preamble = typePreamble
  start:STATE! TYPE! /*compositeStateType*/ identifier typeReferenceProperties

  { ## = #( [CompositeStateTypeReference], ## ); if ( !preamble ) SETLINECOLUMN( ##, start ); }

  ;

/* 8.1.2 Type expression, page 36 */

typeExpression :   // no root node due to ambiguity in 'procedureCallBody'

  baseType ( actualContextParameters )?

  ;

baseType :

  /*unspecified*/ identifier

  ;

/* 8.1.3 Definitions based on types, page 37 */

textualTypebasedAgentDefinition :

    textualTypebasedSystemDefinition
  | textualTypebasedBlockDefinition
  | textualTypebasedProcessDefinition

  ;

/* 8.1.3.1 System definition based on system type, page 38 */

textualTypebasedSystemDefinition :   // keyword 'SYSTEM' taken from 'typebasedSystemHeading'

  start:SYSTEM! typebasedSystemHeading end

  { ## = #( [TypebasedSystemDefinition], ## ); SETLINECOLUMN( ##, start ); }

  ;

typebasedSystemHeading :   // no root nodes for 'typebasedXXXHeading'
                           // keyword 'SYSTEM' moved to 'textualTypebasedSystemDefinition'
  nameDef[ SDLName::system ] Colon! /*system*/ typeExpression

  ;

/* 8.1.3.2 Block definition based on block type, page 38 */

textualTypebasedBlockDefinition :

  start:BLOCK! typebasedBlockHeading end

  { ## = #( [TypebasedBlockDefinition], ## ); SETLINECOLUMN( ##, start ); }

  ;

typebasedBlockHeading :   // no root nodes for 'typebasedXXXHeading'

  nameDef[ SDLName::block ] ( numberOfInstances )? Colon! /*block*/ typeExpression

  ;

/* 8.1.3.3 Process definition based on process type, page 39 */

textualTypebasedProcessDefinition :

  start:PROCESS! typebasedProcessHeading end

  { ## = #( [TypebasedProcessDefinition], ## ); SETLINECOLUMN( ##, start ); }

  ;

typebasedProcessHeading :   // no root nodes for 'typebasedXXXHeading'

  nameDef[ SDLName::process ] ( numberOfInstances )? Colon! /*process*/ typeExpression

  ;

/* 8.1.3.4 Composite state definition based on composite state type, page 39 */

textualTypebasedStatePartitionDefinition :

  start:STATE! AGGREGATION! typebasedStatePartitionHeading end

  { ## = #( [TypebasedStatePartitionDefinition], ## ); SETLINECOLUMN( ##, start ); }

  ;

typebasedCompositeState :

  nameDef[ SDLName::state ] ( actualParameters )? Colon! /*compositeState*/ typeExpression

  { ## = #( [TypebasedCompositeState], ## ); }

  ;

/* 8.1.4 Abstract type, page 39 */

abstract! :

  start:ABSTRACT!

  { ## = #[Abstract]; SETLINECOLUMN( ##, start ); }

  ;

/* 8.1.5 Type references, page 40 */

typeReferenceProperties :

  ( ( WITH ( attributeProperty end )+ ) =>   // inherently nondeterministic
    WITH! ( attributeProperty end )+ )?
  ( WITH! ( behaviourProperty end )+ )?
  REFERENCED! end

  { ## = #( [TypeReferenceProperties], ## ); }

  ;

attributeProperty :

    ( variableProperty ) =>   // inherently nondeterministic
    variableProperty
  | ( fieldProperty ) =>      // inherently nondeterministic
    fieldProperty
  | signalParameterProperty
  | interfaceVariableProperty

  ;

variableProperty :   // no root node due to ambiguity in 'attributeProperty'

  ( local | exported )? /*variable*/ Name ( sort )?

  ;

local! :

  start:Hyphen!

  { ## = #[Local]; SETLINECOLUMN( ##, start ); }

  ;

exported! :

  start:PlusSign!

  { ## = #[Exported]; SETLINECOLUMN( ##, start ); }

  ;

fieldProperty :   // no root node due to ambiguity in 'attributeProperty'

  ( symbolicVisibility )? /*field*/ Name ( sort )?

  ;

symbolicVisibility :

    xprivate
  | xpublic
  | xprotected

  ;

xprivate! :   // private is a reserved keyword

    start1:Hyphen!    { ## = #[Private]; SETLINECOLUMN( ##, start1 ); }
  | start2:PRIVATE!   { ## = #[Private]; SETLINECOLUMN( ##, start2 ); }

  ;

xpublic! :   // public is a reserved keyword

    start1:PlusSign!   { ## = #[Public]; SETLINECOLUMN( ##, start1 ); }
  | start2:PUBLIC!     { ## = #[Public]; SETLINECOLUMN( ##, start2 ); }

  ;

xprotected! :   // protected is a reserved keyword

    start1:NumberSign!   { ## = #[Protected]; SETLINECOLUMN( ##, start1 ); }
  | start2:PROTECTED!    { ## = #[Protected]; SETLINECOLUMN( ##, start2 ); }

  ;

signalParameterProperty :

  sortList

  ;

interfaceVariableProperty :   // no root node due to ambiguity in 'attributeProperty'

  /*remoteVariable*/ Name ( sort )?

  ;

behaviourProperty :   // Attention: Reordering changes the list of tokens that are not allowed due
                      // to ambiguity! (right now: 'ProcedureSignature' and 'FormalParameter')

    ( ( OPERATOR )? operationProperty ) =>             // inherently nondeterministic
    ( ch1:OPERATOR! { ADDCHILD(#[Operator],ch1); } )? operationProperty
  | ( ( METHOD )? operationProperty ) =>               // inherently nondeterministic
    ( ch2:METHOD! { ADDCHILD(#[Method],ch2); } )? operationProperty
  | ( ( PROCEDURE )? procedureProperty ) =>            // inherently nondeterministic
    ( ch3:PROCEDURE! { ADDCHILD(#[Procedure],ch3); } )? procedureProperty
  | ( ( SIGNAL )? signalProperty ) =>                  // inherently nondeterministic
    ( ch4:SIGNAL! { ADDCHILD(#[Signal],ch4); } )? signalProperty
  | ( ( EXCEPTION )? exceptionProperty ) =>            // inherently nondeterministic
    ( ch5:EXCEPTION! { ADDCHILD(#[Exception],ch5); } )? exceptionProperty
  | ( ch6:TIMER! { ADDCHILD(#[Timer],ch6); } )? timerProperty
  | interfaceUseList

  ;

operationProperty :   // no root node due to ambiguity in 'behaviourProperty'

  ( symbolicVisibility )? /*operation*/ Name procedureSignature

  ;

procedureProperty :   // no root node due to ambiguity in 'behaviourProperty'

  ( local | exported )? /*procedure*/ Name procedureSignature

  ;

signalProperty :   // no root node due to ambiguity in 'behaviourProperty'

  /*signal*/ Name ( sortList )?

  ;

exceptionProperty :   // no root node due to ambiguity in 'behaviourProperty'

  /*exception*/ Name ( sortList )?

  ;

timerProperty :   // no root node due to ambiguity in 'behaviourProperty'

  /*timer*/ Name ( sortList )?

  ;

/* 8.1.6 Gate, page 44 */

gateInDefinition :

    textualGateDefinition
  | textualInterfaceGateDefinition

  ;

textualGateDefinition :

  start:GATE! gate[ true ] ( ch:ADDING! { ADDCHILD(#[Adding],ch); } )? gateConstraint ( gateConstraint )? end

  { ## = #( [GateDefinition], ## ); SETLINECOLUMN( ##, start ); }

  ;

gate[ bool defined ] :

    { defined == true }?
    nameDef[ SDLName::gate ]
  | /*gate*/ Name

  ;

gateConstraint :

  (   ch1:OUT! { ADDCHILD(#[Out],ch1); } ( TO! textualEndpointConstraint )?
    | ch2:IN!  { ADDCHILD(#[In],ch2); }  ( FROM! textualEndpointConstraint )? )
  ( WITH! signalList )?

  { ## = #( [GateConstraint], ## ); }

  ;

textualEndpointConstraint :

  ( ch:ATLEAST! { ADDCHILD(#[Atleast],ch); } )? /*unspecified*/ identifier

  { ## = #( [EndpointConstraint], ## ); }

  ;

textualInterfaceGateDefinition :

  start:GATE! (   ch1:IN!  { ADDCHILD(#[In],ch1); }
                | ch2:OUT! { ADDCHILD(#[Out],ch2); } ) WITH! /*interface*/ identifier end

  { ## = #( [InterfaceGateDefinition], ## ); SETLINECOLUMN( ##, start ); }

  ;

/* 8.2 Context parameters, page 46 */

formalContextParameters :

  start:LessThanSign! formalContextParameterList GreaterThanSign!

  { ## = #( [FormalContextParameters], ## ); SETLINECOLUMN( ##, start ); }

  ;

formalContextParameterList :   // root node added in 'formalContextParameters'

  formalContextParameter ( end formalContextParameter )*

  ;

actualContextParameters :

  start:LessThanSign! actualContextParameterList GreaterThanSign!

  { ## = #( [ActualContextParameters], ## ); SETLINECOLUMN( ##, start ); }

  ;

actualContextParameterList :   // root node added in 'actualContextParameters'

  ( actualContextParameter | { ADDVOIDCHILD(#[Void]); } )
  ( ","! ( actualContextParameter | { ADDVOIDCHILD(#[Void]); } ) )*

  ;

actualContextParameter :

  // /*unspecified*/ identifier |   // subsumed by 'primary'
    /*constant*/ primary

  ;

formalContextParameter :

    agentTypeContextParameter
  | agentContextParameter
  | procedureContextParameter
  | remoteProcedureContextParameter
  | signalContextParameter
  | variableContextParameter
  | remoteVariableContextParameter
  | timerContextParameter
  | synonymContextParameter
  | sortContextParameter
  | exceptionContextParameter
  | compositeStateTypeContextParameter
  | gateContextParameter
  | interfaceContextParameter

  ;

/* 8.2.1 Agent type context parameter, page 47 */

agentTypeContextParameter :

  (   start1:PROCESS! TYPE!   { MAKEROOT( #[ProcessTypeContextParameter], start1 ); }
    | start2:BLOCK! TYPE!     { MAKEROOT( #[BlockTypeContextParameter], start2 ); } )
  /*agentType*/ Name ( agentTypeConstraint )?

  ;

agentTypeConstraint :

    start:ATLEAST! /*agent*/ identifier

    { ## = #( [AgentTypeConstraint], ## ); SETLINECOLUMN( ##, start ); }

  | agentSignature

    { ## = #( [AgentTypeConstraint], ## ); }

  ;

/* 8.2.2 Agent context parameter, page 47 */

agentContextParameter :

  (   start1:PROCESS!   { MAKEROOT( #[ProcessContextParameter], start1 ); }
    | start2:BLOCK!     { MAKEROOT( #[BlockContextParameter], start2 ); } )
  /*agent*/ Name ( agentConstraint )?

  ;

agentConstraint :

    ( ch:ATLEAST! { ADDCHILD(#[Atleast],ch); } | Colon! ) /*agentType*/ identifier

    { ## = #( [AgentConstraint], ## ); }

  | agentSignature

    { ## = #( [AgentConstraint], ## ); }

  ;

agentSignature :

  sortList

  ;

/* 8.2.3 Procedure context parameter, page 48 */

procedureContextParameter :

  start:PROCEDURE! /*procedure*/ Name procedureConstraint

  { ## = #( [ProcedureContextParameter], ## ); SETLINECOLUMN( ##, start ); }

  ;

procedureConstraint :

    start:ATLEAST! /*procedure*/ identifier

    { ## = #( [ProcedureConstraint], ## ); SETLINECOLUMN( ##, start ); }

  | procedureSignatureInConstraint

    { ## = #( [ProcedureConstraint], ## ); }

  ;

procedureSignatureInConstraint { bool fp = false; }:

  ( start:"("! { fp = true; } formalParameter ( ","! formalParameter )* ")"! )? ( result )?

  { ## = #( [ProcedureSignatureInConstraint], ## ); if ( fp ) SETLINECOLUMN( ##, start ); }

  ;

/* 8.2.4 Remote procedure context parameter, page 48 */

remoteProcedureContextParameter :

  start:REMOTE! PROCEDURE! /*procedure*/ Name procedureSignatureInConstraint

  { ## = #( [RemoteProcedureContextParameter], ## ); SETLINECOLUMN( ##, start ); }

  ;

/* 8.2.5 Signal context parameter, page 48 */

signalContextParameter :

  start:SIGNAL! /*signal*/ Name ( signalConstraint )? ( ","! /*signal*/ Name ( signalConstraint )? )*

  { ## = #( [SignalContextParameter], ## ); SETLINECOLUMN( ##, start ); }

  ;

signalConstraint :

    start:ATLEAST! /*signal*/ identifier

    { ## = #( [SignalConstraint], ## ); SETLINECOLUMN( ##, start ); }

  | signalSignature

    { ## = #( [SignalConstraint], ## ); }

  ;

signalSignature :

  sortList

  ;

/* 8.2.6 Variable context parameter, page 48 */

variableContextParameter :

  start:DCL! /*variable*/ Name ( ","! /*variable*/ Name )* sort
  ( ","! /*variable*/ Name ( ","! /*variable*/ Name )* sort )*

  { ## = #( [VariableContextParameter], ## ); SETLINECOLUMN( ##, start ); }

  ;

/* 8.2.7 Remote variable context parameter, page 49 */

remoteVariableContextParameter :

  start:REMOTE! /*remoteVariable*/ Name ( ","! /*remoteVariable*/ Name )* sort
  ( ","! /*remoteVariable*/ Name ( ","! /*remoteVariable*/ Name )* sort )*

  { ## = #( [RemoteVariableContextParameter], ## ); SETLINECOLUMN( ##, start ); }

  ;

/* 8.2.8 Timer context parameter, page 49 */

timerContextParameter :

  start:TIMER! /*timer*/ Name ( timerConstraint )? ( ","! /*timer*/ Name ( timerConstraint )? )*

  { ## = #( [TimerContextParameter], ## ); SETLINECOLUMN( ##, start ); }

  ;

timerConstraint :

  sortList

  ;

/* 8.2.9 Synonym context parameter, page 49 */

synonymContextParameter :

  start:SYNONYM! /*synonym*/ Name synonymConstraint ( ","! /*synonym*/ Name synonymConstraint )*

  { ## = #( [SynonymContextParameter], ## ); SETLINECOLUMN( ##, start ); }

  ;

synonymConstraint :

  sort

  ;

/* 8.2.10 Sort context parameter, page 49 */

sortContextParameter :

  (   ch1:VALUE!  { ADDCHILD(#[Value],ch1); }
    | ch2:OBJECT! { ADDCHILD(#[Object],ch2); } )?
  TYPE! /*sort*/ Name ( sortConstraint )?

  { ## = #( [SortContextParameter], ## ); }

  ;

sortConstraint :

    start:ATLEAST! sort

    { ## = #( [SortConstraint], ## ); SETLINECOLUMN( ##, start ); }

  | sortSignature

    { ## = #( [SortConstraint], ## ); }

  ;

sortSignature :

    ch1:LITERALS! { ADDCHILD(#[Literals],ch1); } literalSignature ( ","! literalSignature )*
    ( ch2:OPERATORS! { ADDCHILD(#[Operators],ch2); }
      operationSignatureInConstraint ( ","! operationSignatureInConstraint )* )?
    ( ch3:METHODS! { ADDCHILD(#[Methods],ch3); }
      operationSignatureInConstraint ( ","! operationSignatureInConstraint )* )?

    { ## = #( [SortSignature], ## ); }

  | ch4:OPERATORS! { ADDCHILD(#[Operators],ch4); }
    operationSignatureInConstraint ( ","! operationSignatureInConstraint )*
    ( ch5:METHODS! { ADDCHILD(#[Methods],ch5); }
      operationSignatureInConstraint ( ","! operationSignatureInConstraint )* )?

    { ## = #( [SortSignature], ## ); }

  | ch6:METHODS! { ADDCHILD(#[Methods],ch6); }
    operationSignatureInConstraint ( ","! operationSignatureInConstraint )*

    { ## = #( [SortSignature], ## ); }

  ;

operationSignatureInConstraint :

    operationName ( "("! formalParameter ( ","! formalParameter )* ")"! )? ( result )?

    { ## = #( [OperationSignatureInConstraint], ## ); }

  | nameClassOperation ( result )?

    { ## = #( [OperationSignatureInConstraint], ## ); }

  ;

/* 8.2.11 Exception context parameter, page 50 */

exceptionContextParameter :

  start:EXCEPTION! /*exception*/ Name ( exceptionConstraint )?
  ( ","! /*exception*/ Name ( exceptionConstraint )? )*

  { ## = #( [ExceptionContextParameter], ## ); SETLINECOLUMN( ##, start ); }

  ;

exceptionConstraint :

  sortList

  ;

/* 8.2.12 Composite state type context parameter, page 50 */

compositeStateTypeContextParameter :

  start:STATE! TYPE! /*compositeStateType*/ Name ( compositeStateTypeConstraint )?

  { ## = #( [CompositeStateTypeContextParameter], ## ); SETLINECOLUMN( ##, start ); }

  ;

compositeStateTypeConstraint :

    start:ATLEAST! /*compositeStateType*/ identifier

    { ## = #( [CompositeStateTypeConstraint], ## ); SETLINECOLUMN( ##, start ); }

  | compositeStateTypeSignature

    { ## = #( [CompositeStateTypeConstraint], ## ); }

  ;

compositeStateTypeSignature :

  sortList

  ;

/* 8.2.13 Gate context parameter, page 50 */

gateContextParameter :

  start:GATE! gate[ false ] gateConstraint ( gateConstraint )?

  { ## = #( [GateContextParameter], ## ); SETLINECOLUMN( ##, start ); }

  ;

/* 8.2.14 Interface context parameter, page 50 */

interfaceContextParameter :

  start:INTERFACE! /*interface*/ Name ( interfaceConstraint )?
  ( ","! /*interface*/ Name ( interfaceConstraint )? )*

  { ## = #( [InterfaceContextParameter], ## ); SETLINECOLUMN( ##, start ); }

  ;

interfaceConstraint :

  start:ATLEAST! /*interface*/ identifier

  { ## = #( [InterfaceConstraint], ## ); SETLINECOLUMN( ##, start ); }

  ;

/* 8.3.1 Adding properties, page 51 */

specialization :

  start:INHERITS! typeExpression ( ch:ADDING! { ADDCHILD(#[Adding],ch); } )?

  { ## = #( [Specialization], ## ); SETLINECOLUMN( ##, start ); }

  ;

/* 8.3.2 Virtual type, page 52 */

virtuality! :

    start1:VIRTUAL!     { ## = #[Virtual];   SETLINECOLUMN( ##, start1 ); }
  | start2:REDEFINED!   { ## = #[Redefined]; SETLINECOLUMN( ##, start2 ); }
  | start3:FINALIZED!   { ## = #[Finalized]; SETLINECOLUMN( ##, start3 ); }

  ;

virtualityConstraint :

  start:ATLEAST! /*unspecified*/ identifier

  { ## = #( [VirtualityConstraint], ## ); SETLINECOLUMN( ##, start ); }

  ;

/* 8.4 Associations, page 54 */

association :

  start:ASSOCIATION! ( /*association*/ Name )? associationKind end

  { ## = #( [Association], ## ); SETLINECOLUMN( ##, start ); }

  ;

associationKind :

    ( associationNotBoundKind ) =>
    associationNotBoundKind
  | ( associationEndBoundKind ) =>
    associationEndBoundKind
  | ( associationTwoEndsBoundKind ) =>
    associationTwoEndsBoundKind
  | ( compositionNotBoundKind ) =>
    compositionNotBoundKind
  | ( compositionPartEndBoundKind ) =>
    compositionPartEndBoundKind
  | ( compositionCompositeEndBoundKind ) =>
    compositionCompositeEndBoundKind
  | ( compositionTwoEndsBoundKind ) =>
    compositionTwoEndsBoundKind
  | ( aggregationNotBoundKind ) =>
    aggregationNotBoundKind
  | ( aggregationPartEndBoundKind ) =>
    aggregationPartEndBoundKind
  | aggregationAggregateEndBoundKind
  | aggregationTwoEndsBoundKind

  ;

associationNotBoundKind :

  start:FROM! associationEnd FROM! associationEnd

  { ## = #( [AssociationNotBoundKind], ## ); SETLINECOLUMN( ##, start ); }

  ;

associationEndBoundKind :

  start:FROM! associationEnd TO! associationEnd

  { ## = #( [AssociationEndBoundKind], ## ); SETLINECOLUMN( ##, start ); }

  ;

associationTwoEndsBoundKind :

  start:TO! associationEnd TO! associationEnd

  { ## = #( [AssociationTwoEndsBoundKind], ## ); SETLINECOLUMN( ##, start ); }

  ;

compositionNotBoundKind :

  start:FROM! associationEnd COMPOSITION! associationEnd

  { ## = #( [CompositionNotBoundKind], ## ); SETLINECOLUMN( ##, start ); }

  ;

compositionPartEndBoundKind :

  start:TO! associationEnd COMPOSITION! associationEnd

  { ## = #( [CompositionPartEndBoundKind], ## ); SETLINECOLUMN( ##, start ); }

  ;

compositionCompositeEndBoundKind :

  start:FROM! associationEnd TO! COMPOSITION! associationEnd

  { ## = #( [CompositionCompositeEndBoundKind], ## ); SETLINECOLUMN( ##, start ); }

  ;

compositionTwoEndsBoundKind :

  start:TO! associationEnd TO! COMPOSITION! associationEnd

  { ## = #( [CompositionTwoEndsBoundKind], ## ); SETLINECOLUMN( ##, start ); }

  ;

aggregationNotBoundKind :

  start:FROM! associationEnd AGGREGATION! associationEnd

  { ## = #( [AggregationNotBoundKind], ## ); SETLINECOLUMN( ##, start ); }

  ;

aggregationPartEndBoundKind :

  start:TO! associationEnd AGGREGATION! associationEnd

  { ## = #( [AggregationPartEndBoundKind], ## ); SETLINECOLUMN( ##, start ); }

  ;

aggregationAggregateEndBoundKind :

  start:FROM! associationEnd TO! AGGREGATION! associationEnd

  { ## = #( [AggregationAggregateEndBoundKind], ## ); SETLINECOLUMN( ##, start ); }

  ;

aggregationTwoEndsBoundKind :

  start:TO! associationEnd TO! AGGREGATION! associationEnd

  { ## = #( [AggregationTwoEndsBoundKind], ## ); SETLINECOLUMN( ##, start ); }

  ;

associationEnd :

  ( visibility )? ( AS! /*role*/ Name )? linkedTypeIdentifier
  ( SIZE! multiplicity )? ( ch:ORDERED! { ADDCHILD(#[Ordered],ch); } )?

  { ## = #( [AssociationEnd], ## ); }

  ;

linkedTypeIdentifier :   // Problem: Ambiguity if a name is reused for entities of different kinds

  /*agentType,dataType,interface*/ identifier

  ;

multiplicity :

  rangeCondition

  ;

/* 9 Agents, page 57 */

agentDefinition :

    ( ( packageUseClause )* SYSTEM ) =>
    systemDefinition
  | ( ( packageUseClause )* BLOCK ) =>
    blockDefinition
  | processDefinition

  ;

agentStructure { bool aBody; } :

  ( validInputSignalSet )?
  aBody = loopstar_in_agentStructure
  (   (   ENDSYSTEM | ENDBLOCK | ENDPROCESS
        | STATE ( SUBSTRUCTURE | AGGREGATION )
        | ( packageUseClause )* ( compositeStateHeading | stateAggregationHeading )
          end SUBSTRUCTURE ) =>
      ( statePartitioning )?
    | agentBody { /* use value of 'aBody' for checks and AST construction */ } )

  { ## = #( [AgentStructure], ## ); }

  ;

loopstar_in_agentStructure returns [ bool aBody ] { bool loopVar; } :

    (   ENDSYSTEM | ENDBLOCK | ENDPROCESS
      | STATE ( SUBSTRUCTURE | AGGREGATION )
      | ( packageUseClause )* ( compositeStateHeading | stateAggregationHeading )
        end SUBSTRUCTURE
      | ONEXCEPTION | START ) =>
    { aBody = true; }
    // end of loop
  | (   ( agentReference ) =>
        agentReference                    { aBody = false; }
      | channelToChannelConnection        { aBody = false; }
      | channelDefinition                 { aBody = false; }
      | gateInDefinition                  { aBody = true; }
      | (   SYSTEM typebasedSystemHeading | BLOCK typebasedBlockHeading
          | PROCESS typebasedProcessHeading ) =>
        textualTypebasedAgentDefinition   { aBody = false; }
      | ( ( packageUseClause )* ( systemHeading | blockHeading | processHeading ) ) =>
        agentDefinition                   { aBody = false; }
      | entityInAgent                     { aBody = true; } )
    loopVar = loopstar_in_agentStructure

    { aBody = aBody && loopVar; }

  ;

agentInstantiation :

  ( options { greedy = true; } : numberOfInstances )?   // if 'agentFormalParameters' are present,
  agentAdditionalHeading                            // 'numberOfInstances' must be present as well

  { ## = #( [AgentInstantiation], ## ); }

  ;

agentAdditionalHeading :   // no root node because 'agentAdditionalHeading'
                           // may derive to the empty string
  ( specialization )? ( agentFormalParameters )?

  ;

entityInAgent :

    ( signalReference ) =>
    signalReference
  | ( dataTypeReference ) =>
    dataTypeReference
  | ( interfaceReference ) =>
    interfaceReference
  | ( procedureReference ) =>
    procedureReference
  | ( compositeStateTypeReference ) =>
    compositeStateTypeReference
  | ( agentTypeReference ) =>
    agentTypeReference
  | ( typePreamble SIGNAL ) =>
    signalDefinition
  | signalListDefinition
  | variableDefinition
  | remoteProcedureDefinition
  | remoteVariableDefinition
  | ( typePreamble ( VALUE | OBJECT ) TYPE | ( virtuality )? INTERFACE | SYNTYPE | SYNONYM ) =>
    dataDefinition
  | timerDefinition
  | macroDefinition
  | exceptionDefinition
  | ( PROCEDURE | ( packageUseClause )* procedureHeading ) =>
    procedureDefinition
  | ( ( packageUseClause )* ( virtuality )? STATE ( AGGREGATION )? TYPE ) =>
    compositeStateTypeDefinition
  | selectDefinition
  | agentTypeDefinition
  | association

  ;

agentReference :

    blockReference
  | processReference

  ;

validInputSignalSet :

  start:SIGNALSET! ( signalList )? end

  { ## = #( [ValidInputSignalSet], ## ); SETLINECOLUMN( ##, start ); }

  ;

agentBody :

  ( ( onException )? start )?
  ( state | exceptionHandler | freeAction )*

  { ## = #( [AgentBody], ## ); }

  ;

agentFormalParameters :

  start:"("! parametersOfSort ( ","! parametersOfSort )* ")"!

  { ## = #( [AgentFormalParameters], ## ); SETLINECOLUMN( ##, start ); }

  ;

parametersOfSort :

  /*variable*/ Name ( ","! /*variable*/ Name )* s:sort!

  { ## = #( [ParametersOfSort], #s, ## ); }

  ;

numberOfInstances :

  start:"("! ( initialNumber | { ADDVOIDCHILD(#[Void]); } )
             ( ","! ( maximumNumber | { ADDVOIDCHILD(#[Void]); } ) )? ")"!

  { ## = #( [NumberOfInstances], ## ); SETLINECOLUMN( ##, start ); }

  ;

initialNumber :

  /*natural*/ simpleExpression

  ;

maximumNumber :

  /*natural*/ simpleExpression

  ;

/* 9.1 System, page 63 */

systemDefinition :

  ( packageUseClause )*
  systemHeading end agentStructure
  ENDSYSTEM! ( /*system*/ Name! )? end

  { ## = #( [SystemDefinition], ## ); }

  ;

systemHeading :

  start:SYSTEM! nameDef[ SDLName::system ] agentAdditionalHeading

  { ## = #( [SystemHeading], ## ); SETLINECOLUMN( ##, start ); }

  ;

/* 9.2 Block, page 64 */

blockDefinition :

  ( packageUseClause )*
  blockHeading end agentStructure
  ENDBLOCK! ( ( qualifier! )? /*block*/ Name! )? end

  { ## = #( [BlockDefinition], ## ); }

  ;

blockHeading :

  start:BLOCK! ( qualifier )? nameDef[ SDLName::block ] agentInstantiation

  { ## = #( [BlockHeading], ## ); SETLINECOLUMN( ##, start ); }

  ;

blockReference :

  start:BLOCK! /*block*/ Name ( numberOfInstances )? REFERENCED! end

  { ## = #( [BlockReference], ## ); SETLINECOLUMN( ##, start ); }

  ;

/* 9.3 Process, page 65 */

processDefinition :

  ( packageUseClause )*
  processHeading end agentStructure
  ENDPROCESS! ( ( qualifier! )? /*process*/ Name! )? end

  { ## = #( [ProcessDefinition], ## ); }

  ;

processHeading :

  start:PROCESS! ( qualifier )? nameDef[ SDLName::process ] agentInstantiation

  { ## = #( [ProcessHeading], ## ); SETLINECOLUMN( ##, start ); }

  ;

processReference :

  start:PROCESS! /*process*/ Name ( numberOfInstances )? REFERENCED! end

  { ## = #( [ProcessReference], ## ); SETLINECOLUMN( ##, start ); }

  ;

/* 9.4 Procedure, page 66 */

procedureDefinition { bool entity = false; } :

    ( externalProcedureDefinition ) =>
    externalProcedureDefinition
  | ( packageUseClause )*
    procedureHeading
    (   end entity = loopstar_entityInProcedure_followedBy_misc
        (   procedureBody
            ENDPROCEDURE! ( ( qualifier! )? /*procedure*/ Name! )? end
          | ( virtuality )? LeftCurlyBracket! statementList RightCurlyBracket! { entity == true }? )
      | ( virtuality )? LeftCurlyBracket! statementList RightCurlyBracket! )

    { ## = #( [ProcedureDefinition], ## ); }

  ;

loopstar_entityInProcedure_followedBy_misc returns [ bool entity ] :

    (   ONEXCEPTION | START | state /*!*/ | EXCEPTIONHANDLER | CONNECTION
      | ENDPROCEDURE | ( virtuality )? LeftCurlyBracket ) =>
    // end of loop
    { entity = false; }
  | entityInProcedure entity = loopstar_entityInProcedure_followedBy_misc
    { entity = true; }

  ;

procedurePreamble returns [ bool preamble ] :

  preamble = typePreamble
  ( ch:EXPORTED! { ADDCHILD(#[Exported],ch); preamble = true; }
    ( AS! /*remoteProcedure*/ identifier )? )?

  { if ( preamble ) ## = #( [ProcedurePreamble], ## ); }

  ;

procedureHeading { bool preamble; } :

  preamble = procedurePreamble
  start:PROCEDURE! ( qualifier )? nameDef[ SDLName::procedure ]
  ( formalContextParameters )? ( virtualityConstraint )?
  ( specialization )?
  ( procedureFormalParameters )?
  ( procedureResult )? ( raises )?

  { ## = #( [ProcedureHeading], ## ); if ( !preamble ) SETLINECOLUMN( ##, start ); }

  ;

procedureFormalParameters :

  start:"("! formalVariableParameters ( ","! formalVariableParameters )* ")"!

  { ## = #( [ProcedureFormalParameters], ## ); SETLINECOLUMN( ##, start ); }

  ;

formalVariableParameters :

  parameterKind parametersOfSort

  { ## = #( [FormalVariableParameters], ## ); }

  ;

parameterKind! :

  (   start1:IN! "/"! OUT!   { ## = #[InOut]; SETLINECOLUMN( ##, start1 ); }
    | start2:IN!             { ## = #[In];    SETLINECOLUMN( ##, start2 ); }
    | start3:OUT!            { ## = #[Out];   SETLINECOLUMN( ##, start3 ); } )?

  ;

procedureResult :

  start:ResultSign!
  ( ( Name sort ) =>
    /*variable*/ Name )?
  sort

  { ## = #( [ProcedureResult], ## ); SETLINECOLUMN( ##, start ); }

  ;

raises :

  start:RAISE! /*exception*/ identifier ( ","! /*exception*/ identifier )*

  { ## = #( [Raises], ## ); SETLINECOLUMN( ##, start ); }

  ;

entityInProcedure :

    ( dataTypeReference ) =>
    dataTypeReference
  | ( procedureReference ) =>
    procedureReference
  | ( compositeStateTypeReference ) =>
    compositeStateTypeReference
  | variableDefinition
  | ( typePreamble ( VALUE | OBJECT ) TYPE | ( virtuality )? INTERFACE | SYNTYPE | SYNONYM ) =>
    dataDefinition
  | ( PROCEDURE | ( packageUseClause )* procedureHeading ) =>
    procedureDefinition
  | compositeStateTypeDefinition
  | exceptionDefinition
  | selectDefinition
  | macroDefinition

  ;

procedureSignature :   // no root node due to ambiguity in 'behaviourProperty'

  ( "("! formalParameter ( ","! formalParameter )* ")"! )? ( result )? ( raises )?

  ;

procedureBody :

  ( onException )? ( start )? ( state | exceptionHandler | freeAction )*

  { ## = #( [ProcedureBody], ## ); }

  ;

externalProcedureDefinition :

  start:PROCEDURE! /*procedure*/ Name procedureSignature EXTERNAL! end

  { ## = #( [ExternalProcedureDefinition], ## ); SETLINECOLUMN( ##, start ); }

  ;

procedureReference { bool preamble; } :

  preamble = typePreamble
  start:PROCEDURE! /*procedure*/ identifier typeReferenceProperties

  { ## = #( [ProcedureReference], ## ); if ( !preamble ) SETLINECOLUMN( ##, start ); }

  ;

/* 10.1 Channel, page 70 */

channelDefinition :

  start:CHANNEL! ( nameDef[ SDLName::channel ] )? ( ch:NODELAY! { ADDCHILD(#[Nodelay],ch); } )?
  channelPath ( channelPath )?
  ENDCHANNEL! ( /*channel*/ Name! )? end

  { ## = #( [ChannelDefinition], ## ); SETLINECOLUMN( ##, start ); }

  ;

channelPath :

  start:FROM! channelEndpoint TO! channelEndpoint ( WITH! signalList )? end

  { ## = #( [ChannelPath], ## ); SETLINECOLUMN( ##, start ); }

  ;

channelEndpoint :   // Problem: Ambiguity if a name is reused for entities of different kinds

  (   /*agent,state*/ identifier
    | ch1:ENV!    { ADDCHILD(#[Env],ch1); }
    | ch2:THIS!   { ADDCHILD(#[This],ch2); } )
  ( viaGate )?

  { ## = #( [ChannelEndpoint], ## ); }

  ;

viaGate :

  start:VIA! gate[ false ]

  { ## = #( [ViaGate], ## ); SETLINECOLUMN( ##, start ); }

  ;

/* 10.2 Connection, page 72 */

channelToChannelConnection :

  start:CONNECT! externalChannelIdentifiers AND! channelIdentifiers end

  { ## = #( [ChannelToChannelConnection], ## ); SETLINECOLUMN( ##, start ); }

  ;

externalChannelIdentifiers :

  /*channel*/ identifier ( ","! /*channel*/ identifier )*

  { ## = #( [ExternalChannelIdentifiers], ## ); }

  ;

channelIdentifiers :

  /*channel*/ identifier ( ","! /*channel*/ identifier )*

  { ## = #( [ChannelIdentifiers], ## ); }

  ;

/* 10.3 Signal, page 73 */

signalDefinition { bool preamble; } :

  preamble = typePreamble
  start:SIGNAL! signalDefinitionItem ( ","! signalDefinitionItem )* end

  { ## = #( [SignalDefinition], ## ); if ( !preamble ) SETLINECOLUMN( ##, start ); }

  ;

signalDefinitionItem :

  nameDef[ SDLName::signal ]
  ( formalContextParameters )? ( virtualityConstraint )? ( specialization )? ( sortList )?

  { ## = #( [SignalDefinitionItem], ## ); }

  ;

sortList :

  start:"("! sort ( ","! sort )* ")"!

  { ## = #( [SortList], ## ); SETLINECOLUMN( ##, start ); }

  ;

signalReference { bool preamble; } :

  preamble = typePreamble
  start:SIGNAL! /*signal*/ identifier typeReferenceProperties

  { ## = #( [SignalReference], ## ); if ( !preamble ) SETLINECOLUMN( ##, start ); }

  ;

/* 10.4 Signal list definition, page 74 */

signalListDefinition :

  start:SIGNALLIST! nameDef[ SDLName::signalList ] EqualsSign! signalList end

  { ## = #( [SignalListDefinition], ## ); SETLINECOLUMN( ##, start ); }

  ;

signalList :

  signalListItem ( ","! signalListItem )*

  { ## = #( [SignalList], ## ); }

  ;

signalListItem :

    ( identifier ) =>                                      // inherently nondeterministic
    /*signal,timer*/ identifier

    { ## = #( [SignalListItem], ## ); }

  | start:"("! /*signalList*/ identifier ")"!

    { ## = #( [SignalListItem], ## ); SETLINECOLUMN( ##, start ); }

  | ( ( PROCEDURE )? identifier ) =>                       // inherently nondeterministic
    ( ch1:PROCEDURE! { ADDCHILD(#[Procedure],ch1); } )? /*remoteProcedure*/ identifier

    { ## = #( [SignalListItem], ## ); }

  | ( ( INTERFACE )? identifier ) =>                       // inherently nondeterministic
    ( ch2:INTERFACE! { ADDCHILD(#[Interface],ch2); } )? /*interface*/ identifier

    { ## = #( [SignalListItem], ## ); }

  | ( ch3:REMOTE! { ADDCHILD(#[Remote],ch3); } )? /*remoteVariable*/ identifier

    { ## = #( [SignalListItem], ## ); }

  ;

/* 10.5 Remote procedures, page 74 */

remoteProcedureDefinition :

  start:REMOTE! PROCEDURE! nameDef[ SDLName::remoteProcedure ] ( ch:NODELAY! { ADDCHILD(#[Nodelay],ch); } )?
  procedureSignature end

  { ## = #( [RemoteProcedureDefinition], ## ); SETLINECOLUMN( ##, start ); }

  ;

remoteProcedureCall :   // no dedicated root node due to ambiguity in 'action1'

  start:CALL! remoteProcedureCallBody

  { ## = #( [SomeProcedureCall], ## ); SETLINECOLUMN( ##, start ); }

  ;

remoteProcedureCallBody :   // no dedicated root node due to ambiguity in 'expression'

  /*remoteProcedure*/ identifier ( actualParameters )? communicationConstraints

  { ## = #( [SomeExpression], ## ); }

  ;

communicationConstraints :

  ( options { greedy = true; } :
      ch1:TO! d:destination!              { ADDCHILD(#([CommConDestination],#d),ch1); }
    | ch2:TIMER! /*timer*/ t:identifier!  { ADDCHILD(#([CommConTimer],#t),ch2); }
    | viaPath )*

  ;

/* 10.6 Remote variables, page 77 */

remoteVariableDefinition :

  start:REMOTE! n1:( nameDef[ SDLName::remoteVariable ] ( ","! nameDef[ SDLName::remoteVariable ] )* )!
                s1:sort! { ## = #( #s1, #n1 ); } ( ch1:NODELAY! { ADDCHILD(#[Nodelay],ch1); } )?
                ( ","!
                  n2:( nameDef[ SDLName::remoteVariable ] ( ","! nameDef[ SDLName::remoteVariable ] )* )!
                  s2:sort! { ## = #( ##, #s2, #n2 ); } ( ch2:NODELAY! { ADDCHILD(#[Nodelay],ch2); } )? )*
        end

  { ## = #( [RemoteVariableDefinition], ## ); SETLINECOLUMN( ##, start ); }

  ;

importExpression :

  start:IMPORT! "("! /*remoteVariable*/ identifier communicationConstraints ")"!

  { ## = #( [ImportExpression], ## ); SETLINECOLUMN( ##, start ); }

  ;

xexport :   // export is a reserved keyword

  start:EXPORT! "("! /*variable*/ identifier ( ","! /*variable*/ identifier )* ")"!

  { ## = #( [Export], ## ); SETLINECOLUMN( ##, start ); }

  ;

/* 11.1 Start, page 79 */

start :

  st:START! ( virtuality )? ( /*stateEntryPoint*/ Name )? end ( onException )? transition

  { ## = #( [Start], ## ); SETLINECOLUMN( ##, st ); }

  ;

/* 11.2 State, page 80 */

state :

  // basicState |             // subsumed by 'compositeStateApplication'
    compositeStateApplication

  ;

/* 11.2.1 Basic State, page 80 */

basicState :   // no 'basicState' node due to ambiguity in 'state'

  start:STATE! stateList end ( onException )?
  (   ( INPUT ( virtuality )? inputList ) =>
      inputPart
    | priorityInput
    | savePart
    | spontaneousTransition
    | continuousSignal )*
  ( ENDSTATE! ( /*state*/ Name! )? end )?

  { ## = #( [State], ## ); SETLINECOLUMN( ##, start ); }

  ;

stateList :   // no root node due to ambiguity in 'state' for 'basicState'

    /*state*/ Name ( ","! /*state*/ Name )*
  | asteriskStateList

  ;

asteriskStateList :

  start:Asterisk! ( "("! /*state*/ Name ( ","! /*state*/ Name )* ")"! )?

  { ## = #( [AsteriskStateList], ## ); SETLINECOLUMN( ##, start ); }

  ;

/* 11.2.2 Composite state application, page 82 */

compositeStateApplication :   // no 'compositeStateApplication' node due to ambiguity in 'state'

  start:STATE! compositeStateList end ( onException )?
  (   ( INPUT ( virtuality )? inputList ) =>
      inputPart
    | priorityInput
    | savePart
    | spontaneousTransition
    | continuousSignal
    | connectPart )*
  ( ENDSTATE! ( /*state*/ Name! )? end )?

  { ## = #( [State], ## ); SETLINECOLUMN( ##, start ); }

  ;

compositeStateList :   // no root node due to ambiguity in 'state' for 'compositeStateApplication'

    stateListToken ( ","! stateListToken )*
  | asteriskStateList

  ;

stateListToken :   // no root node due to ambiguity in 'state' for 'compositeStateApplication'

    ( Name ( actualParameters )? Colon ) =>   // these two rules start identically
    typebasedCompositeState
  | /*state*/ Name ( actualParameters )?

  ;

/* 11.3 Input, page 83 */

inputPart :

  start:INPUT! ( virtuality )? inputList end
  ( onException )? ( enablingCondition )? transition

  { ## = #( [InputPart], ## ); SETLINECOLUMN( ##, start ); }

  ;

inputList :

    stimulus ( ","! stimulus )*

    { ## = #( [InputList], ## ); }

  | asteriskInputList

  ;

stimulus :

  signalListItem
  (   "("! ( variable | { ADDVOIDCHILD(#[Void]); } )
           ( ","! ( variable | { ADDVOIDCHILD(#[Void]); } ) )* ")"!
    | remoteProcedureReject )?

  { ## = #( [Stimulus], ## ); }

  ;

remoteProcedureReject :

  start:RAISE! exceptionRaise

  { ## = #( [RemoteProcedureReject], ## ); SETLINECOLUMN( ##, start ); }

  ;

asteriskInputList! :

  start:Asterisk!

  { ## = #[AsteriskInputList]; SETLINECOLUMN( ##, start ); }

  ;

/* 11.4 Priority Input, page 84 */

priorityInput :

  start:PRIORITY! INPUT! ( virtuality )?
  priorityInputList end ( onException )? transition

  { ## = #( [PriorityInput], ## ); SETLINECOLUMN( ##, start ); }

  ;

priorityInputList :

  stimulus ( ","! stimulus )*

  { ## = #( [PriorityInputList], ## ); }

  ;

/* 11.5 Continuous signal, page 85 */

continuousSignal :

  start:PROVIDED! ( virtuality )? continuousExpression end
  ( PRIORITY! priorityName end )? ( onException )? transition

  { ## = #( [ContinuousSignal], ## ); SETLINECOLUMN( ##, start ); }

  ;

continuousExpression :

  /*boolean*/ expression

  ;

priorityName :

  /*naturalLiteral*/ Name

  ;

/* 11.6 Enabling condition, page 86 */

enablingCondition :

  start:PROVIDED! providedExpression end

  { ## = #( [EnablingCondition], ## ); SETLINECOLUMN( ##, start ); }

  ;

providedExpression :

  /*boolean*/ expression

  ;

/* 11.7 Save, page 86 */

savePart :

  start:SAVE! ( virtuality )? saveList end

  { ## = #( [SavePart], ## ); SETLINECOLUMN( ##, start ); }

  ;

saveList :

    signalList
  | asteriskSaveList

  ;

asteriskSaveList! :

  start:Asterisk!

  { ## = #[AsteriskSaveList]; SETLINECOLUMN( ##, start ); }

  ;

/* 11.9 Spontaneous transition, page 87 */

spontaneousTransition :

  start:INPUT! ( virtuality )? spontaneousDesignator! end
  ( onException )? ( enablingCondition )? transition

  { ## = #( [SpontaneousTransition], ## ); SETLINECOLUMN( ##, start ); }

  ;

spontaneousDesignator! :

  NONE!   // you can ignore it for AST construction

  ;

/* 11.10 Label, page 88 */

label :

  nameDef[ SDLName::connector ] ":"!

  { ## = #( [Label], ## ); }

  ;

freeAction :

  start:CONNECTION! transition ( ENDCONNECTION! ( /*connector*/ Name! )? end )?

  { ## = #( [FreeAction], ## ); SETLINECOLUMN( ##, start ); }

  ;

/* 11.11 State machine and Composite state, page 89 */

compositeState :

    ( ( packageUseClause )* compositeStateHeading ) =>
    compositeStateGraph
  | stateAggregation

  ;

/* 11.11.1 Composite State Graph, page 90 */

compositeStateGraph :

  ( packageUseClause )*
  compositeStateHeading end SUBSTRUCTURE!
  ( validInputSignalSet )?
  ( gateInDefinition )*
  ( stateConnectionPoints )*
  loopstar_entityInCompositeState_followedBy_compositeStateBodyOrEndSub
  compositeStateBody
  ENDSUBSTRUCTURE! ( ( qualifier! )? /*compositeState*/ Name! )? end

  { ## = #( [CompositeStateGraph], ## ); }

  ;

compositeStateHeading :

  start:STATE! ( qualifier )? nameDef[ SDLName::compositeState ] ( agentFormalParameters )?

  { ## = #( [CompositeStateHeading], ## ); SETLINECOLUMN( ##, start ); }

  ;

entityInCompositeState :

    ( dataTypeReference ) =>
    dataTypeReference
  | ( compositeStateTypeReference ) =>
    compositeStateTypeReference
  | variableDefinition
  | ( typePreamble ( VALUE | OBJECT ) TYPE | ( virtuality )? INTERFACE | SYNTYPE | SYNONYM ) =>
    dataDefinition
  | selectDefinition
  | macroDefinition
  | ( PROCEDURE | ( packageUseClause )* procedureHeading ) =>
    procedureDefinition
  | exceptionDefinition
  | compositeStateTypeDefinition

  ;

compositeStateBody :

  ( onException )? ( start )* ( state | exceptionHandler | freeAction )*

  { ## = #( [CompositeStateBody], ## ); }

  ;

compositeStateReference :

  start:STATE! SUBSTRUCTURE! /*compositeState*/ Name REFERENCED! end

  { ## = #( [CompositeStateReference], ## ); SETLINECOLUMN( ##, start ); }

  ;

/* 11.11.2 State aggregation, page 91 */

stateAggregation :

  ( packageUseClause )*
  stateAggregationHeading end SUBSTRUCTURE!
  ( stateConnectionPoints )*
  loopstar_entityInCompositeState_followedBy_stateAggregationBodyOrEndSub
  stateAggregationBody
  ENDSUBSTRUCTURE! ( ( qualifier! )? /*compositeState*/ Name! )? end

  { ## = #( [StateAggregation], ## ); }

  ;

stateAggregationHeading :

  start:STATE! AGGREGATION! ( qualifier )? nameDef[ SDLName::compositeState ] ( agentFormalParameters )?

  { ## = #( [StateAggregationHeading], ## ); SETLINECOLUMN( ##, start ); }

  ;

stateAggregationBody :

  (   statePartitioning
    | statePartitionConnection )+

  ;

statePartitioning :

    compositeStateReference
  | ( textualTypebasedStatePartitionDefinition ) =>
    textualTypebasedStatePartitionDefinition
  | compositeState

  ;

statePartitionConnection :

  start:CONNECT! outerEntryPoint AND! innerEntryPoint end

  { ## = #( [StatePartitionConnection], ## ); SETLINECOLUMN( ##, start ); }

  ;

outerEntryPoint :

  /*statePartition*/ identifier VIA! point

  { ## = #( [OuterEntryPoint], ## ); }

  ;

innerEntryPoint :

  /*statePartition*/ identifier VIA! point

  { ## = #( [InnerEntryPoint], ## ); }

  ;

point :

    stateEntryPoint
  // | stateExitPoint   // subsumed by 'stateEntryPoint'
  | start:DEFAULT!   { ## = #[Default]; SETLINECOLUMN( ##, start ); }

  ;

typebasedStatePartitionHeading :   // no root nodes for 'typebasedXXXHeading'

  nameDef[ SDLName::state ] Colon! /*compositeState*/ typeExpression

  ;

/* 11.11.3 State connection point, page 94 */

stateConnectionPoints :

  (   start1:IN!  stateEntryPoints
      { ## = #( [StateConnectionPoints], ## ); SETLINECOLUMN( ##, start1 ); }
    | start2:OUT! stateExitPoints
			{ ## = #( [StateConnectionPoints], ## ); SETLINECOLUMN( ##, start2 ); } )
  end

  ;

stateEntryPoints :

    stateEntryPoint

    { ## = #( [StateEntryPoints], ## ); }

  | start:"("! stateEntryPoint ( ","! stateEntryPoint )* ")"!

    { ## = #( [StateEntryPoints], ## ); SETLINECOLUMN( ##, start ); }

  ;

stateExitPoints :

    stateExitPoint

    { ## = #( [StateExitPoints], ## ); }

  | start:"("! stateExitPoint ( ","! stateExitPoint )* ")"!

    { ## = #( [StateExitPoints], ## ); SETLINECOLUMN( ##, start ); }

  ;

stateEntryPoint :

  /*stateEntryPoint*/ Name

  ;

stateExitPoint :

  /*stateExitPoint*/ Name

  ;

/* 11.11.4 Connect, page 95 */

connectPart :

  start:CONNECT! ( virtuality )? ( connectList )? end ( onException )? exitTransition

  { ## = #( [ConnectPart], ## ); SETLINECOLUMN( ##, start ); }

  ;

connectList :

    /*stateExitPoint*/ Name ( ","! /*stateExitPoint*/ Name )*

    { ## = #( [ConnectList], ## ); }

  | asteriskConnectList

  ;

exitTransition :

  transition

  ;

asteriskConnectList! :

  start:Asterisk!

  { ## = #[AsteriskConnectList]; SETLINECOLUMN( ##, start ); }

  ;

/* 11.12.1 Transition body, page 95 */

transition :

    ( ( label )?
      ( TASK | OUTPUT | CREATE | DECISION | SET | RESET | EXPORT | CALL | ALTERNATIVE ) ) =>
    transitionString ( terminatorStatement )?

    { ## = #( [Transition], ## ); }

  | terminatorStatement

    { ## = #( [Transition], ## ); }
  ;

transitionString :

  actionStatement loopstar_actionStatement

  ;

loopstar_actionStatement :

    ( ( label )?
      ( TASK | OUTPUT | CREATE | DECISION | SET | RESET | EXPORT | CALL | ALTERNATIVE ) ) =>
    actionStatement loopstar_actionStatement
  | // end of loop

  ;

actionStatement :

  ( label )?
  (   action1 end ( onException )?
    | action2 end )

  { ## = #( [ActionStatement], ## ); }

  ;

action1 :

    task
  | output
  | createRequest
  | decision
  | set
  | reset
  | xexport
  | ( CALL (   THIS
             | identifier (   ( actualContextParameters ) => actualContextParameters
                            | ( actualParameters )? ~( TO | TIMER | VIA ) ) ) ) =>
    procedureCall
  | remoteProcedureCall

  ;

action2 :

  transitionOption

  ;

terminatorStatement :

  ( label )?
  (   terminator1 end ( onException )?
    | terminator2 end )

  { ## = #( [TerminatorStatement], ## ); }

  ;

terminator1 :

    xreturn
  | raise

  ;

terminator2 :

    nextstate
  | join
  | stop

  ;

/* 11.12.2.1 Nextstate, page 98 */

nextstate :      // root node added in 'nextstateBody'

  start:NEXTSTATE! nsb:nextstateBody

  { SETLINECOLUMN( #nsb, start ); }

  ;

nextstateBody :

    /*state*/ Name ( actualParameters )? ( VIA! /*stateEntryPoint*/ Name )?

    { ## = #( [Nextstate], ## ); }

  | dashNextstate

  ;

dashNextstate :

    start:Hyphen!

    { ## = #[DashNextstate]; SETLINECOLUMN( ##, start ); }

  | historyDashNextstate

  ;

historyDashNextstate! :

  start:HistoryDashSign!

  { ## = #[HistoryDashNextstate]; SETLINECOLUMN( ##, start ); }

  ;

/* 11.12.2.2 Join, page 99 */

join :

  start:JOIN! /*connector*/ Name

  { ## = #( [Join], ## ); SETLINECOLUMN( ##, start ); }

  ;

/* 11.12.2.3 Stop, page 99 */

stop! :

  start:STOP!

  { ## = #[Stop]; SETLINECOLUMN( ##, start ); }

  ;

/* 11.12.2.4 Return, page 100 */

xreturn! :   // return is a reserved keyword

  start:RETURN!
  (   expression
    | VIA! stateExitPoint )?

  { ## = #( [Return], ## ); SETLINECOLUMN( ##, start ); }

  ;

/* 11.12.2.5 Raise, page 100 */

raise :

  start:RAISE! raiseBody

  { ## = #( [Raise], ## ); SETLINECOLUMN( ##, start ); }

  ;

raiseBody :

  exceptionRaise

  ;

exceptionRaise :

  /*exception*/ identifier ( actualParameters )?

  { ## = #( [ExceptionRaise], ## ); }

  ;

/* 11.13.1 Task, page 101 */

task :

  start:TASK! textualTaskBody

  { ## = #( [Task], ## ); SETLINECOLUMN( ##, start ); }

  ;

textualTaskBody :

    assignment
  | informalText
  | compoundStatement

  ;

/* 11.13.2 Create, page 102 */

createRequest :

  start:CREATE! createBody

  { ## = #( [CreateRequest], ## ); SETLINECOLUMN( ##, start ); }

  ;

createBody :   // root node added in 'createRequest'

  (   /*agent,agentType*/ identifier     // Problem: Ambiguity if a name is reused for entities
    | ch:THIS!   { ADDCHILD(#[This],ch); } )   //          of different kinds
  ( actualParameters )?

  ;

actualParameters :   // root node added in 'actualParameterList'

  start:"("! apl:actualParameterList ")"!

  { SETLINECOLUMN( #apl, start ); }

  ;

actualParameterList :

  ( expression | { ADDVOIDCHILD(#[Void]); } )
  ( ","! ( expression | { ADDVOIDCHILD(#[Void]); } ) )*

  { ## = #( [ActualParameters], ## ); }   // 'ActualParameters' used for consistency with other
                                          // parameter lists, e.g 'ActualContextParameters'
  ;

/* 11.13.3 Procedure call, page 103 */

procedureCall :   // no dedicated root node due to ambiguity in 'action1'

  start:CALL! procedureCallBody

  { ## = #( [SomeProcedureCall], ## ); SETLINECOLUMN( ##, start ); }

  ;

procedureCallBody :   // no dedicated root node due to ambiguity in 'expression'

  ( ch:THIS! { ADDCHILD(#[This],ch); } )?
  ( // /*procedure*/ identifier |   // subsumed by 'typeExpression'
      /*procedure*/ typeExpression )
  ( actualParameters )?

  { ## = #( [SomeExpression], ## ); }

  ;

/* 11.13.4 Output, page 104 */

output :

  start:OUTPUT! outputBody

  { ## = #( [Output], ## ); SETLINECOLUMN( ##, start ); }

  ;

outputBody :   // root node added in 'output'

  /*signal*/ identifier ( actualParameters )?
  ( ","! /*signal*/ identifier ( actualParameters )? )*
  communicationConstraints

  ;

destination :

    /*pid*/ expression
  // | /*agent*/ identifier   // subsumed by 'expression'
  // | THIS                   // subsumed by 'expression'

  ;

viaPath :

  start:VIA! /*channel,gate */ identifier

  { ## = #( [ViaPath], ## ); SETLINECOLUMN( ##, start ); }

  ;

/* 11.13.5 Decision, page 106 */

decision :

  start:DECISION! question end ( onException )? decisionBody ENDDECISION!

  { ## = #( [Decision], ## ); SETLINECOLUMN( ##, start ); }

  ;

decisionBody :

  ( answerPart )+ ( elsePart )?

  { ## = #( [DecisionBody], ## ); }

  ;

answerPart :

  start:"("! ( answer )? ")"! Colon! ( transition )?

  { ## = #( [AnswerPart], ## ); SETLINECOLUMN( ##, start ); }

  ;

answer :

    rangeCondition
  // | informalText   // subsumed by 'rangeCondition'

  ;

elsePart :

  start:ELSE! Colon! ( transition )?

  { ## = #( [ElsePart], ## ); SETLINECOLUMN( ##, start ); }

  ;

question :

     ( ANY end ) =>
     ch:ANY!   { ADDCHILD(#[Any],ch); }
  // | informalText   // subsumed by 'expression'
   | expression

  ;

/* 11.14 Statement list, page 108 */

statementList :   // no root node because 'statementList' may derive to the empty string

  variableDefinitions statements

  ;

variableDefinitions :

  ( variableDefinitionStatement )*

  ;

statements :

  ( statement[ true ] )*

  ;

statement[ bool breakOk ] :

    emptyStatement
  | compoundStatement
  | ( variable IsAssignedSign ) =>
    assignmentStatement
  | algorithmActionStatement
  | ( callStatement ) =>
    callStatement
  | ( ifStatement ) =>              // conflicts with 'conditionalExpression'
    ifStatement                     // in 'expressionStatement'
  | expressionStatement
  | decisionStatement
  | loopStatement
  | terminatingStatement[ breakOk ]
  | labelledStatement
  | exceptionStatement

  ;

terminatingStatement[ bool breakOk ] :

    returnStatement
  | breakStatement { breakOk }?   // ensure that break is allowed
  | loopBreakStatement
  | loopContinueStatement
  | raiseStatement

  ;

variableDefinitionStatement :

  start:DCL! localVariablesOfSort ( ","! localVariablesOfSort )* end

  { ## = #( [VariableDefinitionStatement], ## ); SETLINECOLUMN( ##, start ); }

  ;

localVariablesOfSort :

  nameDef[ SDLName::variable ] ( ","! nameDef[ SDLName::variable ] )* s:sort!
  (   IsAssignedSign! e:expression!
      { ## = #( [LocalVariablesOfSort], #s, ##, #e ); }
    | { ## = #( [LocalVariablesOfSort], #s, ## ); } )

  ;

/* 11.14.1 Compound statement, page 109 */

compoundStatement :

  start:LeftCurlyBracket! statementList RightCurlyBracket!

  { ## = #( [CompoundStatement], ## ); SETLINECOLUMN( ##, start ); }

  ;

/* 11.14.2 Transition actions and terminators as statements, page 110 */

assignmentStatement :

  assignment end

  { ## = #( [AssignmentStatement], ## ); }

  ;

algorithmActionStatement :

    output end

    { ## = #( [AlgorithmActionStatement], ## ); }

  | createRequest end

    { ## = #( [AlgorithmActionStatement], ## ); }

  | set end

    { ## = #( [AlgorithmActionStatement], ## ); }

  | reset end

    { ## = #( [AlgorithmActionStatement], ## ); }

  | xexport end

    { ## = #( [AlgorithmActionStatement], ## ); }

  ;

returnStatement :

  xreturn end

  { ## = #( [ReturnStatement], ## ); }

  ;

raiseStatement :

  raise end

  { ## = #( [RaiseStatement], ## ); }

  ;

callStatement :   // no general root node due to ambiguity in 'statement'

  ( start:CALL! { MAKEROOT( #[ExcCallStatement], start ); } )? procedureCallBody end

  ;

/* 11.14.3 Expressions as Statements, page 111 */

expressionStatement :   // no root node due to ambiguity in 'statement'

  operationApplication end

  ;

/* 11.14.4 If statement, page 111 */

ifStatement :

  start:IF! "("! /*boolean*/ expression ")"! consequenceStatement
  ( options { greedy = true; } : ELSE! alternativeStatement )?

  { ## = #( [IfStatement], ## ); SETLINECOLUMN( ##, start ); }

  ;

consequenceStatement :

  statement[ true ]

  ;

alternativeStatement :

  statement[ true ]

  ;

/* 11.14.5 Decision statement, page 111 */

decisionStatement :

  start:DECISION! "("! expression ")"!
  LeftCurlyBracket! decisionStatementBody RightCurlyBracket!

  { ## = #( [DecisionStatement], ## ); SETLINECOLUMN( ##, start ); }

  ;

decisionStatementBody :

  ( algorithmAnswerPart )+ ( algorithmElsePart )?

  { ## = #( [DecisionStatementBody], ## ); }

  ;

algorithmAnswerPart :

  start:"("! rangeCondition ")"! Colon! statement[ true ]

  { ## = #( [AlgorithmAnswerPart], ## ); SETLINECOLUMN( ##, start ); }

  ;

algorithmElsePart :

  start:ELSE! Colon! alternativeStatement

  { ## = #( [AlgorithmElsePart], ## ); SETLINECOLUMN( ##, start ); }

  ;

/* 11.14.6 Loop statement, page 112 */

loopStatement :

  start:LOOP! "("! ( loopClause ( ";"! loopClause )* )? ")"!
  loopBodyStatement ( options { greedy = true; } : THEN! finalizationStatement )?

  { ## = #( [LoopStatement], ## ); SETLINECOLUMN( ##, start ); }

  ;

loopBodyStatement :

  statement[ true ]

  ;

finalizationStatement :

  statement[ true ]

  ;

loopClause :

  ( loopVariableIndication )? ","! ( /*boolean*/ expression )? loopStep

  { ## = #( [LoopClause], ## ); }

  ;

loopStep :

  ( ","!
    (   expression
      //| ( CALL )? procedureCallBody   // subsumed by 'expression'
      { ## = #( [LoopStep], ## ); } )? )?

  ;

loopVariableIndication :

    loopVariableDefinition

    { ## = #( [LoopVariableIndication], ## ); }

  | /*variable*/ identifier ( IsAssignedSign! expression )?

    { ## = #( [LoopVariableIndication], ## ); }

  ;

loopVariableDefinition :

  start:DCL! nameDef[ SDLName::variable ] sort IsAssignedSign! expression

  { ## = #( [LoopVariableDefinition], ## ); SETLINECOLUMN( ##, start ); }

  ;

loopBreakStatement :

  start:BREAK! end

  { ## = #( [LoopBreakStatement], ## ); SETLINECOLUMN( ##, start ); }

  ;

loopContinueStatement :

  start:CONTINUE! end

  { ## = #( [LoopContinueStatement], ## ); SETLINECOLUMN( ##, start ); }

  ;

/* 11.14.7 Break and labelled statements, page 113 */

breakStatement :

  start:BREAK! /*connector*/ Name end

  { ## = #( [BreakStatement], ## ); SETLINECOLUMN( ##, start ); }

  ;

labelledStatement :

  label statement[ true ]

  { ## = #( [LabelledStatement], ## ); }

  ;

/* 11.14.8 Empty Statement, page 114 */

emptyStatement :

  end

  { ## = #( [EmptyStatement], ## ); }   // for the sake of consistency

  ;

/* 11.14.9 Exception Statement, page 114 */

exceptionStatement :

  start:TRY! tryStatement ( options { greedy = true; } : handleStatement )+

  { ## = #( [ExceptionStatement], ## ); SETLINECOLUMN( ##, start ); }

  ;

tryStatement :

  statement[ false ]

  ;

handleStatement :

  start:HANDLE! "("! exceptionStimulusList ")"! statement[ true ]

  { ## = #( [HandleStatement], ## ); SETLINECOLUMN( ##, start ); }

  ;

/* 11.15 Timer, page 114 */

timerDefinition :

  start:TIMER! timerDefinitionItem ( ","! timerDefinitionItem )* end

  { ## = #( [TimerDefinition], ## ); SETLINECOLUMN( ##, start ); }

  ;


timerDefinitionItem :

  nameDef[ SDLName::timer ] ( sortList )? ( timerDefaultInitialization )?

  { ## = #( [TimerDefinitionItem], ## ); }

  ;

timerDefaultInitialization :

  start:IsAssignedSign! /*duration*/ constantExpression

  { ## = #( [TimerDefaultInitialization], ## ); SETLINECOLUMN( ##, start ); }

  ;

reset :

  start:RESET! "("! resetClause ( ","! resetClause )* ")"!

  { ## = #( [Reset], ## ); SETLINECOLUMN( ##, start ); }

  ;

resetClause :

  /*timer*/ identifier ( "("! expressionList ")"! )?

  { ## = #( [ResetClause], ## ); }

  ;

set :

  start:SET! setClause ( ","! setClause )*

  { ## = #( [Set], ## ); SETLINECOLUMN( ##, start ); }

  ;

setClause :

  start:"("!
  ( ( expression "," ) =>
    /*time*/ expression ","! )?
  /*timer*/ identifier
  ( "("! expressionList ")"! )? ")"!

  { ## = #( [SetClause], ## ); SETLINECOLUMN( ##, start ); }

  ;

/* 11.16 Exception, page 116 */

exceptionDefinition :

  start:EXCEPTION! exceptionDefinitionItem ( ","! exceptionDefinitionItem )* end

  { ## = #( [ExceptionDefinition], ## ); SETLINECOLUMN( ##, start ); }

  ;

exceptionDefinitionItem :

  nameDef[ SDLName::exception ] ( sortList )?

  { ## = #( [ExceptionDefinitionItem], ## ); }

  ;

/* 11.16.1 Exception handler, page 116 */

exceptionHandler :

  start:EXCEPTIONHANDLER! exceptionHandlerList end
  ( onException )? ( handle )*
  ( ENDEXCEPTIONHANDLER! ( /*exceptionHandler*/ Name! )? end )?

  { ## = #( [ExceptionHandler], ## ); SETLINECOLUMN( ##, start ); }

  ;

exceptionHandlerList :

    nameDef[ SDLName::exceptionHandler ] ( ","! nameDef[ SDLName::exceptionHandler ] )*

    { ## = #( [ExceptionHandlerList], ## ); }

  | asteriskExceptionHandlerList

  ;

asteriskExceptionHandlerList :

  start:Asterisk! ( "("! nameDef[ SDLName::exceptionHandler ]
                         ( ","! nameDef[ SDLName::exceptionHandler ] )* ")"! )?

  { ## = #( [AsteriskExceptionHandlerList], ## ); SETLINECOLUMN( ##, start ); }

  ;

/* 11.16.2 On-Exception, page 118 */

onException :

  start:ONEXCEPTION! /*exceptionHandler*/ Name end

  { ## = #( [OnException], ## ); SETLINECOLUMN( ##, start ); }

  ;

/* 11.16.3 Handle, page 120 */

handle :

  start:HANDLE! ( virtuality )? exceptionStimulusList end
  ( onException )? transition

  { ## = #( [Handle], ## ); SETLINECOLUMN( ##, start ); }

  ;

exceptionStimulusList :

    exceptionStimulus ( ","! exceptionStimulus )*

    { ## = #( [ExceptionStimulusList], ## ); }

  | asteriskExceptionStimulusList

  ;

exceptionStimulus :

  /*exception*/ identifier ( "("! ( variable | { ADDVOIDCHILD(#[Void]); } )
                                  ( ","! ( variable | { ADDVOIDCHILD(#[Void]); } ) )* ")"! )?

  { ## = #( [ExceptionStimulus], ## ); }

  ;

asteriskExceptionStimulusList! :

  start:Asterisk!

  { ## = #[AsteriskExceptionStimulusList]; SETLINECOLUMN( ##, start ); }

  ;

/* 12.1 Data definitions, page 122 */

dataDefinition :

    ( dataTypeDefinition ) =>   // Problem: In a _certain_ sense 'syntypeDefinition'
    dataTypeDefinition          //          extends 'dataTypeDefinition' (large lookahead needed)
  | ( ( virtuality )? INTERFACE ) =>
    interfaceDefinition
  | syntypeDefinition
  | synonymDefinition

  ;

sort :

    basicSort
  | anchoredSort
  | expandedSort
  | referenceSort
  // pidSort        // subsumed by 'basicSort'

  ;

basicSort :   // no dedicated root node due to ambiguity in 'sort'

    /*sort*/ identifier   { ## = #( [SomeSort], ## ); }   // sort type not clear
  // | syntype            // subsumed by 'identifier'

  ;

anchoredSort :

  start:THIS! ( basicSort )?

  { ## = #( [AnchoredSort], ## ); SETLINECOLUMN( ##, start ); }

  ;

expandedSort :

  start:VALUE! basicSort

  { ## = #( [ExpandedSort], ## ); SETLINECOLUMN( ##, start ); }

  ;

referenceSort :

  start:OBJECT! basicSort

  { ## = #( [ReferenceSort], ## ); SETLINECOLUMN( ##, start ); }

  ;

pidSort :

  /*sort*/ identifier

  { false }?   // unreachable code

  ;

/* 12.1.1 Data type definition, page 124 */

dataTypeDefinition :

  typePreamble dataTypeHeading ( dataTypeSpecialization )?
  (   end
      ( ( dataTypeDefinitionBody dataTypeClosing ) =>
        dataTypeDefinitionBody dataTypeClosing end )?
    | LeftCurlyBracket! dataTypeDefinitionBody RightCurlyBracket! )

  { ## = #( [DataTypeDefinition], ## ); }

  ;

dataTypeDefinitionBody :

  ( entityInDataType )* ( dataTypeConstructor )? operations ( ( defaultInitialization )? end )?

  { ## = #( [DataTypeDefinitionBody], ## ); }

  ;

dataTypeClosing :

  ( ENDVALUE! | ENDOBJECT! ) TYPE! ( /*dataType*/ Name! )?

  ;

dataTypeHeading :

  (   ch1:VALUE!  { ADDCHILD(#[Value],ch1); }
    | ch2:OBJECT! { ADDCHILD(#[Object],ch2); } )
  TYPE! nameDef[ SDLName::dataType ]
  ( formalContextParameters )? ( virtualityConstraint )?

  { ## = #( [DataTypeHeading], ## ); }

  ;

entityInDataType :

    ( dataTypeDefinition ) =>   // Problem: In a _certain_ sense 'syntypeDefinition'
    dataTypeDefinition          //          extends 'dataTypeDefinition' (large lookahead needed)
  | syntypeDefinition
  | synonymDefinition
  | exceptionDefinition

  ;

operations :   // no root node because 'operations' may derive to the empty string

  operationSignatures operationDefinitions

  ;

dataTypeReference :

  typePreamble
  (   ch1:VALUE!  { ADDCHILD(#[Value],ch1); }
    | ch2:OBJECT! { ADDCHILD(#[Object],ch2); } )
  TYPE! /*dataType*/ identifier typeReferenceProperties

  { ## = #( [DataTypeReference], ## ); }

  ;

/* 12.1.2 Interface definition, page 124 */

interfaceDefinition :   // left-factorized all three alternatives for simpler processing

  ( virtuality )? interfaceHeading ( interfaceSpecialization )?
  (   end
      ( ( ( entityInInterface )* ( interfaceUseList )? interfaceClosing ) =>
        ( entityInInterface )* ( interfaceUseList )? interfaceClosing )?
    | LeftCurlyBracket! ( entityInInterface )* ( interfaceUseList )? RightCurlyBracket! )

  { ## = #( [InterfaceDefinition], ## ); }

  ;

interfaceClosing :

  ENDINTERFACE! ( /*interface*/ Name! )? end

  ;

interfaceHeading :

  start:INTERFACE!
  nameDef[ SDLName::interface ] ( formalContextParameters )? ( virtualityConstraint )?

  { ## = #( [InterfaceHeading], ## ); SETLINECOLUMN( ##, start ); }

  ;

entityInInterface :

    signalDefinition
  | interfaceVariableDefinition
  | interfaceProcedureDefinition
  | exceptionDefinition

  ;

interfaceUseList :

  start:USE! signalList end

  { ## = #( [InterfaceUseList], ## ); SETLINECOLUMN( ##, start ); }

  ;

interfaceVariableDefinition :

  start:DCL! nameDef[ SDLName::remoteVariable ]
             ( ","! nameDef[ SDLName:: remoteVariable ] )* s:sort! end

  { ## = #( [InterfaceVariableDefinition], #s, ## ); SETLINECOLUMN( ##, start ); }

  ;

interfaceProcedureDefinition :

  start:PROCEDURE! nameDef[ SDLName::remoteProcedure ] procedureSignature end

  { ## = #( [InterfaceProcedureDefinition], ## ); SETLINECOLUMN( ##, start ); }

  ;

interfaceReference { bool virt = false; } :

  ( virtuality { virt = true; } )? start:INTERFACE! /*interface*/ identifier typeReferenceProperties

  { ## = #( [InterfaceReference], ## ); if ( !virt ) SETLINECOLUMN( ##, start ); }

  ;

/* 12.1.3 Specialization of data types, page 126 */

dataTypeSpecialization :

  start:INHERITS! /*dataType*/ typeExpression ( renaming )? ( ch:ADDING! { ADDCHILD(#[Adding],ch); } )?

  { ## = #( [DataTypeSpecialization], ## ); SETLINECOLUMN( ##, start ); }

  ;

interfaceSpecialization :

  start:INHERITS! /*interface*/ typeExpression ( ","! /*interface*/ typeExpression )*
  ( ch:ADDING! { ADDCHILD(#[Adding],ch); } )?

  { ## = #( [InterfaceSpecialization], ## ); SETLINECOLUMN( ##, start ); }

  ;

renaming :

  start:"("! renameList ")"!

  { ## = #( [Renaming], ## ); SETLINECOLUMN( ##, start ); }

  ;

renameList :   // root node added in 'renaming'

  renamePair ( ","! renamePair )*

  ;

renamePair :

    ( operationName EqualsSign! operationName ) =>
    operationName EqualsSign! /*baseType*/ operationName

    { ## = #( [RenamePair], ## ); }

  | literalName EqualsSign! /*baseType*/ literalName

    { ## = #( [RenamePair], ## ); }

  ;

/* 12.1.4 Operations, page 127 */

operationSignatures :   // no root node because 'operationSignatures'
                        // may derive to the empty string
  ( operatorList )? ( methodList )?

  ;

operatorList :   // original rule required lookahead > 3

  start:OPERATORS! ( operationSignature end )+

  { ## = #( [OperatorList], ## ); SETLINECOLUMN( ##, start ); }

  ;

methodList :   // original rule required lookahead > 3

  start:METHODS! ( operationSignature end )+

  { ## = #( [MethodList], ## ); SETLINECOLUMN( ##, start ); }

  ;

operationSignature :

  operationPreamble
  (   operationName
    | nameClassOperation )
  ( arguments )? ( result )? ( raises )?

  { ## = #( [OperationSignature], ## ); }

  ;

operationPreamble :   // original rule derived to the empty string in both alternatives

  (   virtuality { ## = #( [OperationPreamble], ## ); } ( visibility )?
    | visibility { ## = #( [OperationPreamble], ## ); } ( virtuality )? )?

  ;

operationName :   // no node type change due to ambiguity in 'renamePair'

    /*operation*/ Name
  | QuotedOperationName

  ;

arguments :

  start:"("! argument ( ","! argument )* ")"!

  { ## = #( [Arguments], ## ); SETLINECOLUMN( ##, start ); }

  ;

argument :

  ( argumentVirtuality )? formalParameter

  { ## = #( [Argument], ## ); }

  ;

formalParameter :   // no root node due to ambiguity in 'behaviourProperty'

  parameterKind sort

  ;

argumentVirtuality! :

  start:VIRTUAL!

  { ## = #[Virtual]; SETLINECOLUMN( ##, start ); }

  ;

result :

  start:ResultSign! sort

  { ## = #( [Result], ## ); SETLINECOLUMN( ##, start ); }

  ;

/* 12.1.7 Data type constructors, page 131 */

dataTypeConstructor :

    ( ( visibility )? LITERALS ) =>
    literalList
  | ( ( visibility )? STRUCT ) =>
    structureDefinition
  | choiceDefinition

  ;

/* 12.1.7.1 Literals, page 131 */

literalList { bool visi = false; } :

  ( visibility { visi = true; } )? start:LITERALS! literalSignature ( ","! literalSignature )* end

  { ## = #( [LiteralList], ## ); if ( !visi ) SETLINECOLUMN( ##, start ); }

  ;

literalSignature :

    literalName
  | nameClassLiteral
  | namedNumber

  ;

literalName :   // no node type change due to ambiguity in 'renamePair'

    /*literal*/ Name
  | stringName

  ;

namedNumber :

  literalName EqualsSign! /*natural*/ simpleExpression

  { ## = #( [NamedNumber], ## ); }

  ;

/* 12.1.7.2 Structure data types, page 133 */

structureDefinition { bool visi = false; } :   // rearranged rule because 'end' requires LA=4

  ( visibility { visi = true; } )? start:STRUCT! ( ( field end )+ | end )

  { ## = #( [StructureDefinition], ## ); if ( !visi ) SETLINECOLUMN( ##, start ); }

  ;

field :   // left-factorized all three alternatives for simpler processing

  fieldsOfSort
  (   ch:OPTIONAL! { ADDCHILD(#[Optional],ch); }
    | fieldDefaultInitialization )?

  { ## = #( [Field], ## ); }

  ;

fieldsOfSort :

  ( visibility )? /*field*/ Name ( ","! /*field*/ Name )* s:fieldSort!

  { ## = #( [FieldsOfSort], #s, ## ); }

  ;

fieldDefaultInitialization :

  start:DEFAULT! constantExpression

  { ## = #( [FieldDefaultInitialization], ## ); SETLINECOLUMN( ##, start ); }

  ;

fieldSort :

  sort

  ;

/* 12.1.7.3 Choice data types, page 134 */

choiceDefinition { bool visi = false; } :   // rearranged rule because 'end' requires LA=4

  ( visibility { visi = true; } )? start:CHOICE! ( ( choiceOfSort end )+ | end )

  { ## = #( [ChoiceDefinition], ## ); if ( !visi ) SETLINECOLUMN( ##, start ); }

  ;

choiceOfSort :

  ( visibility )? /*field*/ Name fieldSort

  { ## = #( [ChoiceOfSort], ## ); }

  ;

/* 12.1.8 Behaviour of operations, page 135 */

operationDefinitions :

  (   ( textualOperationReference ) =>
      textualOperationReference
    | ( externalOperationDefinition ) =>
      externalOperationDefinition
    | operationDefinition )*

  ;

operationDefinition { bool entity = false; } :

  ( packageUseClause )*
  operationHeading
  (   end ( entityInOperation { entity = true; } )*
      (   operationBody
          ( ENDOPERATOR! | ENDMETHOD! ) ( ( qualifier! )? operationName! )? end
        | LeftCurlyBracket! statementList RightCurlyBracket! { entity == true }? )
    | LeftCurlyBracket! statementList RightCurlyBracket! )

  { ## = #( [OperationDefinition], ## ); }

  ;

operationHeading :

  (   ch1:OPERATOR! { ADDCHILD(#[Operator],ch1); }
    | ch2:METHOD! { ADDCHILD(#[Method],ch2); } )
  operationPreamble ( qualifier )? operationName
  ( formalOperationParameters )?
  ( operationResult )? ( raises )?

  { ## = #( [OperationHeading], ## ); }

  ;

operationIdentifier :   // no root node due to ambiguity in 'primaryExtension'

  ( qualifier { ## = #( [Identifier], ## ); } )? operationName

  ;

formalOperationParameters :

  start:"("! operationParameters ( ","! operationParameters )* ")"!

  { ## = #( [FormalOperationParameters], ## ); SETLINECOLUMN( ##, start ); }

  ;

operationParameters :

  ( argumentVirtuality )? parameterKind /*variable*/ Name
  ( ","! /*variable*/ Name )* s:sort!

  { ## = #( [OperationParameters], #s, ## ); }

  ;

entityInOperation :

    dataDefinition
  | variableDefinition
  | exceptionDefinition
  | selectDefinition
  | macroDefinition

  ;

operationBody :

  ( onException )? start ( freeAction | exceptionHandler )*

  { ## = #( [OperationBody], ## ); }

  ;

operationResult :

  start:ResultSign!
  ( ( Name sort ) =>
    /*variable*/ Name )?
  sort

  { ## = #( [OperationResult], ## ); SETLINECOLUMN( ##, start ); }

  ;

textualOperationReference :

  operationHeading REFERENCED! end

  { ## = #( [OperationReference], ## ); }

  ;

externalOperationDefinition :

  operationHeading EXTERNAL! end

  { ## = #( [ExternalOperationDefinition], ## ); }

  ;

/* 12.1.9.1 Name class, page 138 */

nameClassLiteral :

  start:NAMECLASS! regularExpression

  { ## = #( [NameClassLiteral], ## ); SETLINECOLUMN( ##, start ); }

  ;

nameClassOperation :

  /*operation*/ Name IN! NAMECLASS! regularExpression

  { ## = #( [NameClassOperation], ## ); }

  ;

regularExpression :   // greediness is justified by the last sentence on p. 138 (in-in-indirectly)

  partialRegularExpression ( options { greedy = true; } :
                             (   OR! { ## = #( [OpOr], ## ); }
                               | { ## = #( [OpConcat], ## ); } ) partialRegularExpression )*

  ;

partialRegularExpression :

  regularElement
  (   /*naturalLiteral*/ Name   { ## = #( [OpStar], ## ); }
    | PlusSign!                 { ## = #( [OpPlus], ## ); }
    | Asterisk!                 { ## = #( [OpStar], ## ); } )?

  ;

regularElement :

    "("! regularExpression ")"!
  | CharacterString
  | regularInterval

  ;

regularInterval :

  CharacterString ( Colon! | RangeSign! ) CharacterString

  { ## = #( [RegularInterval], ## ); }

  ;

/* 12.1.9.2 Name class mapping, page 139 */

spellingTerm :

  start:SPELLING! "("! /*operation*/ Name ")"!

  { ## = #( [SpellingTerm], ## ); SETLINECOLUMN( ##, start ); }

  ;

/* 12.1.9.3 Restricted visibility, page 140 */

visibility! :

    start1:PUBLIC!      { ## = #[Public];    SETLINECOLUMN( ##, start1 ); }
  | start2:PROTECTED!   { ## = #[Protected]; SETLINECOLUMN( ##, start2 ); }
  | start3:PRIVATE!     { ## = #[Private];   SETLINECOLUMN( ##, start3 ); }

  ;

/* 12.1.9.4 Syntypes, page 141 */

syntype :

  /*syntype*/ identifier

  ;

syntypeDefinition :

    start:SYNTYPE! nameDef[ SDLName::syntype ] EqualsSign! parentSortIdentifier
    ( defaultInitialization )?
    ( ( ( end )? constraint ) =>   // I assume the ambiguity is caused by LL_approx analysis
      ( end )? constraint )?
    ( ( ( end )? syntypeClosing ) =>   // I assume the ambiguity is caused by LL_approx analysis
      ( end )? syntypeClosing )?
    end

    { ## = #( [SyntypeDefinition], ## ); SETLINECOLUMN( ##, start ); }

  | typePreamble dataTypeHeading ( dataTypeSpecialization )?
    (   constraint end
      | end dataTypeDefinitionBody constraint end dataTypeClosing end
      | LeftCurlyBracket! dataTypeDefinitionBody constraint end RightCurlyBracket! )

    { ## = #( [SyntypeDefinition], ## ); }

  ;

syntypeClosing :

  ENDSYNTYPE! ( /*syntype*/ Name! )?

  ;

parentSortIdentifier :

  sort

  ;

/* 12.1.9.5 Constraint, page 142 */

constraint :

    start:CONSTANTS! "("! rangeCondition ")"!

    { ## = #( [Constraint], ## ); SETLINECOLUMN( ##, start ); }

  | sizeConstraint

    { ## = #( [Constraint], ## ); }

  ;

rangeCondition :

  range ( ","! range )*

  { ## = #( [RangeCondition], ## ); }

  ;

range :

    ( constant ( Colon! | RangeSign! ) ) =>
    closedRange
  | openRange

  ;

closedRange :

  constant ( Colon! | RangeSign! ) constant

  { ## = #( [ClosedRange], ## ); }

  ;

openRange :

    constant

    { ## = #( [OpenRange], ## ); }

  | (   start1:EqualsSign!                { MAKEROOT( #[OpEqual], start1 ); }
      | start2:NotEqualsSign!             { MAKEROOT( #[OpNotEqual], start2 ); }
      | start3:LessThanSign!              { MAKEROOT( #[OpLess], start3 ); }
      | start4:GreaterThanSign!           { MAKEROOT( #[OpGreater], start4 ); }
      | start5:LessThanOrEqualsSign!      { MAKEROOT( #[OpLessOrEqual], start5 ); }
      | start6:GreaterThanOrEqualsSign!   { MAKEROOT( #[OpGreaterOrEqual], start6 ); } )
    constant

    { ## = #( [OpenRange], ## ); }

  ;

sizeConstraint :

  start:SIZE! "("! rangeCondition ")"!

  { ## = #( [SizeConstraint], ## ); SETLINECOLUMN( ##, start ); }

  ;

constant :

  constantExpression

  ;

/* 12.1.9.6 Synonym definition, page 143 */

synonymDefinition :

  start:SYNONYM! synonymDefinitionItem ( ","! synonymDefinitionItem )* end

  { ## = #( [SynonymDefinition], ## ); SETLINECOLUMN( ##, start ); }

  ;

synonymDefinitionItem :

    ( externalSynonymDefinitionItem ) =>   // swapping alternatives simplifies LA analysis
    externalSynonymDefinitionItem
  | internalSynonymDefinitionItem

  ;

internalSynonymDefinitionItem :

  nameDef[ SDLName::synonym ] ( sort )? EqualsSign! constantExpression

  { ## = #( [InternalSynonymDefinitionItem], ## ); }

  ;

externalSynonymDefinitionItem :

  nameDef[ SDLName::synonym ] /*predefined*/ sort "="! EXTERNAL!

  { ## = #( [ExternalSynonymDefinitionItem], ## ); }

  ;

/* 12.2.1 Expressions, page 144,
 * 12.2.4 Extended primary, page 147
 * 12.2.5 Equality expression, page 147
 * 12.2.7 Operation application, page 149 (also kept for reference from 11.14.3)
 * 12.2.8 Range check expression, page 151
 *
 * Complete rewrite due to left recursion
 */

expression :

    (   CALL | THIS identifier   // sequences that are not subsumed by 'operand'
      | identifier ( actualContextParameters | ( actualParameters )? ( TO | TIMER | VIA ) ) ) =>
    valueReturningProcedureCall   // inherently nondeterministic
  | operand
  | createExpression

  ;

operand :   // I didn't find the reason for the ambiguity; nevertheless, I used the greedy option

  operand0 ( options { greedy = true; } :
             ImpliesSign!   { ## = #( [OpImplies], ## ); }
             operand0 )*

  ;

operand0 :   // I didn't find the reason for the ambiguity; nevertheless, I used the greedy option

  operand1 ( options { greedy = true; } :
             (   OR!    { ## = #( [OpOr], ## ); }
               | XOR!   { ## = #( [OpXor], ## ); } )
             operand1 )*

  ;

operand1 :   // I didn't find the reason for the ambiguity; nevertheless, I used the greedy option

  operand2 ( options { greedy = true; } :
             AND!   { ## = #( [OpAnd], ## ); }
             operand2 )*

  ;

operand2 :   // I didn't find the reason for the ambiguity; nevertheless, I used the greedy option

  operand3
  ( options { greedy = true; } :
      ( (   GreaterThanSign!           { ## = #( [OpGreater], ## ); }
          | GreaterThanOrEqualsSign!   { ## = #( [OpGreaterOrEqual], ## ); }
          | LessThanSign!              { ## = #( [OpLess], ## ); }
          | LessThanOrEqualsSign!      { ## = #( [OpLessOrEqual], ## ); }
          | IN!                        { ## = #( [OpIn], ## ); }
          | EqualsSign!                { ## = #( [OpEqual], ## ); }
          | NotEqualsSign!             { ## = #( [OpNotEqual], ## ); } )
        operand3 )
    | { ## = #( [OpRangeCheck], ## ); } rangeCheck )*

  ;

rangeCheck :

  IN! TYPE!
  (   ( identifier constraint ) =>
      /*sort*/ identifier constraint
    | sort )

  ;

operand3 :

  operand4
  ( (   PlusSign!            { ## = #( [OpPlus], ## ); }
      | Hyphen!              { ## = #( [OpBinaryMinus], ## ); }
      | ConcatenationSign!   { ## = #( [OpConcat], ## ); } )
    operand4 )*

  ;

operand4 :

  operand5
  ( (   Asterisk!   { ## = #( [OpTimes], ## ); }
      | Solidus!    { ## = #( [OpDivide], ## ); }
      | MOD!        { ## = #( [OpMod], ## ); }
      | REM!        { ## = #( [OpRem], ## ); } )
    operand5 )*

  ;

operand5 :

  (   start1:Hyphen!   { MAKEROOT( #[OpUnaryMinus], start1 ); }
    | start2:NOT!      { MAKEROOT( #[OpNot], start2 ); } )?
  primary

  ;

primary returns [ bool application = false ] :   // no dedicated root node due to ambiguity
                                                 // in various rules
  (   "("! expression ")"!
    | conditionalExpression
    | spellingTerm
    | ( ( qualifier )? CompositeBeginSign ) =>
      structurePrimary                  // taken from 'extendedPrimary'
    | ( operatorApplication ) =>        // taken from 'operationApplication'
      operatorApplication   { ( application = true ) }?  // set 'application' even in guess mode
    // | fieldName                      // taken from 'extendedPrimary'; subsumed by 'operatorApplication'
    | ( literal ) =>
      literal
    // | synonym                        // subsumed by 'operatorApplication'
    | activePrimary )
  ( application = primaryExtension )*   // check last 'primaryExtension' for 'methodApplication'

  { ## = #( [SomeExpression], ## ); }

  ;

primaryExtension returns [ bool application = false ] :   // no general root node due to ambiguity
                                                          // in 'expression'
    start1:FullStop! operationIdentifier          // taken from 'methodApplication'; there is NO conflict
    ( options { greedy = true; } : actualParameters )?         // with the former 'indexedPrimary'

    { ( application = true ) }?                  // set 'application' even in guess mode

    { ## = #( [SomePrimaryExtension], ## ); SETLINECOLUMN( ##, start1 ); }

  | start2:ExclamationMark! /*field*/ Name

    { ## = #( [SomePrimaryExtension], ## ); SETLINECOLUMN( ##, start2 ); }

  // | FullStop! /*field*/ Name                  // subsumed by 'FullStop operationIdentifier'

  | start3:"("! apl3:actualParameterList ")"!

    { SETLINECOLUMN( #apl3, start3 ); }

  | start4:LeftSquareBracket! apl4:actualParameterList RightSquareBracket!

    { SETLINECOLUMN( #apl4, start4 ); }

  ;

fieldName :

  /*field*/ Name   // name should be of type 'field', shouldn't it?

  ;

structurePrimary { bool qual = false; } :

  ( qualifier { qual = true; } )? start:CompositeBeginSign! actualParameterList CompositeEndSign!

  { ## = #( [StructurePrimary], ## ); if ( !qual ) SETLINECOLUMN( ##, start ); }

  ;

/* End of rewrite */

activePrimary :

    variableAccess
  | imperativeExpression

  ;

expressionList :

  expression ( ","! expression )*

  { ## = #( [ExpressionList], ## ); }

  ;

simpleExpression :

  constantExpression

  ;

constantExpression :

  /*constant*/ expression

  ;

/* 12.2.2 Literal, page 146 */

literal :

  literalIdentifier

  ;

literalIdentifier :   // no root node due to ambiguity in 'primary' for 'literal'

  ( qualifier { ## = #( [Identifier], ## ); } )? literalName

  ;

/* 12.2.3 Synonym, page 146 */

synonym :

  /*synonym*/ identifier

  ;

/* 12.2.6 Conditional expression, page 148 */

conditionalExpression :

  start:IF! /*boolean*/ expression THEN! consequenceExpression ELSE! alternativeExpression FI!

  { ## = #( [ConditionalExpression], ## ); SETLINECOLUMN( ##, start ); }

  ;

consequenceExpression :

  expression

  ;

alternativeExpression :

  expression

  ;

/* 12.2.7 Operation application, page 149 */

operationApplication { bool application; } :

  application = primary   // make sure that either an 'operatorApplication' is parsed
  { application }?        // without a 'primaryExtension', or a 'methodApplication' is
                          // parsed as the last 'primaryExtension'

  ;

operatorApplication :   // no root node due to ambiguity in 'primary'

  operationIdentifier ( options { greedy = true; } : actualParameters )?
  // there is NO conflict with 'primaryExtension'

  ;

// 'methodApplication' has been moved into the definition of 'primaryExtension'

/* 12.3.1 Variable definition, page 151 */

variableDefinition :

  start:DCL! ( ch:EXPORTED! { ADDCHILD(#[Exported],ch); } )? variablesOfSort ( ","! variablesOfSort )* end

  { ## = #( [VariableDefinition], ## ); SETLINECOLUMN( ##, start ); }

  ;

variablesOfSort :

  nameDef[ SDLName::variable ] ( exportedAs )?
  ( ","! nameDef[ SDLName::variable ] ( exportedAs )? )* s:sort!
  (   IsAssignedSign! e:constantExpression!
      { ## = #( [VariablesOfSort], #s, ##, #e ); }
    | { ## = #( [VariablesOfSort], #s, ## ); } )

  ;

exportedAs :

  start:AS! /*remoteVariable*/ identifier

  { ## = #( [ExportedAs], ## ); SETLINECOLUMN( ##, start ); }

  ;

/* 12.3.2 Variable access, page 152 */

variableAccess :   // no root node due to ambiguity in 'primary'

    /*variable*/ identifier
  |! start:THIS!   { ## = #[This]; SETLINECOLUMN( ##, start ); }

  ;

/* 12.3.3   Assignment and assignment attempt, page 152 &
 * 12.3.3.1 Extended variable, page 153
 *
 * Complete rewrite due to left recursion
 */

assignment :

  variable IsAssignedSign! expression

  { ## = #( [Assignment], ## ); }

  ;

variable :

  /*variable*/ identifier ( variableExtension )*

  { ## = #( [Variable], ## ); }

  ;

variableExtension :

    start1:"("! actualParameterList ")"!

    { ## = #( [VariableIndex], ## ); SETLINECOLUMN( ##, start1 ); }

  | start2:LeftSquareBracket! actualParameterList RightSquareBracket!

    { ## = #( [VariableIndex], ## ); SETLINECOLUMN( ##, start2 ); }

  | start3:ExclamationMark! fieldName

    { ## = #( [VariableField], ## ); SETLINECOLUMN( ##, start3 ); }

  | start4:FullStop! fieldName

    { ## = #( [VariableField], ## ); SETLINECOLUMN( ##, start4 ); }

  ;

/* 12.3.3.2 Default initialization, page 154 */

defaultInitialization :

  start:DEFAULT! ( virtuality )? ( constantExpression )?

  { ## = #( [DefaultInitialization], ## ); SETLINECOLUMN( ##, start ); }

  ;

/* 12.3.4 Imperative expressions, page 154 */

imperativeExpression :

    nowExpression
  | importExpression
  | pidExpression
  | timerActiveExpression
  | anyExpression
  | stateExpression

  ;

/* 12.3.4.1 Now expression, page 155 */

nowExpression! :

  start:NOW!

  { ## = #[NowExpression]; SETLINECOLUMN( ##, start ); }

  ;

/* 12.3.4.3 Pid expression, page 155 */

pidExpression! :

    start1:SELF!        { ## = #[Self];      SETLINECOLUMN( ##, start1 ); }
  | start2:PARENT!      { ## = #[Parent];    SETLINECOLUMN( ##, start2 ); }
  | start3:OFFSPRING!   { ## = #[Offspring]; SETLINECOLUMN( ##, start3 ); }
  | start4:SENDER!      { ## = #[Sender];    SETLINECOLUMN( ##, start4 ); }

  ;

createExpression :

  createRequest

  ;

/* 12.3.4.4 Timer active expression, page 156 */

timerActiveExpression :

  start:ACTIVE! "("! /*timer*/ identifier ( "("! expressionList ")"! )? ")"!

  { ## = #( [TimerActiveExpression], ## ); SETLINECOLUMN( ##, start ); }

  ;

/* 12.3.4.5 Any expression, page 156 */

anyExpression :

  start:ANY! "("! sort ")"!

  { ## = #( [AnyExpression], ## ); SETLINECOLUMN( ##, start ); }

  ;

/* 12.3.4.6 State expression, page 157 */

stateExpression! :

  start:STATE!

  { ## = #[StateExpression]; SETLINECOLUMN( ##, start ); }

  ;

/* 12.3.5 Value returning procedure call, page 157 */

valueReturningProcedureCall :   // no general root node due to ambiguity in 'expression'

    ( ( CALL )? (   THIS
                  | identifier (   ( actualContextParameters ) => actualContextParameters
                                 | ( actualParameters )? ~( TO | TIMER | VIA ) ) ) ) =>
    ( start1:CALL! { MAKEROOT( #[SomeProcedureCall], start1 ); } )?
    procedureCallBody
  | ( start2:CALL! { MAKEROOT( #[SomeProcedureCall], start2 ); } )?
    remoteProcedureCallBody

  ;

/* 13.1 Optional definition, page 158 */

selectDefinition :

  start:SELECT! IF! "("! /*boolean*/ simpleExpression ")"! end
  (   ( agentTypeReference ) =>
      agentTypeReference
    | ( dataTypeReference ) =>
      dataTypeReference
    | ( compositeStateTypeReference ) =>
      compositeStateTypeReference
    | ( procedureReference ) =>
      procedureReference
    | ( ( packageUseClause )* ( SYSTEM TYPE | typePreamble ( BLOCK TYPE | PROCESS TYPE ) ) ) =>
      agentTypeDefinition
    | ( ( packageUseClause )* ( SYSTEM | BLOCK | PROCESS ) ) =>
      agentDefinition
    | channelDefinition
    | ( typePreamble SIGNAL ) =>
      signalDefinition
    | signalListDefinition
    | remoteVariableDefinition
    | remoteProcedureDefinition
    | ( typePreamble ( VALUE | OBJECT ) TYPE | ( virtuality )? INTERFACE | SYNTYPE | SYNONYM ) =>
      dataDefinition
    | ( PROCEDURE | ( packageUseClause )* procedureHeading ) =>   // moved up for simpler
      procedureDefinition                                         // disambiguation
    | ( ( packageUseClause )* ( virtuality )? STATE ( AGGREGATION )? TYPE ) =>
      compositeStateTypeDefinition
    | statePartitioning
    | timerDefinition
    | variableDefinition
    | channelToChannelConnection
    | selectDefinition
    | macroDefinition
    | exceptionDefinition )+
  ENDSELECT! end

  { ## = #( [SelectDefinition], ## ); SETLINECOLUMN( ##, start ); }

  ;

/* 13.2 Optional transition string, page 159 */

transitionOption :

  start:ALTERNATIVE! alternativeQuestion end decisionBody ENDALTERNATIVE!

  { ## = #( [TransitionOption], ## ); SETLINECOLUMN( ##, start ); }

  ;

alternativeQuestion :

    simpleExpression
  // | informalText   // subsumed by 'simpleExpression'

  ;
