/*
 *     TreeParser for SDL-2000
 *
 * Copyright (C) 2003 Michael Schmitt <Michael.Schmitt@teststep.org>
 *
 * This program is free software; you can redistribute it and/or modify it under the terms
 * of the GNU General Public License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program;
 * if not, write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 *
 * $Id: SDLTreeParser.g,v 1.30 2003/05/21 21:31:39 schmitt Exp $
 *
 * This SDL-2000 grammar is based on the document 'SDL Forum Society Z.100 (11/99) version 3'
 * aka 'TDG04r22.doc' aka 'z.100_1001.pdf' (SDL-2000 as approved by ITU-T in September 2001).
 *
 * For the latest release of all grammar files please have a look at http://www.teststep.org.
 *
 * This file is intended to be used with ANTLR 2.7.2. ANTLR is a parser generator
 * that is based on 'linear approximate' lookahead analysis (a kind of weak LL(k) analysis).
 * ANTLR is free software provided by jGuru.com (MageLang Institute). For further information
 * please refer to http://www.antlr.org. My special thanks go to Terence Parr from jGuru.com.
 *
 * If you make use of this SDL-2000 grammar specification, please send a little email in which
 * you tell me what you are using it for and whether you consider it useful. Any grammar
 * corrections and extensions (e.g. for semantics checks and abstract syntax tree transformation)
 * are welcome.
 *
 * General considerations:
 *
 *   - the SDL parser constructs an abstract syntax tree (AST) based on the token stream delivered
 *     by 'SDLLexer'. However, as no type information about names is available at this stage of
 *     processing, several decisions have to be postponed. E.g., the parser cannot always
 *     determine whether a statement represents a remote or a 'normal' procedure call. Such
 *     ambiguities are supposed to be resolved by this tree parser that operates on the
 *     intermediate AST. For a full list of unresolvable nondeterminisms please see file
 *     SDLParser.g
 *
 * Unsolved problems:
 *
 *   - macros are not supported (and will never ever be!)
 *
 * Errors in the Z.100 SDL standard:
 *
 *   - cf. SDLParser.g
 *
 */

header "pre_include_hpp"
{
  #include "InfoAST.hpp"
  #include "SDLName.h"

  // $Id: SDLTreeParser.g,v 1.30 2003/05/21 21:31:39 schmitt Exp $
}


header "post_include_cpp"
{
  #include <string>

  using std::string;

  static string   sdlTreeParserVersion = "$Id: SDLTreeParser.g,v 1.30 2003/05/21 21:31:39 schmitt Exp $";
}


options
{
  language  = "Cpp";
  namespace = "SDLGrammar";
  namespaceStd = "std";
  namespaceAntlr = "antlr";
  genHashLines = false;
}


class SDLTreeParser extends TreeParser;

options
{
  k = 1;
  importVocab = SDLParser;
  exportVocab = SDLTreeParser;
  defaultErrorHandler = true;
  buildAST = false;
}

/* 6.3 Visibility rules, names and identifiers, page 19 */

name[ SDLName::Type type ] :

  Name

  { true /* check type */ }?

  ;

identifier[ SDLName::Type type ] :

    #( Identifier qualifier Name )

    { true /* check type */ }?

  | Name

    { true /* check type */ }?

  ;

qualifier :

  #( Qualifier ( pathItem )+ )

  ;

stringName :

    CharacterString
  | BitString
  | HexString

  ;

pathItem :

  #( PathItem ( scopeUnitKind )? Name )

  ;

scopeUnitKind :

    Package
  | SystemType
  | System
  | Block
  | BlockType
  | Process
  | ProcessType
  | State
  | StateType
  | Procedure
  | Signal
  | Type
  | Operator
  | Method
  | Interface

  ;

/* 6.4 Informal text, page 23 */

informalText :

  CharacterString

  ;

/* 6.7 Comment, page 24 */

end :

  #( End ( Comment )? )

  ;

/* 7.1 Framework, page 25 */

sdlSpecification :

  #( SDLSpecification
     ( packageDefinition | textualSystemSpecification )
     // ( packageDefinition )*   BETA: needs to be fixed; also in SDLParser
     ( definition )* )

  ;

textualSystemSpecification :

    agentDefinition
  | textualTypebasedAgentDefinition

  ;

/* 7.2 Package, page 26 */

packageDefinition :

  #( PackageDefinition ( packageUseClause )* packageHeading end ( entityInPackage )* end )

  ;

packageHeading :

  #( PackageHeading ( qualifier )? Name ( packageInterface )? )

  ;

entityInPackage :

    agentTypeDefinition
  | agentTypeReference
  | packageDefinition
  | packageReference
  | signalDefinition
  | signalReference
  | signalListDefinition
  | remoteVariableDefinition
  | dataDefinition
  | dataTypeReference
  | procedureDefinition
  | procedureReference
  | remoteProcedureDefinition
  | compositeStateTypeDefinition
  | compositeStateTypeReference
  | exceptionDefinition
  | selectDefinition
  | interfaceReference
  | association

  ;

packageReference :

  #( PackageReference ( qualifier )? name[ SDLName::package ] end )

  ;

packageUseClause :

  #( PackageUseClause identifier[ SDLName::package ] ( definitionSelectionList )? end )

  ;

definitionSelectionList :

  #( DefinitionSelectionList ( definitionSelection )+ )

  ;

definitionSelection :

  #( DefinitionSelection ( selectedEntityKind )? Name )

  ;

selectedEntityKind :

    SystemType
  | BlockType
  | ProcessType
  | Package
  | Signal
  | Procedure
  | RemoteProcedure
  | Type
  | Signallist
  | StateType
  | Synonym
  | Remote
  | Exception
  | Interface

  ;

packageInterface :

  #( PackageInterface ( definitionSelection )* )

  ;

/* 7.3 Referenced definition, page 30 */

definition :

    packageDefinition
  | agentDefinition
  | agentTypeDefinition
  | compositeState
  | compositeStateTypeDefinition
  | procedureDefinition
  | operationDefinition

  ;

/* 8.1.1.1 Agent types, page 36 */

agentTypeDefinition :

    systemTypeDefinition
  | blockTypeDefinition
  | processTypeDefinition

  ;

typePreamble :

  ( virtuality | Abstract )?

  ;

agentTypeAdditionalHeading :

  ( formalContextParameters )? ( virtualityConstraint )? agentAdditionalHeading

  ;

agentTypeReference :

    systemTypeReference
  | blockTypeReference
  | processTypeReference

  ;

/* 8.1.1.2 System type, page 33 */

systemTypeDefinition :

  #( SystemTypeDefinition ( packageUseClause )* systemTypeHeading end agentStructure end )

  ;

systemTypeHeading :

  #( SystemTypeHeading ( qualifier )? Name agentTypeAdditionalHeading )

  ;

systemTypeReference :

  #( SystemTypeReference identifier[ SDLName::systemType ] typeReferenceProperties )

  ;

/* 8.1.1.3 Block type, page 34 */

blockTypeDefinition :

  #( BlockTypeDefinition ( packageUseClause )* blockTypeHeading end agentStructure end )

  ;

blockTypeHeading :

  #( BlockTypeHeading typePreamble ( qualifier )? Name agentTypeAdditionalHeading )

  ;

blockTypeReference :

  #( BlockTypeReference typePreamble identifier[ SDLName::blockType ] typeReferenceProperties )

  ;

/* 8.1.1.4 Process type, page 34 */

processTypeDefinition :

  #( ProcessTypeDefinition ( packageUseClause )* processTypeHeading end agentStructure end )

  ;

processTypeHeading :

  #( ProcessTypeHeading typePreamble ( qualifier )? Name agentTypeAdditionalHeading )

  ;

processTypeReference :

  #( ProcessTypeReference typePreamble identifier[ SDLName::processType ] typeReferenceProperties )

  ;

/* 8.1.1.5 Composite state type, page 35 */

compositeStateTypeDefinition :

    #( CompositeStateTypeDefinition
       ( packageUseClause )*
       compositeStateTypeHeading end
       ( validInputSignalSet )?
       ( gateInDefinition )*
       ( stateConnectionPoints )*
       ( entityInCompositeState )*
       compositeStateBody end )
  | #( StateAggregationTypeDefinition
       ( packageUseClause )*
       stateAggregationTypeHeading end
       ( validInputSignalSet )?
       ( gateInDefinition )*
       ( stateConnectionPoints )*
       ( entityInCompositeState )*
       stateAggregationBody end )

  ;

compositeStateTypeHeading :

  #( CompositeStateTypeHeading
     ( virtuality )? ( qualifier )? Name
     ( formalContextParameters )? ( virtualityConstraint )?
     ( specialization )? ( agentFormalParameters )? )

  ;

stateAggregationTypeHeading :

  #( StateAggregationTypeHeading
     ( virtuality )? ( qualifier )? Name
     ( formalContextParameters )? ( virtualityConstraint )?
     ( specialization )?
     ( agentFormalParameters )? )

  ;

compositeStateTypeReference :

  #( CompositeStateTypeReference
     typePreamble identifier[ SDLName::compositeStateType ] typeReferenceProperties )

  ;

/* 8.1.2 Type expression, page 36 */

typeExpression :

  identifier[ SDLName::unspecified ] ( actualContextParameters )?

  ;

/* 8.1.3 Definitions based on types, page 37 */

textualTypebasedAgentDefinition :

    textualTypebasedSystemDefinition
  | textualTypebasedBlockDefinition
  | textualTypebasedProcessDefinition

  ;

/* 8.1.3.1 System definition based on system type, page 38 */

textualTypebasedSystemDefinition :

  #( TypebasedSystemDefinition typebasedSystemHeading end )

  ;

typebasedSystemHeading :

  Name /*system*/ typeExpression

  ;

/* 8.1.3.2 Block definition based on block type, page 38 */

textualTypebasedBlockDefinition :

  #( TypebasedBlockDefinition typebasedBlockHeading end )

  ;

typebasedBlockHeading :

  Name ( numberOfInstances )? /*block*/ typeExpression

  ;

/* 8.1.3.3 Process definition based on process type, page 39 */

textualTypebasedProcessDefinition :

  #( TypebasedProcessDefinition typebasedProcessHeading end )

  ;

typebasedProcessHeading :

  Name ( numberOfInstances )? /*process*/ typeExpression

  ;

/* 8.1.3.4 Composite state definition based on composite state type, page 39 */

textualTypebasedStatePartitionDefinition :

  #( TypebasedStatePartitionDefinition typebasedStatePartitionHeading end )

  ;

typebasedCompositeState :

  #( TypebasedCompositeState Name ( actualParameters )? /*compositeState*/ typeExpression )

  ;

/* 8.1.5 Type references, page 40 */

typeReferenceProperties :

  #( TypeReferenceProperties
     ( ( attributeProperty ) =>
       attributeProperty end )*
     ( behaviourProperty end )* end )

  ;

attributeProperty :

    ( variableProperty ) =>   // inherently nondeterministic
    variableProperty
  | ( fieldProperty ) =>      // inherently nondeterministic
    fieldProperty
  | signalParameterProperty
  | interfaceVariableProperty

  ;

variableProperty :

  ( Local | Exported )? name[ SDLName::variable ] ( sort )?

  ;

fieldProperty :

  ( symbolicVisibility )? name[ SDLName::field ] ( sort )?

  ;

symbolicVisibility :

    Private
  | Public
  | Protected

  ;

signalParameterProperty :

  sortList

  ;

interfaceVariableProperty :

  name[ SDLName::remoteVariable ] ( sort )?

  ;

behaviourProperty :

    ( ( Operator )? operationProperty ) =>             // inherently nondeterministic
    ( Operator )? operationProperty
  | ( ( Method )? operationProperty ) =>               // inherently nondeterministic
    ( Method )? operationProperty
  | ( ( Procedure )? procedureProperty ) =>            // inherently nondeterministic
    ( Procedure )? procedureProperty
  | ( ( Signal )? signalProperty ) =>                  // inherently nondeterministic
    ( Signal )? signalProperty
  | ( ( Exception )? exceptionProperty ) =>            // inherently nondeterministic
    ( Exception )? exceptionProperty
  | ( Timer )? timerProperty
  | interfaceUseList

  ;

operationProperty :

  ( symbolicVisibility )? name[ SDLName::operation ] procedureSignature

  ;

procedureProperty :

  ( Local | Exported )? name[ SDLName::procedure ] procedureSignature

  ;

signalProperty :

  name[ SDLName::signal ] ( sortList )?

  ;

exceptionProperty :

  name[ SDLName::exception ] ( sortList )?

  ;

timerProperty :

  name[ SDLName::timer ] ( sortList )?

  ;

/* 8.1.6 Gate, page 44 */

gateInDefinition :

    textualGateDefinition
  | textualInterfaceGateDefinition

  ;

textualGateDefinition :

  #( GateDefinition gate ( Adding )? gateConstraint ( gateConstraint )? end )

  ;

gate :

  Name

  ;

gateConstraint :

  #( GateConstraint ( Out | In ) ( textualEndpointConstraint )? ( signalList )? )

  ;

textualEndpointConstraint :

  #( EndpointConstraint ( Atleast )? identifier[ SDLName::unspecified ] )

  ;

textualInterfaceGateDefinition :

  #( InterfaceGateDefinition ( In | Out ) identifier[ SDLName::interface ] end )

  ;

/* 8.2 Context parameters, page 46 */

formalContextParameters :

  #( FormalContextParameters formalContextParameter ( end formalContextParameter )* )

  ;

actualContextParameters :

  #( ActualContextParameters ( actualContextParameter | Void )+ )

  ;

actualContextParameter :

  (   ( /*constant*/ primary ) =>
      /*constant*/ primary
    | identifier[ SDLName::unspecified ] )

  ;

formalContextParameter :

    agentTypeContextParameter
  | agentContextParameter
  | procedureContextParameter
  | remoteProcedureContextParameter
  | signalContextParameter
  | variableContextParameter
  | remoteVariableContextParameter
  | timerContextParameter
  | synonymContextParameter
  | sortContextParameter
  | exceptionContextParameter
  | compositeStateTypeContextParameter
  | gateContextParameter
  | interfaceContextParameter

  ;

/* 8.2.1 Agent type context parameter, page 47 */

agentTypeContextParameter :

    #( ProcessTypeContextParameter name[ SDLName::agentType ] ( agentTypeConstraint )? )
  | #( BlockTypeContextParameter name[ SDLName::agentType ] ( agentTypeConstraint )? )
  ;

agentTypeConstraint :

  #( AgentTypeConstraint
     (   identifier[ SDLName::agent ]
       | agentSignature ) )

  ;

/* 8.2.2 Agent context parameter, page 47 */

agentContextParameter :

    #( ProcessContextParameter name[ SDLName::agent ] ( agentConstraint )? )
  | #( BlockContextParameter name[ SDLName::agent ] ( agentConstraint )? )

  ;

agentConstraint :

  #( AgentConstraint
     (   ( Atleast )? identifier[ SDLName::agentType ]
       | agentSignature ) )

  ;

agentSignature :

  sortList

  ;

/* 8.2.3 Procedure context parameter, page 48 */

procedureContextParameter :

  #( ProcedureContextParameter name[ SDLName::procedure ] procedureConstraint )

  ;

procedureConstraint :

  #( ProcedureConstraint
     (   identifier[ SDLName::procedure ]
       | procedureSignatureInConstraint ) )

  ;

procedureSignatureInConstraint :

  #( ProcedureSignatureInConstraint ( formalParameter )* ( result )? )

  ;

/* 8.2.4 Remote procedure context parameter, page 48 */

remoteProcedureContextParameter :

  #( RemoteProcedureContextParameter name[ SDLName::procedure ] procedureSignatureInConstraint )

  ;

/* 8.2.5 Signal context parameter, page 48 */

signalContextParameter :

  #( SignalContextParameter ( name[ SDLName::signal ] ( signalConstraint )? )+ )

  ;

signalConstraint :

  #( SignalConstraint
     (   identifier[ SDLName::signal ]
       | sortList ) )

  ;

/* 8.2.6 Variable context parameter, page 48 */

variableContextParameter :

  #( VariableContextParameter ( ( name[ SDLName::variable ] )+ sort )+ )

  ;

/* 8.2.7 Remote variable context parameter, page 49 */

remoteVariableContextParameter :

  #( RemoteVariableContextParameter ( ( name[ SDLName::remoteVariable ] )+ sort )+ )

  ;

/* 8.2.8 Timer context parameter, page 49 */

timerContextParameter :

  #( TimerContextParameter ( name[ SDLName::timer ] ( timerConstraint )? )+ )

  ;

timerConstraint :

  sortList

  ;

/* 8.2.9 Synonym context parameter, page 49 */

synonymContextParameter :

  #( SynonymContextParameter ( name[ SDLName::synonym ] synonymConstraint )+ )

  ;

synonymConstraint :

  sort

  ;

/* 8.2.10 Sort context parameter, page 49 */

sortContextParameter :

  #( SortContextParameter ( Value | Object )? name[ SDLName::sort ]
                          ( sortConstraint )? )

  ;

sortConstraint :

  #( SortConstraint ( sort | sortSignature ) )

  ;

sortSignature :

  #( SortSignature
     ( Literals  ( literalSignature )+ )?
     ( Operators ( operationSignatureInConstraint )+ )?
     ( Methods   ( operationSignatureInConstraint )+ )? )

  ;

operationSignatureInConstraint :

  #( OperationSignatureInConstraint
     (   operationName ( formalParameter )* ( result )?
       | nameClassOperation ( result )? ) )

  ;

/* 8.2.11 Exception context parameter, page 50 */

exceptionContextParameter :

  #( ExceptionContextParameter ( name[ SDLName::exception ] ( exceptionConstraint )? )+ )

  ;

exceptionConstraint :

  sortList

  ;

/* 8.2.12 Composite state type context parameter, page 50 */

compositeStateTypeContextParameter :

  #( CompositeStateTypeContextParameter name[ SDLName::compositeStateType ]
                                        ( compositeStateTypeConstraint )? )

  ;

compositeStateTypeConstraint :

  #( CompositeStateTypeConstraint
     (   identifier[ SDLName::compositeStateType ]
       | compositeStateTypeSignature ) )

  ;

compositeStateTypeSignature :

  sortList

  ;

/* 8.2.13 Gate context parameter, page 50 */

gateContextParameter :

  #( GateContextParameter gate /*false*/ gateConstraint ( gateConstraint )? )

  ;

/* 8.2.14 Interface context parameter, page 50 */

interfaceContextParameter :

  #( InterfaceContextParameter ( name[ SDLName::interface ] ( interfaceConstraint )? )+ )

  ;

interfaceConstraint :

  #( InterfaceConstraint identifier[ SDLName::interface ] )

  ;

/* 8.3.1 Adding properties, page 51 */

specialization :

  #( Specialization typeExpression ( Adding )? )

  ;

/* 8.3.2 Virtual type, page 52 */

virtuality :

    Virtual
  | Redefined
  | Finalized

  ;

virtualityConstraint :

  #( VirtualityConstraint identifier[ SDLName::unspecified ] )

  ;

/* 8.4 Associations, page 54 */

association :

  #( Association ( name[ SDLName::association ] )? associationKind end )

  ;

associationKind :

    #( AssociationNotBoundKind          associationEnd associationEnd )
  | #( AssociationEndBoundKind          associationEnd associationEnd )
  | #( AssociationTwoEndsBoundKind      associationEnd associationEnd )
  | #( CompositionNotBoundKind          associationEnd associationEnd )
  | #( CompositionPartEndBoundKind      associationEnd associationEnd )
  | #( CompositionCompositeEndBoundKind associationEnd associationEnd )
  | #( CompositionTwoEndsBoundKind      associationEnd associationEnd )
  | #( AggregationNotBoundKind          associationEnd associationEnd )
  | #( AggregationPartEndBoundKind      associationEnd associationEnd )
  | #( AggregationAggregateEndBoundKind associationEnd associationEnd )
  | #( AggregationTwoEndsBoundKind      associationEnd associationEnd )

  ;

associationEnd :

  #( AssociationEnd
     ( visibility )? ( ( name[ SDLName::role ] ) => name[ SDLName::role ] )?
     linkedTypeIdentifier ( rangeCondition )? ( Ordered )? )

  ;

linkedTypeIdentifier :   // Problem: Ambiguity if a name is reused for entities of different kinds

    ( identifier[ SDLName::agentType ] ) =>
    identifier[ SDLName::agentType ]
  | ( identifier[ SDLName::dataType ] ) =>
    identifier[ SDLName::dataType ]
  | identifier[ SDLName::interface ]

  ;

/* 9 Agents, page 57 */

agentDefinition :

    systemDefinition
  | blockDefinition
  | processDefinition

  ;

agentStructure :

  #( AgentStructure ( validInputSignalSet )?
                      (   entityInAgent
                        | channelToChannelConnection
                        | channelDefinition
                        | gateInDefinition
                        | agentDefinition
                        | agentReference
                        | textualTypebasedAgentDefinition )*
                      (   ( statePartitioning )?
                        | agentBody ) )

  ;

agentInstantiation :

  #( AgentInstantiation ( numberOfInstances )? agentAdditionalHeading )

  ;

agentAdditionalHeading :

  ( specialization )? ( agentFormalParameters )?

  ;

entityInAgent :

    signalDefinition
  | signalReference
  | signalListDefinition
  | variableDefinition
  | remoteProcedureDefinition
  | remoteVariableDefinition
  | dataDefinition
  | dataTypeReference
  | timerDefinition
  | interfaceReference
  | exceptionDefinition
  | procedureReference
  | procedureDefinition
  | compositeStateTypeDefinition
  | compositeStateTypeReference
  | selectDefinition
  | agentTypeDefinition
  | agentTypeReference
  | association

  ;

agentReference :

    blockReference
  | processReference

  ;

validInputSignalSet :

  #( ValidInputSignalSet ( signalList )? end )

  ;

agentBody :

  #( AgentBody ( ( onException )? start )? ( state | exceptionHandler | freeAction )* )

  ;

agentFormalParameters :

  #( AgentFormalParameters ( parametersOfSort )+ )

  ;

parametersOfSort :

  #( ParametersOfSort sort ( name[ SDLName::variable ] )+ )

  ;

numberOfInstances :

  #( NumberOfInstances ( /*natural*/ simpleExpression | Void )
                       ( /*natural*/ simpleExpression | Void )? )

  ;

/* 9.1 System, page 63 */

systemDefinition :

  #( SystemDefinition ( packageUseClause )* systemHeading end agentStructure end )

  ;

systemHeading :

  #( SystemHeading Name agentAdditionalHeading )

  ;

/* 9.2 Block, page 64 */

blockDefinition :

  #( BlockDefinition ( packageUseClause )* blockHeading end agentStructure end )

  ;

blockHeading :

  #( BlockHeading ( qualifier )? Name agentInstantiation )

  ;

blockReference :

  #( BlockReference name[ SDLName::block ] ( numberOfInstances )? end )

  ;

/* 9.3 Process, page 65 */

processDefinition :

  #( ProcessDefinition ( packageUseClause )* processHeading end agentStructure end )

  ;

processHeading :

  #( ProcessHeading ( qualifier )? Name agentInstantiation )

  ;

processReference :

  #( ProcessReference name[ SDLName::process ] ( numberOfInstances )? end )

  ;

/* 9.4 Procedure, page 66 */

procedureDefinition :

    externalProcedureDefinition
  | #( ProcedureDefinition ( packageUseClause )*
                             procedureHeading
                             ( end ( entityInProcedure )* )?
                             (   procedureBody
                               | ( virtuality )? statementList ) )

  ;

procedurePreamble :

    #( ProcedurePreamble typePreamble ( Exported ( identifier[ SDLName::remoteProcedure ] )? )? )
  | /* empty preamble */

  ;

procedureHeading :

  #( ProcedureHeading
     procedurePreamble
     ( qualifier )? Name
     ( formalContextParameters )? ( virtualityConstraint )?
     ( specialization )?
     ( procedureFormalParameters )?
     ( procedureResult )? ( raises )? )

  ;

procedureFormalParameters :

  #( ProcedureFormalParameters ( formalVariableParameters )+ )

  ;

formalVariableParameters :

  #( FormalVariableParameters parameterKind parametersOfSort )

  ;

parameterKind :

  (   InOut
    | In
    | Out )?

  ;

procedureResult :

  #( ProcedureResult ( name[ SDLName::variable ] )? sort )

  ;

raises :

  #( Raises ( identifier[ SDLName::exception ] )+ )

  ;

entityInProcedure :

    variableDefinition
  | dataDefinition
  | dataTypeReference
  | procedureReference
  | procedureDefinition
  | compositeStateTypeDefinition
  | compositeStateTypeReference
  | exceptionDefinition
  | selectDefinition

  ;

procedureSignature :

  ( formalParameter )* ( result )? ( raises )?


  ;

procedureBody :

  #( ProcedureBody ( onException )? ( start )? ( state | exceptionHandler | freeAction )* )

  ;

externalProcedureDefinition :

  #( ExternalProcedureDefinition name[ SDLName::procedure ] procedureSignature end )

  ;

procedureReference :

  #( ProcedureReference typePreamble identifier[ SDLName::procedure ] typeReferenceProperties )

  ;

/* 10.1 Channel, page 70 */

channelDefinition :

  #( ChannelDefinition ( Name )? ( Nodelay )? channelPath ( channelPath )? end )

  ;

channelPath :

  #( ChannelPath channelEndpoint channelEndpoint ( signalList )? end )

  ;

channelEndpoint :   // Problem: Ambiguity if a name is reused for entities of different kinds

  #( ChannelEndpoint
     (   ( identifier[ SDLName::agent ] ) =>
         identifier[ SDLName::agent ]
       | ( identifier[ SDLName::state ] ) =>
         identifier[ SDLName::state ]
       | Env
       | This )
     ( viaGate )? )

  ;

viaGate :

  #( ViaGate gate /*false*/ )

  ;

/* 10.2 Connection, page 72 */

channelToChannelConnection :

  #( ChannelToChannelConnection externalChannelIdentifiers channelIdentifiers end )

  ;

externalChannelIdentifiers :

  #( ExternalChannelIdentifiers ( identifier[ SDLName::channel ] )+ )

  ;

channelIdentifiers :

  #( ChannelIdentifiers ( identifier[ SDLName::channel ] )+ )

  ;

/* 10.3 Signal, page 73 */

signalDefinition :

  #( SignalDefinition typePreamble ( signalDefinitionItem )+ end )

  ;

signalDefinitionItem :

  #( SignalDefinitionItem
     Name
     ( formalContextParameters )? ( virtualityConstraint )?
     ( specialization )? ( sortList )? )

  ;

sortList :

  #( SortList ( sort )+ )

  ;

signalReference :

  #( SignalReference typePreamble identifier[ SDLName::signal ] typeReferenceProperties )

  ;

/* 10.4 Signal list definition, page 74 */

signalListDefinition :

  #( SignalListDefinition Name signalList end )

  ;

signalList :

  #( SignalList ( signalListItem )+ )

  ;

signalListItem :

  #( SignalListItem
     (   ( identifier[ SDLName::signal ] ) =>
         identifier[ SDLName::signal ]
       | ( identifier[ SDLName::timer ] ) =>
         identifier[ SDLName::timer ]
       | ( identifier[ SDLName::signalList ] ) =>
         identifier[ SDLName::signalList ]
       | ( ( Procedure )? identifier[ SDLName::remoteProcedure ] ) =>
         ( Procedure )? identifier[ SDLName::remoteProcedure ]
       | ( ( Interface )? identifier[ SDLName::interface ] ) =>
         ( Interface )? identifier[ SDLName::interface ]
       | ( Remote )? identifier[ SDLName::remoteVariable ] ) )

  ;

/* 10.5 Remote procedures, page 74 */

remoteProcedureDefinition :

  #( RemoteProcedureDefinition Name ( Nodelay )? procedureSignature end )

  ;

remoteProcedureCall :

  #( SomeProcedureCall remoteProcedureCallBody )

  ;

remoteProcedureCallBody :

  #( SomeExpression
     identifier[ SDLName::remoteProcedure ] ( actualParameters )? communicationConstraints )

  ;

communicationConstraints :

  (   #( CommConDestination destination )
    | #( CommConTimer identifier[ SDLName::timer ] )
    | viaPath )*

  ;

/* 10.6 Remote variables, page 77 */

remoteVariableDefinition :

  #( RemoteVariableDefinition ( ( Name )+ sort ( Nodelay )? )+ end )

  ;

importExpression :

  #( ImportExpression identifier[ SDLName::remoteVariable ] communicationConstraints )

  ;

xexport :   // export is a reserved keyword

  #( Export ( identifier[ SDLName::variable ] )+ )

  ;

/* 11.1 Start, page 79 */

start :

  #( Start ( virtuality )? ( name[ SDLName::stateEntryPoint ] )? end ( onException )? transition )

  ;

/* 11.2 State, page 80 */

state :

  #( State // basicState |             // 'basicState' is subsumed by 'compositeStateApplication'
     compositeStateApplication )

  ;

/* 11.2.1 Basic State, page 80 */

asteriskStateList :

  #( AsteriskStateList ( name[ SDLName::state ] )* )

  ;

/* 11.2.2 Composite state application, page 82 */

compositeStateApplication :

  compositeStateList end ( onException )?
  (   inputPart
    | priorityInput
    | savePart
    | spontaneousTransition
    | continuousSignal
    | connectPart )*
  ( end )?

  ;

compositeStateList :

    ( stateListToken )+
  | asteriskStateList

  ;

stateListToken :

    ( typebasedCompositeState ) =>
    typebasedCompositeState
  | name[ SDLName::state ] ( actualParameters )?

  ;

/* 11.3 Input, page 83 */

inputPart :

  #( InputPart ( virtuality )? inputList end
                 ( onException )? ( enablingCondition )? transition )

  ;

inputList :

    #( InputList ( stimulus )+ )
  | AsteriskInputList

  ;

stimulus :

  #( Stimulus
     signalListItem
     (   ( variable | Void )+
       | remoteProcedureReject )? )

  ;

remoteProcedureReject :

  #( RemoteProcedureReject exceptionRaise )

  ;

/* 11.4 Priority Input, page 84 */

priorityInput :

  #( PriorityInput ( virtuality )? priorityInputList end ( onException )? transition )

  ;

priorityInputList :

  #( PriorityInputList ( stimulus )+ )

  ;

/* 11.5 Continuous signal, page 85 */

continuousSignal :

  #( ContinuousSignal ( virtuality )? continuousExpression end
                        ( priorityName end )? ( onException )? transition )

  ;

continuousExpression :

  /*boolean*/ expression

  ;

priorityName :

  name[ SDLName::naturalLiteral ]

  ;

/* 11.6 Enabling condition, page 86 */

enablingCondition :

  #( EnablingCondition /*boolean*/ expression end )

  ;

/* 11.7 Save, page 86 */

savePart :

  #( SavePart ( virtuality )? saveList end )

  ;

saveList :

    signalList
  | AsteriskSaveList

  ;

/* 11.9 Spontaneous transition, page 87 */

spontaneousTransition :

  #( SpontaneousTransition ( virtuality )? end
                           ( onException )? ( enablingCondition )? transition )

  ;

/* 11.10 Label, page 88 */

label :

  #( Label Name )

  ;

freeAction :

  #( FreeAction transition ( end )? )

  ;

/* 11.11 State machine and Composite state, page 89 */

compositeState :

    compositeStateGraph
  | stateAggregation

  ;

/* 11.11.1 Composite State Graph, page 90 */

compositeStateGraph :

  #( CompositeStateGraph ( packageUseClause )* compositeStateHeading end
                           ( validInputSignalSet )?
                           ( gateInDefinition )*
                           ( stateConnectionPoints )*
                           entityInCompositeState
                           compositeStateBody end )

  ;

compositeStateHeading :

  #( CompositeStateHeading ( qualifier )? Name ( agentFormalParameters )? )

  ;

entityInCompositeState :

    variableDefinition
  | dataDefinition
  | selectDefinition
  | dataTypeReference
  | procedureDefinition
  | exceptionDefinition
  | compositeStateTypeDefinition
  | compositeStateTypeReference

  ;

compositeStateBody :

  #( CompositeStateBody ( onException )? ( start )* ( state | exceptionHandler | freeAction )* )

  ;

compositeStateReference :

  #( CompositeStateReference name[ SDLName::compositeState ] end )

  ;

/* 11.11.2 State aggregation, page 91 */

stateAggregation :

  #( StateAggregation ( packageUseClause )*
                        stateAggregationHeading end
                        ( stateConnectionPoints )*
                        entityInCompositeState
                        stateAggregationBody end )

  ;

stateAggregationHeading :

  #( StateAggregationHeading ( qualifier )? Name ( agentFormalParameters )? )

  ;

stateAggregationBody :

  (   statePartitioning
    | statePartitionConnection )+

  ;

statePartitioning :

    compositeStateReference
  | textualTypebasedStatePartitionDefinition
  | compositeState

  ;

statePartitionConnection :

  #( StatePartitionConnection outerEntryPoint innerEntryPoint end )

  ;

outerEntryPoint :

  #( OuterEntryPoint identifier[ SDLName::statePartition ] point )

  ;

innerEntryPoint :

  #( InnerEntryPoint identifier[ SDLName::statePartition ] point )

  ;

point :

    ( name[ SDLName::stateEntryPoint ] ) =>
    name[ SDLName::stateEntryPoint ]
  | name[ SDLName::stateExitPoint ]
  | Default

  ;

typebasedStatePartitionHeading :

  Name /*compositeState*/ typeExpression

  ;

/* 11.11.3 State connection point, page 94 */

stateConnectionPoints :

  #( StateConnectionPoints
     (   #( StateEntryPoints ( name[ SDLName::stateEntryPoint ] )+ )
       | #( StateExitPoints  ( name[ SDLName::stateExitPoint  ] )+ ) end ) )

  ;

/* 11.11.4 Connect, page 95 */

connectPart :

  #( ConnectPart ( virtuality )? ( connectList )? end ( onException )? transition )

  ;

connectList :

    #( ConnectList ( name[ SDLName::stateExitPoint ] )+ )
  | AsteriskConnectList

  ;

/* 11.12.1 Transition body, page 95 */

transition :

  #( Transition (   ( actionStatement )+ ( terminatorStatement )?
                  | terminatorStatement ) )

  ;

actionStatement :

  #( ActionStatement ( label )? (   action1 end ( onException )?
                                  | action2 end ) )

  ;

action1 :

    task
  | output
  | createRequest
  | decision
  | set
  | reset
  | xexport
  | ( procedureCall ) =>
    procedureCall
  | remoteProcedureCall

  ;

action2 :

  transitionOption

  ;

terminatorStatement :

  #( TerminatorStatement ( label )? (   terminator1 end ( onException )?
                                        | terminator2 end ) )

  ;

terminator1 :

    xreturn
  | raise

  ;

terminator2 :

    nextstate
  | join
  | Stop

  ;

/* 11.12.2.1 Nextstate, page 98 */

nextstate :

    #( Nextstate /*state*/ Name ( actualParameters )? ( /*stateEntryPoint*/ Name )? )
  | DashNextstate
  | HistoryDashNextstate

  ;

/* 11.12.2.2 Join, page 99 */

join :

  #( Join /*connector*/ Name )

  ;

/* 11.12.2.4 Return, page 100 */

xreturn :   // return is a reserved keyword

  #( Return
     (   expression
       | /*stateExitPoint*/ Name )? )

  ;

/* 11.12.2.5 Raise, page 100 */

raise :

  #( Raise exceptionRaise )

  ;

exceptionRaise :

  #( ExceptionRaise identifier[ SDLName::exception ] ( actualParameters )? )

  ;

/* 11.13.1 Task, page 101 */

task :

  #( Task textualTaskBody )

  ;

textualTaskBody :

    assignment
  | informalText
  | compoundStatement

  ;

/* 11.13.2 Create, page 102 */

createRequest :

  #( CreateRequest createBody )

  ;

createBody :   // Problem: Ambiguity if a name is reused for entities of different kinds

  (   ( identifier[ SDLName::agent ] ) =>
      identifier[ SDLName::agent ]
    | ( identifier[ SDLName::agentType ] ) =>
      identifier[ SDLName::agentType ]
    | This )
  ( actualParameters )?

  ;

actualParameters :

  #( ActualParameters ( expression | Void )* )

  ;

/* 11.13.3 Procedure call, page 103 */

procedureCall :

  #( SomeProcedureCall procedureCallBody )

  ;

procedureCallBody :

  #( SomeExpression
     ( This )?
     (   ( /*procedure*/ typeExpression ) =>
         /*procedure*/ typeExpression
       | identifier[ SDLName::procedure ] )
     ( actualParameters )? )

  ;

/* 11.13.4 Output, page 104 */

output :

  #( Output outputBody )

  ;

outputBody :

  ( identifier[ SDLName::signal ] ( actualParameters )? )+ communicationConstraints

  ;

destination :

    ( /*pid*/ expression ) =>
    /*pid*/ expression
  | identifier[ SDLName::agent ]
  | This

  ;

viaPath :

  #( ViaPath ( /*channel,gate */ identifier[ SDLName::channel ] ) )

  ;

/* 11.13.5 Decision, page 106 */

decision :

  #( Decision question end ( onException )? decisionBody )

  ;

decisionBody :

  #( DecisionBody ( answerPart )+ ( elsePart )? )

  ;

answerPart :

  #( AnswerPart ( answer )? ( transition )? )

  ;

answer :

    ( rangeCondition ) =>
    rangeCondition
  | informalText

  ;

elsePart :

  #( ElsePart ( transition )? )

  ;

question :

    ( expression ) =>
    expression
  | informalText
  | Any

  ;

/* 11.14 Statement list, page 108 */

statementList :

  ( variableDefinitionStatement )* ( statement )*

  ;

statement :

    emptyStatement
  | compoundStatement
  | assignmentStatement
  | algorithmActionStatement
  | ( callStatement ) =>
    callStatement
  | ifStatement
  | expressionStatement
  | decisionStatement
  | loopStatement
  | terminatingStatement
  | labelledStatement
  | exceptionStatement

  ;

terminatingStatement :

    returnStatement
  | breakStatement
  | loopBreakStatement
  | loopContinueStatement
  | raiseStatement

  ;

variableDefinitionStatement :

  #( VariableDefinitionStatement ( localVariablesOfSort )+ end )

  ;

localVariablesOfSort :

  #( LocalVariablesOfSort sort ( Name )+ ( expression )? )

  ;

/* 11.14.1 Compound statement, page 109 */

compoundStatement :

  #( CompoundStatement statementList )

  ;

/* 11.14.2 Transition actions and terminators as statements, page 110 */

assignmentStatement :

  #( AssignmentStatement assignment end )

  ;

algorithmActionStatement :

  #( AlgorithmActionStatement
     (   output end
       | createRequest end
       | set end
       | reset end
       | xexport end ) )

  ;

returnStatement :

  #( ReturnStatement xreturn end )

  ;

raiseStatement :

  #( RaiseStatement raise end )

  ;

callStatement :

  procedureCallBody end

  ;

/* 11.14.3 Expressions as Statements, page 111 */

expressionStatement :

  operationApplication end

  ;

/* 11.14.4 If statement, page 111 */

ifStatement :

  #( IfStatement /*boolean*/ expression statement ( statement )? )

  ;

/* 11.14.5 Decision statement, page 111 */

decisionStatement :

  #( DecisionStatement expression decisionStatementBody )

  ;

decisionStatementBody :

  #( DecisionStatementBody ( algorithmAnswerPart )+ ( algorithmElsePart )? )

  ;

algorithmAnswerPart :

  #( AlgorithmAnswerPart rangeCondition statement )

  ;

algorithmElsePart :

  #( AlgorithmElsePart statement )

  ;

/* 11.14.6 Loop statement, page 112 */

loopStatement :

  #( LoopStatement ( loopClause )* statement ( statement )? )

  ;

loopClause :

  #( LoopClause ( loopVariableIndication )? ( /*boolean*/ expression )?
                ( loopStep )? )

  ;

loopStep :

  #( LoopStep (   expression
                //| ( CALL )? procedureCallBody   // subsumed by 'expression'
                | Void ) )

  ;

loopVariableIndication :

  #( LoopVariableIndication (   loopVariableDefinition
                              | identifier[ SDLName::variable ] ( expression )? ) )

  ;

loopVariableDefinition :

  #( LoopVariableDefinition Name sort expression )

  ;

loopBreakStatement :

  #( LoopBreakStatement end )

  ;

loopContinueStatement :

  #( LoopContinueStatement end )

  ;

/* 11.14.7 Break and labelled statements, page 113 */

breakStatement :

  #( BreakStatement /*connector*/ Name end )

  ;

labelledStatement :

  #( LabelledStatement label statement )

  ;

/* 11.14.8 Empty Statement, page 114 */

emptyStatement :

  #( EmptyStatement end )

  ;

/* 11.14.9 Exception Statement, page 114 */

exceptionStatement :

  #( ExceptionStatement statement ( handleStatement )+ )

  ;

handleStatement :

  #( HandleStatement exceptionStimulusList statement )

  ;

/* 11.15 Timer, page 114 */

timerDefinition :

  #( TimerDefinition ( timerDefinitionItem )+ end )

  ;


timerDefinitionItem :

  #( TimerDefinitionItem Name ( sortList )? ( timerDefaultInitialization )? )

  ;

timerDefaultInitialization :

  /*duration*/ constantExpression

  ;

reset :

  #( Reset ( resetClause )+ )

  ;

resetClause :

  #( ResetClause identifier[ SDLName::timer ] ( expressionList )? )

  ;

set :

  #( Set ( setClause )+ )

  ;

setClause :

  #( SetClause
     ( /*time*/ expression )?
     identifier[ SDLName::timer ] ( expressionList )? )

  ;

/* 11.16 Exception, page 116 */

exceptionDefinition :

  #( ExceptionDefinition ( exceptionDefinitionItem )+ end )

  ;

exceptionDefinitionItem :

  #( ExceptionDefinitionItem Name ( sortList )? )

  ;

/* 11.16.1 Exception handler, page 116 */

exceptionHandler :

  #( ExceptionHandler
     exceptionHandlerList end ( onException )? ( handle )* ( end )? )

  ;

exceptionHandlerList :

    #( ExceptionHandlerList ( Name )+ )
  | #( AsteriskExceptionHandlerList ( Name )* )

  ;

/* 11.16.2 On-Exception, page 118 */

onException :

  #( OnException /*exceptionHandler*/ Name end )

  ;

/* 11.16.3 Handle, page 120 */

handle :

  #( Handle ( virtuality )? exceptionStimulusList end ( onException )? transition )

  ;

exceptionStimulusList :

    #( ExceptionStimulusList ( exceptionStimulus )+ )
  | AsteriskExceptionStimulusList

  ;

exceptionStimulus :

  #( ExceptionStimulus identifier[ SDLName::exception ] ( variable | Void )* )

  ;

/* 12.1 Data definitions, page 122 */

dataDefinition :

    dataTypeDefinition
  | interfaceDefinition
  | syntypeDefinition
  | synonymDefinition

  ;

sort :

    ( basicSort ) =>   // conflict with 'pidSort'
    basicSort
  | anchoredSort
  | expandedSort
  | referenceSort
  | pidSort

  ;

basicSort :

  #( SomeSort (   ( identifier[ SDLName::sort ] ) =>
                  identifier[ SDLName::sort ]
                | syntype ) )

  ;

anchoredSort :

  #( AnchoredSort ( basicSort )? )

  ;

expandedSort :

  #( ExpandedSort basicSort )

  ;

referenceSort :

  #( ReferenceSort basicSort )

  ;

pidSort :

  #( SomeSort identifier[ SDLName::sort ] )

  ;

/* 12.1.1 Data type definition, page 124 */

dataTypeDefinition :

  #( DataTypeDefinition typePreamble dataTypeHeading ( dataTypeSpecialization )?
                        (   end
                            ( dataTypeDefinitionBody end )?
                          | dataTypeDefinitionBody ) )

  ;

dataTypeDefinitionBody :

  #( DataTypeDefinitionBody ( entityInDataType )* ( dataTypeConstructor )?
                              operations ( defaultInitialization )? )

  ;

dataTypeHeading :

  #( DataTypeHeading
     ( Value | Object ) Name ( formalContextParameters )? ( virtualityConstraint )? )

  ;

entityInDataType :

    dataTypeDefinition
  | syntypeDefinition
  | synonymDefinition
  | exceptionDefinition

  ;

operations :

  operationSignatures operationDefinitions

  ;

dataTypeReference :

  #( DataTypeReference typePreamble ( Value | Object ) identifier[ SDLName::dataType ]
                       typeReferenceProperties )

  ;

/* 12.1.2 Interface definition, page 124 */

interfaceDefinition :

  #( InterfaceDefinition ( virtuality )? interfaceHeading ( interfaceSpecialization )?
                           (   end
                               ( ( entityInInterface )* ( interfaceUseList )? end )?
                             | ( entityInInterface )* ( interfaceUseList )? ) )

  ;

interfaceHeading :

  #( InterfaceHeading Name ( formalContextParameters )? ( virtualityConstraint )? )

  ;

entityInInterface :

    signalDefinition
  | interfaceVariableDefinition
  | interfaceProcedureDefinition
  | exceptionDefinition

  ;

interfaceUseList :

  #( InterfaceUseList signalList end )

  ;

interfaceVariableDefinition :

  #( InterfaceVariableDefinition ( Name )+ sort end )

  ;

interfaceProcedureDefinition :

  #( InterfaceProcedureDefinition Name procedureSignature end )

  ;

interfaceReference :

  #( InterfaceReference ( virtuality )? identifier[ SDLName::interface ] typeReferenceProperties )

  ;

/* 12.1.3 Specialization of data types, page 126 */

dataTypeSpecialization :

  #( DataTypeSpecialization /*dataType*/ typeExpression ( renaming )? ( Adding )? )

  ;

interfaceSpecialization :

  #( InterfaceSpecialization ( /*interface*/ typeExpression )+ ( Adding )? )

  ;

renaming :

  #( Renaming ( renamePair )+ )

  ;

renamePair :

  #( RenamePair (   ( operationName /*baseType*/ operationName ) =>
                      operationName /*baseType*/ operationName
                    | literalName /*baseType*/ literalName ) )

  ;

/* 12.1.4 Operations, page 127 */

operationSignatures :

  ( operatorList )? ( methodList )?

  ;

operatorList :

  #( OperatorList ( operationSignature end )+ )

  ;

methodList :

  #( MethodList ( operationSignature end )+ )

  ;

operationSignature :

  #( OperationSignature operationPreamble
                          (   operationName
                            | nameClassOperation )
                          ( arguments )? ( result )? ( raises )? )

  ;

operationPreamble :   // original rule derived to the empty string in both alternatives

  (   virtuality ( visibility )?
    | visibility ( virtuality )? )?

  ;

operationName :

    /*operation*/ Name
  | QuotedOperationName

  ;

arguments :

  #( Arguments ( argument )+ )

  ;

argument :

  #( Argument ( argumentVirtuality )? formalParameter )

  ;

formalParameter :

  parameterKind sort

  ;

argumentVirtuality :

  Virtual

  ;

result :

  #( Result sort )

  ;

/* 12.1.7 Data type constructors, page 131 */

dataTypeConstructor :

    literalList
  | structureDefinition
  | choiceDefinition

  ;

/* 12.1.7.1 Literals, page 131 */

literalList :

  #( LiteralList ( visibility )? ( literalSignature )+ end )

  ;

literalSignature :

    literalName
  | nameClassLiteral
  | namedNumber

  ;

literalName :

    name[ SDLName::literal ]
  | stringName

  ;

namedNumber :

  #( NamedNumber literalName /*natural*/ simpleExpression )

  ;

/* 12.1.7.2 Structure data types, page 133 */

structureDefinition :

  #( StructureDefinition ( visibility )? ( ( field end )+ | end ) )

  ;

field :

  #( Field fieldsOfSort ( Optional | fieldDefaultInitialization )? )

  ;

fieldsOfSort :

  #( FieldsOfSort fieldSort ( visibility )? ( /*field*/ Name )+ )

  ;

fieldDefaultInitialization :

  #( FieldDefaultInitialization constantExpression )

  ;

fieldSort :

  sort

  ;

/* 12.1.7.3 Choice data types, page 134 */

choiceDefinition :

  #( ChoiceDefinition ( visibility )? ( ( choiceOfSort end )+ | end ) )

  ;

choiceOfSort :

  #( ChoiceOfSort ( visibility )? /*field*/ Name fieldSort )

  ;

/* 12.1.8 Behaviour of operations, page 135 */

operationDefinitions :

  (   textualOperationReference
    | externalOperationDefinition
    | operationDefinition )*

  ;

operationDefinition :

  #( OperationDefinition
     ( packageUseClause )* operationHeading
     (   end ( entityInOperation )*
         (   operationBody end
           | statementList )
       | statementList ) )

  ;

operationHeading :

  #( OperationHeading
     ( Operator | Method ) operationPreamble ( qualifier )? operationName
     ( formalOperationParameters )? ( operationResult )? ( raises )? )

  ;

operationIdentifier :

    #( Identifier qualifier operationName )
  | operationName

  ;

formalOperationParameters :

  #( FormalOperationParameters ( operationParameters )+ )

  ;

operationParameters :

  #( OperationParameters sort ( argumentVirtuality )? parameterKind ( /*variable*/ Name )+ )

  ;

entityInOperation :

    dataDefinition
  | variableDefinition
  | exceptionDefinition
  | selectDefinition

  ;

operationBody :

  #( OperationBody ( onException )? start ( freeAction | exceptionHandler )* )

  ;

operationResult :

  #( OperationResult ( /*variable*/ Name )? sort )

  ;

textualOperationReference :

  #( OperationReference operationHeading end )

  ;

externalOperationDefinition :

  #( ExternalOperationDefinition operationHeading end )

  ;

/* 12.1.9.1 Name class, page 138 */

nameClassLiteral :

  #( NameClassLiteral regularExpression )

  ;

nameClassOperation :

  #( NameClassOperation /*operation*/ Name regularExpression )

  ;

regularExpression :

    CharacterString
  | #( RegularInterval CharacterString CharacterString )
  | #( OpStar regularExpression ( Name )? )
  | #( OpPlus regularExpression )
  | #( OpOr   regularExpression regularExpression )

  ;

/* 12.1.9.2 Name class mapping, page 139 */

spellingTerm :

  #( SpellingTerm /*operation*/ Name )

  ;

/* 12.1.9.3 Restricted visibility, page 140 */

visibility :

    Public
  | Protected
  | Private

  ;

/* 12.1.9.4 Syntypes, page 141 */

syntype :

  identifier[ SDLName::syntype ]

  ;

syntypeDefinition :

  #( SyntypeDefinition
     (   Name parentSortIdentifier ( defaultInitialization )? ( constraint end )? ( end )?
       | typePreamble dataTypeHeading ( dataTypeSpecialization )?
         (   constraint end
           | end dataTypeDefinitionBody constraint end end
           | dataTypeDefinitionBody constraint end ) ) )

  ;

parentSortIdentifier :

  sort

  ;

/* 12.1.9.5 Constraint, page 142 */

constraint :

  #( Constraint
     (   rangeCondition
       | sizeConstraint ) )

  ;

rangeCondition :

  #( RangeCondition ( range )+ )

  ;

range :

    closedRange
  | openRange

  ;

closedRange :

  #( ClosedRange constant constant )

  ;

openRange :

  #( OpenRange (   ( constant ) => constant   // constant may be a boolean expression
                 | #( OpEqual constant )
                 | #( OpNotEqual constant )
                 | #( OpLess constant )
                 | #( OpGreater constant )
                 | #( OpLessOrEqual constant )
                 | #( OpGreaterOrEqual constant ) ) )

  ;

sizeConstraint :

  #( SizeConstraint rangeCondition )

  ;

constant :

  constantExpression

  ;

/* 12.1.9.6 Synonym definition, page 143 */

synonymDefinition :

  #( SynonymDefinition ( synonymDefinitionItem )+ end )

  ;

synonymDefinitionItem :

    externalSynonymDefinitionItem
  | internalSynonymDefinitionItem

  ;

internalSynonymDefinitionItem :

  #( InternalSynonymDefinitionItem Name ( sort )? constantExpression )

  ;

externalSynonymDefinitionItem :

  #( ExternalSynonymDefinitionItem Name /*predefined*/ sort )

  ;

/* 12.2.1 Expressions, page 144,
 * 12.2.4 Extended primary, page 147
 * 12.2.5 Equality expression, page 147
 * 12.2.7 Operation application, page 149 (also kept for reference from 11.14.3)
 * 12.2.8 Range check expression, page 151
 *
 * Complete rewrite due to left recursion
 */

expression :

    ( valueReturningProcedureCall ) =>
    valueReturningProcedureCall
  | #( OpImplies             expression expression )
  | #( OpOr                  expression expression )
  | #( OpXor                 expression expression )
  | #( OpAnd                 expression expression )
  | #( OpGreater             expression expression )
  | #( OpGreaterOrEqual      expression expression )
  | #( OpLess                expression expression )
  | #( OpLessOrEqual         expression expression )
  | #( OpIn                  expression expression )
  | #( OpEqual               expression expression )
  | #( OpNotEqual            expression expression )
  | #( OpPlus                expression expression )
  | #( OpBinaryMinus         expression expression )
  | #( OpConcat              expression expression )
  | #( OpTimes               expression expression )
  | #( OpDivide              expression expression )
  | #( OpMod                 expression expression )
  | #( OpRem                 expression expression )
  | #( OpUnaryMinus          expression )
  | #( OpNot                 expression )
  | #( OpRangeCheck          expression rangeCheck )
  | primary
  | createExpression

  ;

rangeCheck :

    identifier[ SDLName::sort ] constraint
  | sort

  ;


primary :

  #( SomeExpression
     (   expression
       | conditionalExpression
       | spellingTerm
       | structurePrimary                  // taken from 'extendedPrimary'
       | ( operatorApplication ) =>        // taken from 'operationApplication'
         operatorApplication
       | ( fieldName ) =>
         fieldName                         // taken from 'extendedPrimary'; subsumed by 'literal'
       | ( literal ) =>
         literal
       | ( synonym ) =>
         synonym
       | activePrimary )
     ( primaryExtension )* )

  ;

primaryExtension :

    #( SomePrimaryExtension
       (   ( operationIdentifier ) =>
           operationIdentifier ( actualParameters )?
         | /*field*/ Name ) )
  | actualParameters

  ;

fieldName :

  /*field*/ Name   // name should be of type 'field', shouldn't it?

  ;

structurePrimary :

  #( StructurePrimary ( qualifier )? actualParameters )

  ;

/* End of rewrite */

activePrimary :

    variableAccess
  | imperativeExpression

  ;

expressionList :

  ( expression )+

  ;

simpleExpression :

  constantExpression

  ;

constantExpression :

  /*constant*/ expression

  ;

/* 12.2.2 Literal, page 146 */

literal :

  literalIdentifier

  ;

literalIdentifier :

    #( Identifier qualifier literalName )
  | literalName

  ;

/* 12.2.3 Synonym, page 146 */

synonym :

  identifier[ SDLName::synonym ]

  ;

/* 12.2.6 Conditional expression, page 148 */

conditionalExpression :

  #( ConditionalExpression /*boolean*/ expression consequenceExpression alternativeExpression )

  ;

consequenceExpression :

  expression

  ;

alternativeExpression :

  expression

  ;

/* 12.2.7 Operation application, page 149 */

operationApplication :

  primary

  ;

operatorApplication :

  operationIdentifier ( options { greedy = true; } : actualParameters )?
  // there is NO conflict with 'primaryExtension'

  ;

// 'methodApplication' has been moved into the definition of 'primaryExtension'

/* 12.3.1 Variable definition, page 151 */

variableDefinition :

  #( VariableDefinition ( Exported )? ( variablesOfSort )+ end )

  ;

variablesOfSort :

  #( VariablesOfSort sort ( Name ( exportedAs )? )+ ( expression )? )

  ;

exportedAs :

  #( ExportedAs identifier[ SDLName::remoteVariable ] )

  ;

/* 12.3.2 Variable access, page 152 */

variableAccess :

    identifier[ SDLName::variable ]
  | This

  ;

/* 12.3.3   Assignment and assignment attempt, page 152 &
 * 12.3.3.1 Extended variable, page 153
 *
 * Complete rewrite due to left recursion
 */

assignment :

  #( Assignment variable expression )

  ;

variable :

  #( Variable identifier[ SDLName::variable ] ( variableExtension )* )

  ;

variableExtension :

    actualParameters
  | fieldName

  ;

/* 12.3.3.2 Default initialization, page 154 */

defaultInitialization :

  #( DefaultInitialization ( virtuality )? ( constantExpression )? ( end )? )

  ;

/* 12.3.4 Imperative expressions, page 154 */

imperativeExpression :

    nowExpression
  | importExpression
  | pidExpression
  | timerActiveExpression
  | anyExpression
  | stateExpression

  ;

/* 12.3.4.1 Now expression, page 155 */

nowExpression :

  NowExpression

  ;

/* 12.3.4.3 Pid expression, page 155 */

pidExpression :

    Self
  | Parent
  | Offspring
  | Sender

  ;

createExpression :

  createRequest

  ;

/* 12.3.4.4 Timer active expression, page 156 */

timerActiveExpression :

  #( TimerActiveExpression identifier[ SDLName::timer ] ( expressionList )? )

  ;

/* 12.3.4.5 Any expression, page 156 */

anyExpression :

  #( AnyExpression sort )

  ;

/* 12.3.4.6 State expression, page 157 */

stateExpression :

  StateExpression

  ;

/* 12.3.5 Value returning procedure call, page 157 */

valueReturningProcedureCall :

    #( SomeProcedureCall
       (   ( procedureCallBody ) =>
           procedureCallBody
         | remoteProcedureCallBody ) )
  | ( procedureCallBody ) =>
    procedureCallBody
  | remoteProcedureCallBody

  ;

/* 13.1 Optional definition, page 158 */

selectDefinition :

  #( SelectDefinition /*boolean*/ simpleExpression  end
                        (   agentTypeReference
                          | dataTypeReference
                          | compositeStateTypeReference
                          | procedureReference
                          | agentTypeDefinition
                          | agentDefinition
                          | channelDefinition
                          | signalDefinition
                          | signalListDefinition
                          | remoteVariableDefinition
                          | remoteProcedureDefinition
                          | dataDefinition
                          | procedureDefinition
                          | compositeStateTypeDefinition
                          | statePartitioning
                          | timerDefinition
                          | variableDefinition
                          | channelToChannelConnection
                          | selectDefinition
                          | exceptionDefinition )+ end )

  ;

/* 13.2 Optional transition string, page 159 */

transitionOption :

  #( TransitionOption alternativeQuestion end decisionBody )

  ;

alternativeQuestion :

    simpleExpression
  // | informalText

  ;
